/*
 * File:	PiePopup.c
 *
 * Purpose: Contains the function DrawPieMenu
 *			that is callable from HyperCard via an 
 *			XCMD interface.
 */
 
#include "circledata.h" /* circle definitions */


/* character separating menu string items */
#define	MENU_SEPARATOR		';'
/* 1/2 the width of the popup menu */
#define	ONE_HALF_POPUP_X	180
/* 1/2 the height of the popup menu */
#define	ONE_HALF_POPUP_Y	80
/* minimum # of pixels to have around popup */
#define	BORDER_SIZE			4
/* Indicates no region is currently selected */
#define	NO_REGION			-1
/* # of pixels in Mac's menu bar */
#define	MENUBAR_HEIGHT		20

/*
 * Definitions used in warping the mouse.
 */
#define	RawMouse		(Point *) 0x82c
#define	MTemp			(Point *) 0x828
#define	CrsrNew			(char *)  0x8CE
#define	CrsrCouple		(char *)  0x8CF


/* Local Function Prototypes */
int DrawPieMenu(Point,char*,int,int,int,int, int);
Rect  GetMenuRect(Point, int, WindowPtr *);
char *strcpy(char *dest, char *src);
int   strlen(char *str);


/* 
 * Number of different pie wedge patterns and how
 * many different ones to use when displaying
 * menus containing an even/odd number of menu
 * items.  The intent is to not have the first and
 * last pattern used be the same making the two
 * wedges look like one big wedge.
 */
#define	NUM_PIE_PATTERNS	5
/* # of patterns to use with an odd # of items */
#define	NUM_ODD_PATTERNS	5
/* # of patterns to use with an even # of items */
#define	NUM_EVEN_PATTERNS	4

/* The actual patterns used for the pie wedges */
long	pattern[NUM_PIE_PATTERNS][2] =
{
		0x77DD77DD, 0x77DD77DD,	/* Dark Grey */
		0x88228822, 0x88228822,	/* Light Grey */
		0xAA55AA55, 0xAA55AA55,	/* Grey */
		0xCC33CC33, 0xCC33CC33,	/* Textured */
		0xAAAAAAAA, 0xAAAAAAAA,	/* Striped */
};

/*
 * We need to have our own black pattern because
 * the normal pen pattern of black is referenced
 * off A5 as part of the QuickDraw globals. Since
 * this is a code resource, it's built referencing
 * globals off A4 and doesn't have access to the
 * A5-based QuickDraw globals.
 */
long	black_pattern[2] = {0xFFFFFFFF,0xFFFFFFFF};


/* Maximum number of pie wedges allowed */
#define	MAX_MENU_ITEMS	40

/*
 * Rectangles used to hold the coordinates of the
 * pie wedge label's rectangles
 */
struct	{
	Rect	r;
} label_rects[MAX_MENU_ITEMS];

/* 
 * Region table used to hold the region handles
 * for the pie wedges.
 */
RgnHandle	pie_rgns[MAX_MENU_ITEMS];

/*
 * The region for the center "non-hot" region is
 * treated separately.
 */
RgnHandle	center_rgn;


/*
 * Function:DrawPieMenu()
 *
 * Outputs: Displays a pie menu for item selection
 *
 * Returns:The # of the selected menu item or zero
 *
 */
int
DrawPieMenu(
	Point	windowLoc,		/* popup location     */
	char 	*menuList,		/* list of menu items */
	int		num_menu_items,	/* # of menu items    */
	int		font,			/* font to display    */
	int		font_size,	 	/* font point size    */
	int		use_patterns,	/* use/dont use ptrns */
	int		warp_mouse)	 	/* do/dont allow warp */
{
	Rect		popupRect;   /* menu's rectangle  */
	WindowPtr	origWindow;  /* original grafport */
	WindowPtr	pieWindow;   /* ptr to pie popup  */
	WindowRecord  window_rec;/*pie popup window   */
	int		num_patterns;	 /* # patterns to use */
	int		selection;   	 /* user's selection  */	

	
	/*
	 * Before doing any work, ensure the mouse is
	 * still depressed.
	 */
	if (! StillDown())
		return (0);

	/* Nothing to do if the menu is empty */
	if (num_menu_items < 1)
		return (0);
	
	/* Save the grafport of the current window */
	GetPort(&origWindow);

	/* Compute the enclosing rectangle. */
	popupRect = GetMenuRect(windowLoc,
							warp_mouse,
						    &origWindow);
		
	/* 
	 * Open a dialog box window to place the pie
	 * menu within, and make it current.
	 */				
	pieWindow = NewWindow((Ptr) &window_rec,
						  &popupRect,
						  "\pPieMenu",
						  true,
						  altDBoxProc,
						  (WindowPtr) -1L,
						  true,
						  0);
	SetPort(pieWindow);

	/*
	 * Construct the array of regions that
	 * represent pie wedges.
	 */
	BuildPieRegions(&num_menu_items,
					use_patterns,
					&num_patterns);

	/* Display the menu on the screen. */
	DrawTheMenu(menuList,
				num_menu_items,
				&num_patterns,
				font,
				font_size,
				use_patterns);

	/* Obtain the user's menu choice. */
	selection = GetMenuPick(use_patterns,
							num_menu_items,
							num_patterns);

	/* Dispose of the pie wedge regions */
	DisposePieRegions();

	/* Switch back to the original grafport */
	SetPort(origWindow);
	CloseWindow(pieWindow);

	return selection;
}


Rect
GetMenuRect(
	Point		windowLoc,	/* popup location  */
	int			warp_mouse,	/* do/dont warp    */
	WindowPtr	*origWindow)/* window coord    */
{
	Point		mouseLoc;	/* mouse location  */
	Point		oldmouseLoc;/* save mouse loc  */
	Rect 		popupRect;
	short		scrn_left;	/* left of screen  */
	short		scrn_right;	/* right of screen */
	short		scrn_top;	/* top of screen   */
	short		scrn_bottom;/* bot of screen   */


	/* Obtain the dimensions of the screen. */
	scrn_left = 
			(*origWindow)->portBits.bounds.left;
	scrn_right = 
			(*origWindow)->portBits.bounds.right;
	scrn_top = 
			(*origWindow)->portBits.bounds.top;
	scrn_bottom = 
			(*origWindow)->portBits.bounds.bottom;
	

	/*
	 * Obtain the current location of the mouse to
	 * center the popup around.
	 */
	GetMouse(&mouseLoc);
	oldmouseLoc = mouseLoc;
	mouseLoc.h += windowLoc.h;
	mouseLoc.v += windowLoc.v;


	/* 
	 * Make sure the pie popup will fit on the
	 * screen when centered about the cursor
	 * location, adjust coordinates if necessary
	 * to keep the entire popup menu visible.
	 */
	if (mouseLoc.h <= 
		(scrn_left + ONE_HALF_POPUP_X + BORDER_SIZE))
	  mouseLoc.h = 
		scrn_left + ONE_HALF_POPUP_X + BORDER_SIZE;
	
	if ((mouseLoc.h + ONE_HALF_POPUP_X + BORDER_SIZE)
		>= scrn_right)
	  mouseLoc.h = 
		scrn_right - (ONE_HALF_POPUP_X + BORDER_SIZE);
	
	if (mouseLoc.v <= (scrn_top + ONE_HALF_POPUP_Y + 
					MENUBAR_HEIGHT + BORDER_SIZE))
	  mouseLoc.v = scrn_top + ONE_HALF_POPUP_Y +
				   MENUBAR_HEIGHT + BORDER_SIZE;
	
	if ((mouseLoc.v + ONE_HALF_POPUP_Y + BORDER_SIZE)
		 >= scrn_bottom )
	  mouseLoc.v = 
		scrn_bottom - (ONE_HALF_POPUP_Y + BORDER_SIZE);
	

	/* 
	 * Reposition the mouse, if necessary and
	 * requested.
	 */
	if (warp_mouse) {
		*RawMouse = *MTemp = mouseLoc;
		*CrsrNew = *CrsrCouple;
	}


	/* 
	 * Create a rectangle for the popup using the
	 * adjusted coordinates
	 */
	SetRect(&popupRect, 
			mouseLoc.h - ONE_HALF_POPUP_X,
			mouseLoc.v - ONE_HALF_POPUP_Y,
			mouseLoc.h + ONE_HALF_POPUP_X,
			mouseLoc.v + ONE_HALF_POPUP_Y);
	
	return popupRect;

}


BuildPieRegions(
	int	*num_menu_items,
	int	use_patterns)
{
	int	i;		
	
	/* 
	 * The region points headers assume a 0,0
	 * based coordinate system
	 */
	SetOrigin(-(ONE_HALF_POPUP_X), 
			  -(ONE_HALF_POPUP_Y));

	/*
	 * Create the center "non-hot" region in the
	 * pie menu when using patterns.
	 */
	if (use_patterns) {
		center_rgn = NewRgn();
		PenSize(1,1);
		PenPat((Pattern *) black_pattern);
		MoveTo(c_circle_points[0].x,
			   c_circle_points[0].y);
		OpenRgn();
		for (i = 1; i < NUM_C_POINTS; i++) {
			LineTo(c_circle_points[i].x,
				   c_circle_points[i].y);
		}
		CloseRgn(center_rgn);
		if (use_patterns)
			FrameRgn(center_rgn);
	}
	
	/*
	 * Dissallow more menu items than we have
	 * regions to display.
	 */
	if (*num_menu_items > MAX_MENU_ITEMS)
		*num_menu_items = MAX_MENU_ITEMS;
		
		
	/*
	 * Create as many regions as their are pie
	 * slices to display.
	 */
	for (i = 0; i < *num_menu_items; i++)
		pie_rgns[i] = NewRgn();
}


DrawTheMenu(
	char 	*menuList,
	int		num_menu_items,
	int		*num_patterns,
	int		font,	
	int		font_size,
	int		use_patterns)
{
	int		line_start;			/* start of wedge */
	int		line_end;			/* label end point */
	int		points_this_i_rgn;  /* # inr rgn pts */
	int		points_this_o_rgn;  /* # outr rgn pts*/
	int		i_pts;				/* tot inr points */
	int		o_pts;				/* tot outr pts */	
	int		extra_i_pts;		/* used to spread */	
	int		extra_o_pts;		/*  out extra pts */
	int		next_i_pt;			/* next inr point */
	int		next_o_pt;			/* next outr point */
	char	*cp;				/* temp pointer */
	int	i,j;					/* temp indexes */
	Point	start_pt;			/* label start pt */	
	Point	end_pt;				/* label end pt */
	char	label_buf[80];		/* temp label buf */
	short	label_width;		/* label x pixels */	
	int		label_len;			/* # chars in label*/


	/* 
	 * Calculate the number of points to use in
	 * defining a region.
	 */
	i_pts   	= NUM_I_POINTS / num_menu_items;
	extra_i_pts = NUM_I_POINTS % num_menu_items;
	o_pts  		= NUM_O_POINTS / num_menu_items;
	extra_o_pts = NUM_O_POINTS % num_menu_items;

	/*
	 * Regions have no width so we need to overlap
	 * with next regions first point.
	 */
	++i_pts;
	++o_pts;

	/*
	 * Pick a number of patterns that won't cause
	 * 1st pattern to be equal last.
	 */
	if (num_menu_items % 2)
		*num_patterns = NUM_ODD_PATTERNS;
	else
		*num_patterns = NUM_EVEN_PATTERNS;

	/*
	 * Initialize next points and next menu
	 * selection to use.
	 */
	next_i_pt = 0;
	next_o_pt = 0;
	cp = menuList;
	
	/* Plot the regions. */
	for (i = 0; i < num_menu_items; i++) {
		points_this_i_rgn = i_pts;
		if (extra_i_pts) {
			++points_this_i_rgn;
			--extra_i_pts;
		}
		
		points_this_o_rgn = o_pts;
		if (extra_o_pts) {
			++points_this_o_rgn;
			--extra_o_pts;
		}
		
		/*
	     * Draw a line from the center of this arc
		 * to the outer circle.
		 */
		line_start =
				(points_this_i_rgn / 2) + next_i_pt;
		line_end = 
				(points_this_o_rgn / 2) + next_o_pt;
		PenSize(1,1);
		MoveTo(i_circle_points[line_start].x,
			   i_circle_points[line_start].y);
		LineTo(o_circle_points[line_end].x,
			   o_circle_points[line_end].y);
		
		/*
		 * Put a horizontal tab 6 pixels long on
		 * the radiating line.
		 */
		if ( i < (num_menu_items / 2) )
			LineTo(o_circle_points[line_end].x + 6,
				   o_circle_points[line_end].y);
		else
			LineTo(o_circle_points[line_end].x - 6,
				   o_circle_points[line_end].y);
		next_o_pt += points_this_o_rgn;
		
		/* Set the typeface and style for text. */
		TextFont(font);
		TextFace(0);
		TextSize(font_size);
				
		/* Isolate next label and display it. */
		j = 0;

		while ((*cp != MENU_SEPARATOR) && 
			   (*cp != '\0'))
			label_buf[j++] = *cp++;

		label_buf[j] = '\0';
		++cp;			

		/*
		 * Calculate the number of pixels contained
		 * in this label.
		 */
		label_len = strlen(label_buf);
		label_width = 
				TextWidth(label_buf,0,label_len);
		
		/*
		 * Calculate where the start point of the
		 * label should be located.
		 */
		if ( i < (num_menu_items / 2) ) {
			SetPt(&start_pt,
				  o_circle_points[line_end].x + 10,
				  o_circle_points[line_end].y + 4);
		}
		else {
			SetPt(&start_pt,
				  o_circle_points[line_end].x -
									(8+label_width),
				  o_circle_points[line_end].y + 4);
		}

		/*
		 * The end point is to the left of the
		 * start by the length of the label.
		 */
		SetPt(&end_pt,
			  start_pt.h + label_width,
			  start_pt.v);
		
		MoveTo(start_pt.h,start_pt.v);	
		DrawText(label_buf,0,label_len);

		/*
		 * Calculate the rectangle that encloses
		 * the label
		 */		
		SetRect((Rect *) &label_rects[i],
				start_pt.h - 3, 
				start_pt.v - font_size, 
				end_pt.h + 3,
				end_pt.v + 3);
				
				
		/*
		 * Open the next region, assign points,
		 * close it, fill it, remove center.
		 */
		MoveTo(0,0);
		PenSize(1,1);
		OpenRgn();
		for (j = points_this_i_rgn;  j != 0;  
			 j--, next_i_pt++)

			LineTo(i_circle_points[next_i_pt].x,
				   i_circle_points[next_i_pt].y);
	
		LineTo(0,0);
		CloseRgn(pie_rgns[i]);
		if (use_patterns) {
			DiffRgn(pie_rgns[i],
					center_rgn,pie_rgns[i]);
			FillRgn(pie_rgns[i], (Pattern *) 
					pattern[i % (*num_patterns)]);
		}
		else
			 FrameRgn(pie_rgns[i]);

		/*
		 * The next region actually starts on our
		 * ending point.
		 */
		--next_i_pt;
		--next_o_pt;
	}
}


GetMenuPick(
	int	use_patterns,
	int	num_menu_items,
	int	num_patterns)
{
	Point	mouseLoc;	/* current loc of mouse */
	int		old_rgn;	/* previous rgn selection */
	int		in_a_region;/* TRUE iff crsr in pie */
	int		i;			/* general purpose index */

	/* Indicate that we aren't in any region */
	old_rgn = NO_REGION;		

	/*
	 * If we're not using patterns put a 2 pixel
	 * border around the wedges.
	 */
	if (! use_patterns) {
		for (i = 0; i < num_menu_items; i++)
			InsetRgn(pie_rgns[i], 2, 2);
	}
	
	/*
	 * Loop, tracking the mouse until the user
	 * releases the button.
	 */
	while (Button()) {
		SystemTask();
		in_a_region = FALSE;
		GetMouse(&mouseLoc);
		for (i = 0; i < num_menu_items; i++) {
		  if (PtInRgn(mouseLoc, pie_rgns[i]) ) {
			in_a_region = TRUE;
			if (old_rgn == i) 
			  break;
			else {
			  in_a_region = TRUE;
			  PaintRgn(pie_rgns[i]);
			  InvertRect((Rect *) &label_rects[i]);
			  if (old_rgn != NO_REGION) {
				if (use_patterns)
				  FillRgn(pie_rgns[old_rgn],
						  (Pattern *)
						  pattern[old_rgn %
								      num_patterns]);
				else
				  InvertRgn(pie_rgns[old_rgn]);
			    InvertRect((Rect *)
							  &label_rects[old_rgn]);
			  }
			  old_rgn = i;
			}
		  }
		}
		
		/*
		 * Deselect the previously selected region
		 * when we exit the pie regions
		 */
		if ((in_a_region == FALSE) && 
		    (old_rgn != NO_REGION) ) {
		  if (use_patterns)
			FillRgn(pie_rgns[old_rgn],
			  (Pattern *) 
				  pattern[old_rgn % num_patterns]);
		  else
			InvertRgn(pie_rgns[old_rgn]);
		  InvertRect((Rect *) &label_rects[old_rgn]);
		  old_rgn = NO_REGION;
		}
	}

	/* Return the index of the selected item */
	return ((old_rgn == NO_REGION) ? 0 : ++old_rgn);
}


DisposePieRegions(
	int		num_menu_items,
	int		use_patterns)
{
	int		i;			
	
	
	/* Dispose of the pie regions */
	for (i = 0; i < num_menu_items; i++)
		DisposeRgn(pie_rgns[i]);
		
	/* Dispose of the center 'non-hot' region */
	if (use_patterns)
		DisposeRgn(center_rgn);
}

