/**********************************************************************
 * PatchGNE.c
 *
 * This INIT installs a patch on GetNextEvent and SystemEvent
 * that intercepts keyDown and autoKey events. For this example,
 * the intercepted key events are converted to lower case if both
 * the capsLock key and the shiftKey are down (thus making the
 * Mac keyboard behave like an IBM keyboard). However, you can
 * use this shell to do generalized event intercepting as well
 * as generalized trap patching (with no asm and no self-modifying
 * code). If your patches need globals, put them in the PatchGlobals
 * struct and initialize them in main.
 *
 * In Think C, set the Project Type to Code Resource, the File Type
 * to INIT, the Creator to anything, the Type to INIT, the ID to
 * something like 55 (55 will work but it doesn't have to be 55),
 * turn Custom Header ON and Attrs to 20 (purgeable) and Multi
 * Segment OFF.
 *
 * Mike Scanlin. 16 May 1992.
 **********************************************************************/

#include "Traps.h"

/**********************************************************************
 * typedefs
 **********************************************************************/
typedef pascal short (*GNEProcPtr)(short eventMask, EventRecord *theEvent);
typedef pascal short (*SEProcPtr)(EventRecord *theEvent);

typedef struct PatchGlobals {
    GNEProcPtr      pgOldGNE;
    SEProcPtr       pgOldSE;
} PatchGlobals, *PatchGlobalsPtr;


/**********************************************************************
 * prototypes
 **********************************************************************/
void main(void);
void StartPatchCode(void);
pascal short MyGetNextEvent(short eventMask, EventRecord *theEvent); pascal
short MySystemEvent(EventRecord *theEvent);
void CheckKeyCase(EventRecord *theEvent);
void EndPatchCode(void);


/**********************************************************************
 * main
 *
 * Gets some memory in the system heap and installs the GNE and SE
 * patches (as well as allocating and initializing the patch globals).
 * This is the only routine that gets executed at startup time
 * (by the INIT mechanism).
 *
 * The block of memory that main allocates will look like this
 * when main has finished:
 *
 *                   +--------------------+
 *                   |    PatchGlobals    |
 *                   +--------------------+
 *                   |  StartPatchCode()  |
 *  GNE trap addr -> +--------------------+
 *                   |  MyGetNextEvent()  |
 *   SE trap addr -> +--------------------+
 *                   |  MySystemEvent()   |
 *                   +--------------------+
 *                   |   CheckKeyCase()   |
 *                   +--------------------+
 *                   |   EndPatchCode()   |
 *                   +--------------------+
 *
 **********************************************************************/
void main()
{
    Ptr             patchPtr;
    PatchGlobalsPtr pgPtr;
    long            codeSize, offset;

    /* try and get some memory in the system heap for code and globals */
    codeSize = (long) EndPatchCode - (long) StartPatchCode;
    patchPtr = NewPtrSys(codeSize + sizeof(PatchGlobals));
    if (!patchPtr)
        return; /* out of memory -- abort patching */

    /* initialize the patch globals at the beginning of the block */
    pgPtr = (PatchGlobalsPtr) patchPtr;
    pgPtr->pgOldGNE = (GNEProcPtr) GetTrapAddress(_GetNextEvent);
    pgPtr->pgOldSE = (SEProcPtr) GetTrapAddress(_SystemEvent);

    /* move the code into place after the globals */
    BlockMove(StartPatchCode, patchPtr + sizeof(PatchGlobals), codeSize);

    /* set the patches */
    patchPtr += sizeof(PatchGlobals);
    offset = (long) MyGetNextEvent - (long) StartPatchCode;
    SetTrapAddress((long) patchPtr + offset, _GetNextEvent);
    offset = (long) MySystemEvent - (long) StartPatchCode;
    SetTrapAddress((long) patchPtr + offset, _SystemEvent);
}


/**********************************************************************
 * StartPatchCode
 *
 * Dummy proc to mark the beginning of the code for the patches.
 * Make sure all of your patch code is between here and EndPatchCode.
 **********************************************************************/
void StartPatchCode()
{
}


/**********************************************************************
 * MyGetNextEvent
 *
 * Tail patch on GetNextEvent.
 *
 * The reason this returns a short instead of a Boolean is because
 * we need to make sure the low byte of the top word on the stack is zero
 * because some programs do a Tst.W (SP)+ when this returns instead
 * of Tst.B (SP)+ like they should (which is technically their bug but,
 * we might as well work around it since it's not hard).
 *
 * If you want to eat the event and not pass it on to the caller
 * then set returnValue to zero.
 **********************************************************************/
pascal short MyGetNextEvent(short eventMask, EventRecord *theEvent)
{
    PatchGlobalsPtr pgPtr;
    short           returnValue;

    /* find our globals */
    pgPtr = (PatchGlobalsPtr) ((long) StartPatchCode - sizeof(PatchGlobals));

    /* call original GNE first */
    returnValue = (*pgPtr->pgOldGNE)(eventMask, theEvent);

    /* do some post-processing */
    CheckKeyCase(theEvent);

    /* return to original caller */
    return (returnValue);
}


/**********************************************************************
 * MySystemEvent
 *
 * Tail patch on SystemEvent.
 *
 * The reason this returns a short instead of a Boolean is because
 * we need to make sure the low byte of the top word on the stack is zero
 * because some programs do a Tst.W (SP)+ when this returns instead
 * of Tst.B (SP)+ like they should (which is technically their bug but,
 * we might as well work around it since it's not hard).
 *
 * We need this patch as well as the one on GetNextEvent because of
 * desk accessories. If you don't patch SystemEvent then the patch
 * will not apply to events that are sent to DAs.
 *
 * If you want to eat the event and not pass it on to the caller
 * then set returnValue to zero.
 **********************************************************************/
pascal short MySystemEvent(EventRecord *theEvent)
{
    PatchGlobalsPtr pgPtr;
    short           returnValue;

    /* find our globals */
    pgPtr = (PatchGlobalsPtr) ((long) StartPatchCode - sizeof(PatchGlobals));

    /* call original GNE first */
    returnValue = (*pgPtr->pgOldSE)(theEvent);

    /* do some post-processing */
    CheckKeyCase(theEvent);

    /* return to original caller */
    return (returnValue);
}


/**********************************************************************
 * CheckKeyCase
 *
 * If theEvent was a keyDown or autoKey event, this checks if both
 * the shiftKey and the capsLock key were down. If so, it changes
 * theEvent to be a lowercase letter. If not, nothing is changed.
 * Also, if either the optionKey or cmdKey is down then nothing is
 * changed.
 **********************************************************************/
void CheckKeyCase(EventRecord *theEvent)
{
    register long   theMods, theMessage;
    register char   theChar;

    if (theEvent->what == keyDown || theEvent->what == autoKey) {
        theMods = theEvent->modifiers;
        theMods &= shiftKey | alphaLock | optionKey | cmdKey;
        theMods ^= shiftKey | alphaLock;
        if (!theMods) {
            theMessage = theEvent->message;
            theChar = theMessage & charCodeMask;
            if (theChar >= 'A' && theChar <= 'Z') {
                theMessage &= ~charCodeMask;
                theMessage |= theChar + 'a' - 'A';
                theEvent->message = theMessage;
            }
        }
    }
}


/**********************************************************************
 * EndPatchCode
 *
 * Dummy proc to mark the end of the code for the patches.
 * Make sure all of your patch code is between here and StartPatchCode.
 **********************************************************************/
void EndPatchCode()
{
}

