// 
// calc.cp
//
// a simple calculator program
// based on an example from Dewhurst/Stark "Programming in C++"
//
// J. Langowski November 1992
//

#include <Ctype.h>
#include <StdIO.h>
#include <String.h>
#include <Stream.h>
#include <StdLib.h>

#define NIL 0

// basic syntax tree structure
//
class node {
	protected:
		node() {}			
	public:
		virtual ~node() {}
		virtual int set (int)
			{ cout << "Error: node::set(int) undefined" << endl; return 0;} 
		virtual int eval() 
			{ cout << "Error: node::eval() undefined" << endl; return 0;} 
};

class dyad : public node {
	protected:
		node *left, *right;
		dyad(node *l, node *r)	{left=l; right=r;}
		~dyad() {delete left; delete right;}
};

// operators
//
class plus : public dyad {
	public:
		plus(node *l, node *r) : dyad(l,r) {}
		int eval() { return left->eval() + right->eval();} 
};

class minus : public dyad {
	public:
		minus(node *l, node *r) : dyad(l,r) {}
		int eval() { return left->eval() - right->eval();}
};

class times : public dyad {
	public:
		times(node *l, node *r) : dyad(l,r) {}
		int eval() { return left->eval() * right->eval();}
};

class divide : public dyad {
	public:
		divide(node *l, node *r) : dyad(l,r) {}
		int eval() { return left->eval() / right->eval();}
};

class uminus : public node {
		node *operand;
	public:
		uminus(node *o) {operand = o;}
		~uminus() {delete operand;}
		int eval() {return -operand->eval();}
};

class inumber : public node {
		int value;
	public:
		inumber(int v) {value = v;}
		int eval() {return value;}
};


// identifier table
//
class id;

class entry {
	char *name;
	int value;
	entry *next;
	entry (char *nm, entry *n) {
		name = strcpy (new char[strlen(nm) + 1], nm);
		value = 0;
		next = n;
	}
	friend id;
};

class id : public node {
		static entry *symtab;
		entry *ent;
		entry *look(char *);
	public:
		id(char *nm) {ent = look(nm);}
		int set(int i) {return ent->value = i;}
		int eval() {return ent->value;}
};

entry *id::look(char *nm) {
	for(entry *p = symtab; p; p = p->next)
		if(strcmp(p->name,nm) == 0) return p;
	return symtab = new entry(nm,symtab);
}

entry *id::symtab = NIL;

// assignment
//
class equals : public dyad {
	public:
		equals(node *t,node *e) : dyad(t,e) {}
		int eval()
			{ return left->set(right->eval()); }
};


// parsing + evaluation
//

static char token;	// current token 
static char line[81];		// for reading identifiers + numbers
enum {ID = char(128), INT, EOLN, BAD}; // special tokens

node *e(), *t(), *f();		// parser routines 
							// for expression, term, factor

// input stream scanner, returns next token
//
char scan() {
	char c;
		while (1)
			switch (c = cin.get()) {
			case '+': case '-': case '*': case '/':
			case '(': case ')': case '=':
				return c;
			case ' ': case '\t':
				continue;
			case '\n': case '\r':
				return EOLN;
			default:
				if (isdigit(c)) {
					char *s = line;
					do	*s++ = c;
						while(isdigit(c = cin.get()));
					*s = '\0'; // terminate string just read
					cin.putback(c); // had read one too much
					return INT;
				}
				
				if (isalpha(c)) {
					char *s = line;
					do	*s++ = c;
						while(isalnum(c = cin.get()));
					*s = '\0'; // terminate string just read
					cin.putback(c); // had read one too much
					return ID;
				}
				
				return BAD; // if nothing fits, syntax error
			}
}

// simple error routine
//
void error() { cout << "Syntax error." << endl; exit(255); }

// parser routines for expression, term, factor
//
node *e() {
	node *root = t();
	while (1)
		switch (token) {
		case '+':
			token = scan();
			root = new plus(root, t()); 
			break;
		case '-':
			token = scan();
			root = new minus(root, t());
			break;
		default:
			return root;
		}
}

node *t() {
	node *root = f();
	while (1)
		switch (token) {
		case '*':
			token = scan();
			root = new times(root, t()); 
			break;
		case '/':
			token = scan();
			root = new divide(root, t()); 
			break;
		default:
			return root;
		}
}

node *f() {
	node *root = NIL;
	switch (token) {
	case ID:
		root = new id(line);
		token = scan();
		if (token == '=') {
			token = scan();
			root = new equals(root, e());
		}
		return root;
	case INT:
		root = new inumber(atoi(line));
		token = scan();
		return root;
	case '(':
		token = scan();
		root = e();	
		if (token != ')' ) error();
		token = scan();
		return root;
	case '-':
		token = scan();
		return new uminus(f());
	default:
		error();
	}
}

// main program
//

void main() {
	node *root = NIL;
	while (1) {
		cout << "Enter expression: " << endl;
		token = scan();
		root = e();

		if (token == BAD) error();
		
		if (root != NIL)
			{ 	cout << "Result = " << root->eval() << endl;
				delete root; 
			}
	}
}




			
					
			





