/* Hiccup.c - allocate handles in lumps to minimize
time spent moving memory around.
Usage:
-call HicNew(), passing address of an HH (hP), desired
 starting size and increment size
-if entries will be all the same size, set hP->dataSize
-if entries will be all from the same (locked) place, set
 hP->dataPtr
-call any of HicAdd(), HicAddPtr(), HicAddPtrSize()
 repeatedly to add entries. If data is being added from
 an existing handle, make sure it's locked!
-when done adding call HicClose()
-note entries are always added at end of handle.
-"dataSize" can be varied, and "inc" need not be larger
than "dataSize"
-to restart with existing handle;
	set up HH with h, realSize, inc, freePos,
	and then proceed as above.
-to start with a dummy entry, call HicNew() followed by
	hP->freePos += size of dummy entry,
	or Add a dummy entry.
-Dispos of a handle in the usual way, or call HicDispose.
-comment out "H_FULL_CHECKING" after final testing
*/

#define H_FULL_CHECKING // if defined, check more errors
// struct "HH" and function prototypes for hiccups
#include "Hiccup.h"

// Just loads segment - sometimes handy while testing
void InitHiccup(void);
void InitHiccup(void) { ; }

// Create Handle hP->h, init tracking variables.
Boolean HicNew(HHPtr hP, long startSize, long increment)
	{
	if (startSize < 0L || increment <= 0L)
		return FALSE;
	hP->h = NewHandle(startSize);
	if (MemError() != noErr)
		return FALSE;
	hP->realSize = startSize;
	hP->freePos = 0;
	hP->inc = increment;
	return TRUE;
	}

/* Add data to the Handle hP->h. Use when  both  hP->dataPtr
and hP->dataSize have been set and  won't  change.  Note  if
hP->dataPtr is NULL then handle size will be increased,  and
hP->freePos bumped by dataSize - for occasions when size  is
known but copying data must be delayed.
*/
Boolean HicAdd(HHPtr hP)
	{
#ifdef H_FULL_CHECKING
	long	first4, laterFirst4;
	Boolean checkPlease = FALSE, movedIt = FALSE;
	
	if (!hP->h || hP->inc <= 0L)
		return FALSE;
	if (hP->dataSize < 0L)
		return FALSE;
	else if (hP->dataPtr && hP->dataSize >= 4L)
		{
		BlockMove(hP->dataPtr, &first4, 4L);
		checkPlease = TRUE;
		}
#endif

	while (hP->freePos + hP->dataSize > hP->realSize)
		{
		SetHandleSize(hP->h, hP->realSize + hP->inc);
		if (MemError() != noErr)
			return FALSE;
		hP->realSize += hP->inc;
#ifdef H_FULL_CHECKING
		movedIt = TRUE;
#endif
		}

#ifdef H_FULL_CHECKING
	if (checkPlease && movedIt)
		{
		BlockMove(hP->dataPtr, &laterFirst4, 4L);
		if (laterFirst4 != first4)
			return FALSE;
		}
#endif

	if (hP->dataPtr)
		BlockMove(hP->dataPtr, *(hP->h) + hP->freePos,
					hP->dataSize);
	hP->freePos += hP->dataSize;
	return TRUE;
	}

// Use when hP->dataSize is set and won't change.
Boolean HicAddPtr(HHPtr hP, void *dataPtr)
	{
	Boolean result;
	
	hP->dataPtr = dataPtr;
	result = HicAdd(hP);
	return result;
	}

// Use when both dataPtr and dataSize vary.
Boolean HicAddPtrSize(HHPtr hP, void *dataPtr,
	long dataSize)
	{
	Boolean result;
	
	hP->dataPtr = dataPtr;
	hP->dataSize = dataSize;
	result = HicAdd(hP);
	return result;
	}

// Vital to call at end, shrinks Handle hP->h.
void HicClose(HHPtr hP)
	{
	
	SetHandleSize(hP->h, hP->freePos);
	hP->realSize = hP->freePos;
	}

void HicDispose(HHPtr hP)
	{
	if (hP->h)
		DisposHandle(hP->h);
	hP->h = (void *)0;
	}

#ifdef H_FULL_CHECKING
#undef H_FULL_CHECKING
#endif
