/* Tree.c : Generic code for handling  trees  with  Handles.
The array of tree nodes is kept in a Handle, as  opposed  to
the usual linked  list  of  pointers.  Data  comparison  and
addition is done by pointers-to-functions (TreeCompProc  and
TreeAddProc respectively) and there are no static variables:
keep around a pointer to each tree in use, and create  Procs
for each kind of tree you need. Only  a  long  dataIndex  is
recorded in each node (TreeNode) for each data  entry  -  it
can be an index or a pointer. The tree proper (Tree) records
a "void *dataHolder" which can be a pointer or  Handle,  and
should exist before calling TreeNew().

Restrictions:
1	<= 32,767 nodes in tree
2	no deletions from tree (infrequent deletions could be
	handled - anyone care to write DeleteFromTree()?)
3	TreeProc, TreeAddProc, and TreeAppendProc functions may
	move memory, but TreeCompProc functions must not.
4	Call TreeClose before calling NextNode or
	ForEachNodeInOrderDo (no need to "reopen" tree after to
	add more nodes).
 ********Three ways to approach adding to a tree:***********
1 data is preloaded into a handle or pointer, and  an  entry
in tree to be made for each data chunk.  In  this  case  the
tree  does  not  need  to  manage  the  data   handle,   and
TreeAddProc should just return a data offset.

2 data not preloaded and may not be unique - load into  data
handle (or linked list etc) only if don't have it  treed  up
already. In this case TreeAddProc should add the data to the
"dataHolder", as well as returning data offset. If  data  is
being added to a Handle it should be added at the end, since
there is no provision for correcting the "dataIndex" in tree
nodes previously created.

3 data not preloaded, multiple data entries may be  required
for a single tree node. In  this  case,  TreeAddProc  should
make the first data entry, and subsequent entries should  be
made by TreeAppendProc. Note TreeAppendProc  returns  a  new
dataIndex to be stored in the tree node.

If multiple data entries for a  node  are  not  needed,  set
TreeAppendProc to NULL.

Loosely based on DD March 88 ~pg 60, which uses linked lists
for the tree nodes  and  is  for  only  one  kind  of  data.
Alpha-order is in-order. Note the root node is set to  point
to a value that is greater than any legal value. The  entire
tree proper hangs from the left node of the root. A positive
left or right index in a  node  indicates  a  true  "child",
whereas if  <=  0  it  is  a  thread.  As  a  special  case,
root->right is a thread to the node with lowest value in the
tree, set in TreeClose().
*/

// #include "Hiccup.h" is done in Tree.h
#define ACCESS_TREE_DATA
// structs Tree, TreeNode, and prototypes
#include "Tree.h"
#undef ACCESS_TREE_DATA
// This file only
#define LEFTCHILD 0
#define RIGHTCHILD 1
static short AddTreeEntry(short which, short old,
	TreePtr tP, void *dataPtr);

/*  Create  handle  for  nodes,  of  a   comfortable   size.
"dataHolder" is for later passing to procs which access data
- it should exist in some form before calling this function.
"numToStart" and "numInc" are measured in nodes, not  bytes.
compProc and addProc are required, appendProc can optionally
be NULL. */
Boolean TreeNew(TreePtr tP, short numToStart, short numInc,
	TreeCompProc compProc, TreeAddProc addProc,
	TreeAppendProc appendProc, void *dataHolder)
	{
	HHPtr		hP = &tP->nodeHH;
	TreeNodePtr	tnp;
	
	tP->compProc = compProc;
	tP->addProc = addProc;
	tP->appendProc = appendProc;
	tP->dataHolder = dataHolder;
	if (!HicNew(hP, sizeof(TreeNode) * numToStart,
				sizeof(TreeNode) * numInc))
		return(FALSE);
	hP->dataSize = sizeof(TreeNode);
	hP->freePos += hP->dataSize; // root, dummy entry
	hP->dataPtr = NULL; // not used
	tnp = (TreeNodePtr)*(hP->h);
	tnp->dataIndex = -1; // = above all legal entries
	tnp->left = tnp->right = 0;
	return(TRUE);
	}

/* Note cmp(data, root) < 0.  Returns  "new"  if  new  entry
added, neg of index if found an  old  one,  0  if  ran  into
trouble. Position 0 is occupied  by  root,  which  is  never
matched. */
short AddToTree(TreePtr tP, void *dataPtr)
	{
	HHPtr			hP = &tP->nodeHH;
	register TreeNodePtr basetnp = (TreeNodePtr)*(hP->h),
					tnp = basetnp;
	register short 	cmp, new, old;
	
	/* comp(a,b) should return similar to strcmp(a,b):
	<0 if data(a) < data(b)
	0 if data(a) == data(b)
	>0 if data(a) > data(b)
	BUT return -1 if data(b) == -1 (this for the root).
	*/
	while ((cmp = (*(tP->compProc))(dataPtr,
				tP->dataHolder, tnp->dataIndex)) != 0)
		{
		if (cmp < 0) // look left
			{
			if (tnp->left > 0)
				tnp = basetnp + tnp->left;
			else
				{
				new = AddTreeEntry(LEFTCHILD, tnp - basetnp,
						tP, dataPtr);
				break;
				}
			}
		else // cmp > 0, look right
			{
			if (tnp->right > 0)
				tnp = basetnp + tnp->right;
			else
				{
				new = AddTreeEntry(RIGHTCHILD, tnp - basetnp,
						tP, dataPtr);
				break;
				}
			}
		}
	if (!cmp) // entry was found
		{
		old = (short)(tnp - basetnp);
		if (tP->appendProc)
			{
			long	oldDataIndex, newDataIndex;
			
			oldDataIndex = tnp->dataIndex;
			newDataIndex =
					(*(tP->appendProc))(tP->dataHolder,
							tnp->dataIndex, dataPtr);
			// ...that may have moved memory...
			if (newDataIndex < 0)
				return(0);
			if (oldDataIndex != newDataIndex)
				{
				tnp = (TreeNodePtr)*(hP->h) + old;
				tnp->dataIndex = newDataIndex;
				}
			}
		return(-old);
		}
	else // new entry was added, maybe
		return(new);
	}

// Note you may also need to HicClose your data handle.
Handle TreeClose(TreePtr tP)
	{
	HHPtr	hP = &tP->nodeHH;
	TreeNodePtr	roottnp = (TreeNodePtr)*(tP->nodeHH.h),
				nexttnp = roottnp;
	short		next;
	//right from root requires special maintenance.
	while (nexttnp->left > 0)
			nexttnp = roottnp + nexttnp->left;
	roottnp->right = -(short)(nexttnp - roottnp);
	HicClose(hP);
	return hP->h;
	}

// Note only nodes are disposed, not data.
void TreeDispose(TreePtr tP)
	{
	HHPtr	hP = &tP->nodeHH;
	
	if (hP->h)
		{
		DisposHandle(hP->h);
		hP->h = NULL;
		}
	}

static short AddTreeEntry(short which, short old,
	TreePtr tP, void *dataPtr)
	{
	HHPtr	hP = &tP->nodeHH;
	short 	new = hP->freePos/hP->dataSize;
	long	dataIndex;
	TreeNodePtr	oldtnp, newtnp;
	
	if (!HicAdd(hP))
		return(0);
	dataIndex = (*(tP->addProc))(tP->dataHolder, dataPtr);
	// ...that may have moved memory...
	if (dataIndex < 0)
		return(0);
	oldtnp = (TreeNodePtr)*(hP->h) + old;
	newtnp = (TreeNodePtr)*(hP->h) + new;
	newtnp->dataIndex = dataIndex;
	if (which == LEFTCHILD)
		{
		newtnp->left = oldtnp->left;
		newtnp->right = -old;
		oldtnp->left = new;
		}
	else
		{
		newtnp->right = oldtnp->right;
		newtnp->left = -old;
		oldtnp->right = new;
		}
	return new;
	}

/* Return index of in-order successor of  oldIndex  in  tree
tP, which is in the same order as set by  your  TreeCompProc
function. For an example,  see  ForEachNodeInOrderDo  below.
Note  TreeClose()  must  be   called   before   using   this
function, to set the root's right pointer. */
short NextNode(TreePtr tP, short oldIndex)
	{
	TreeNodePtr	roottnp = (TreeNodePtr)*(tP->nodeHH.h),
			tnp = roottnp + oldIndex, nexttnp;
	short		next;
	
	if (tnp->right > 0)
		{
		nexttnp = roottnp + tnp->right;
		while (nexttnp->left > 0)
			nexttnp = roottnp + nexttnp->left;
		}
	else
		nexttnp = roottnp - tnp->right;
	next = (short)(nexttnp - roottnp);
	return next;
	}

short PreviousNode(TreePtr tP, short oldIndex)
	{
	TreeNodePtr	roottnp = (TreeNodePtr)*(tP->nodeHH.h),
			tnp = roottnp + oldIndex, nexttnp;
	short		next;
	
	if (tnp->left > 0)
		{
		nexttnp = roottnp + tnp->left;
		while (nexttnp->right > 0)
			nexttnp = roottnp + nexttnp->right;
		}
	else
		nexttnp = roottnp - tnp->left;
	next = (short)(nexttnp - roottnp);
	
	return next;
	}

/* Retrieve index of node matching some criteria; returns
node index or 0. If "tcp" was also used to build the tree
then calling "NextNode" after this makes sense. */
short FirstNodeMatching(TreePtr tP,
	TreeCompProc tcp, void *dataPtr)
	{
	register TreeNodePtr
				roottnp = (TreeNodePtr)*(tP->nodeHH.h),
				tnp = roottnp;
	register short cmp, old;
	
	/* tcp(a,b) should return similar to strcmp(a,b):
	<0 if a < b (or a comes before b)
	0 if a == b
	>0 if a > b
	*/
	while ((cmp = (*tcp)(dataPtr, tP->dataHolder,
					tnp->dataIndex)) != 0)
		{
		if (cmp < 0) /* look left */
			{
			if (tnp->left > 0)
				tnp = roottnp + tnp->left;
			else
				break;
			}
		else /* cmp > 0, look right */
			{
			if (tnp->right > 0)
				tnp = roottnp + tnp->right;
			else
				break;
			}
		}
	if (!cmp) /* entry was found */
		{
		old = (short)(tnp - roottnp);
		return(old);
		}
	else
		return(0);
	}

/* Traverse the tree in order (forwards or  reverse),  doing
"theDoer" on each node. Note "theDoer" may move memory. */
Boolean ForEachNodeInOrderDo(Boolean forwards, 
	TreePtr tP, TreeProc theDoer)
	{
	register TreeNodePtr	tnp;
	register short	index = 0;
	
	if (forwards)
		{
		while (index = NextNode(tP, index))
			{
			tnp = (TreeNodePtr)*(tP->nodeHH.h) + index;
			if (!((*theDoer)(tP->dataHolder, tnp->dataIndex)))
				return FALSE;
			}
		}
	else
		{
		while (index = PreviousNode(tP, index))
			{
			tnp = (TreeNodePtr)*(tP->nodeHH.h) + index;
			if (!((*theDoer)(tP->dataHolder, tnp->dataIndex)))
				return FALSE;
			}
		}
	return TRUE;
	}


