// ** TRegion.cp **
//
// TRegion - a class wrapper for Quickdraw regions
//
// Chris Prinos 1993, Symantec Corp.

#include "TRegion.h"

TRegion::TRegion()				// default constructor
{
	fRgn = new RegionData;
}

TRegion::~TRegion()				// delete if data not referenced by
{								// more than one TRegion
	if ( --fRgn->fCount <= 0 )
		delete fRgn;
}

TRegion::TRegion(const Rect &r)	// build a TRegion from a Rect
{
	fRgn = new RegionData;
	RectRgn(fRgn->fRgnH, &r);
}

TRegion::TRegion(const TRegion &r)	// copy constructor
{
	fRgn = r.fRgn;			// adjust pointers, increment reference 
	fRgn->fCount++;			// count
}

TRegion::TRegion(const RgnHandle r)	// create TRegion from RgnHandle
{
	// assumes r is an allocated Quickdraw region
	fRgn = new RegionData(r); 	
}

TRegion::TRegion(short left, short top, 
	short right, short bottom)
{
	fRgn = new RegionData;			
	SetRectRgn(fRgn->fRgnH, left, top, right, bottom);
}

TRegion::TRegion(const PicHandle p) // build a TRegion from a PicHandle
{
    Rect 		destRect;  
    RgnHandle	tmpRgn;

    destRect = (*p)->picFrame;      // use current picture location
    OpenRgn();
        DrawPicture( p, &destRect );
    CloseRgn(tmpRgn);               // OpenRgn/CloseRgn allocates region
    fRgn = new RegionData(tmpRgn);  // pass the allocated region to RegionData
                                    // ctor
}

TRegion::TRegion(const PolyHandle p) // build a TRegion from a PolyHandle
{
    Rect 		destRect;
    RgnHandle	tmpRgn;
    
    destRect = (*p)->polyBBox;  // use current polygon's frame
    OpenRgn();
        FramePoly( p );
    CloseRgn(tmpRgn);               
    fRgn = new RegionData(tmpRgn);
    
}

void TRegion::Empty()			// clear a region
{
	DetachSharedRgn();
	SetEmptyRgn(fRgn->fRgnH);
}

Boolean TRegion::IsEmpty() const	// TRUE if empty region
{
	return EmptyRgn(fRgn->fRgnH);
}

void TRegion::Offset(short h, short v)	// like OffsetRgn
{
	if (h != 0 && v != 0)		// don't bother if offsets are zero
	{
		DetachSharedRgn();
		OffsetRgn(fRgn->fRgnH, h, v);
	}
};

void TRegion::Offset(Point p)	 // like Offset
{
	if (p.h !=0 && p.v != 0)	// don't bother if point is (0,0)
	{
		DetachSharedRgn();
		OffsetRgn(fRgn->fRgnH, p.h, p.v);
	}
}

void TRegion::Inset(short h, short v)	// like InsetRgn
{
	if (h != 0 && v != 0)		// don't bother if insets are zero
	{
		DetachSharedRgn();
		InsetRgn(fRgn->fRgnH, h, v);
	}
}

void TRegion::Inset(Point p)		// like InsetRgn
{
	if (p.h !=0 && p.v !=0)		// don't bother if point is (0,0)
	{
		DetachSharedRgn();
		InsetRgn(fRgn->fRgnH, p.h, p.v);
	}
}

Boolean TRegion::ContainsRect(const Rect &r) const
{
	return RectInRgn(&r, fRgn->fRgnH);
}

Boolean TRegion::ContainsPoint(Point p) const
{
	return PtInRgn(p, fRgn->fRgnH);
}

TRegion& TRegion::operator=(const TRegion &r) // assignment 
{
	r.fRgn->fCount++;				
	if (--fRgn->fCount == 0)	// delete current region only if 
	{							// no other references
		delete fRgn;
	}
	fRgn = r.fRgn;
	return *this;				// *this = r;
}

TRegion& TRegion::operator+=(const TRegion &r)  
{
	DetachSharedRgn();
	// *this = *this + r
	UnionRgn(fRgn->fRgnH, r.fRgn->fRgnH, fRgn->fRgnH);	
	return *this;
}

TRegion& TRegion::operator-=(const TRegion &r)	
{
	DetachSharedRgn();
	// *this = *this - r
	DiffRgn(fRgn->fRgnH, r.fRgn->fRgnH, fRgn->fRgnH);	
	return *this;
}

TRegion& TRegion::operator^=(const TRegion &r)
{
	DetachSharedRgn();
	// *this = *this ^ r
	XorRgn(fRgn->fRgnH, r.fRgn->fRgnH, fRgn->fRgnH); 	
	return *this;
}

TRegion& TRegion::operator|=(const TRegion &r)
{
	DetachSharedRgn();
	// *this = *this | r
	UnionRgn(fRgn->fRgnH, r.fRgn->fRgnH, fRgn->fRgnH);	
	return *this;
}

TRegion& TRegion::operator&=(const TRegion &r)
{
	DetachSharedRgn();
	// *this = *this & r
	SectRgn(fRgn->fRgnH, r.fRgn->fRgnH, fRgn->fRgnH);	
	return *this;
}

TRegion operator+(const TRegion &a, const TRegion &b) 	// union
{
	TRegion tempRgn;
	//  a + b;
	UnionRgn(a.fRgn->fRgnH, b.fRgn->fRgnH, tempRgn.fRgn->fRgnH);	
	return tempRgn;
}

TRegion operator-(const TRegion &a, const TRegion &b)		// difference
{
	TRegion tempRgn;
	// a - b;
	DiffRgn(a.fRgn->fRgnH, b.fRgn->fRgnH, tempRgn.fRgn->fRgnH);	
	return tempRgn;
}
							
TRegion operator&(const TRegion &a, const TRegion &b)		// intersection
{
	TRegion tempRgn;
	// a & b;
	SectRgn(a.fRgn->fRgnH, b.fRgn->fRgnH, tempRgn.fRgn->fRgnH);	
	return tempRgn;
}

TRegion operator|(const TRegion &a, const TRegion &b)		// union, same as operator+
{
	TRegion tempRgn;
	// a | b;
	UnionRgn(a.fRgn->fRgnH, b.fRgn->fRgnH, tempRgn.fRgn->fRgnH);
	return tempRgn;
}

TRegion operator^(const TRegion &a, const TRegion &b) 	// XOR
{
	TRegion tempRgn;
	// a ^ b;
	XorRgn(a.fRgn->fRgnH, b.fRgn->fRgnH, tempRgn.fRgn->fRgnH);	
	return tempRgn;
}

Boolean operator==(const TRegion &a, const TRegion &b)	// comparison
{
	if (a.fRgn == b.fRgn)  	// compare the RegionData pointers. 
	{		// This is quicker than an EqualRgn() call, but isn't
			// conclusive if the pointers aren't the same
		return true;		
	} 
	else
	{
		return EqualRgn(a.fRgn->fRgnH, b.fRgn->fRgnH);
	}
}

Boolean operator&&(const TRegion &a, const TRegion &b)
{
	return 						// a && b true if both are non-empty
	( !EmptyRgn(a.fRgn->fRgnH) && !EmptyRgn(b.fRgn->fRgnH) );
}

Boolean operator||(const TRegion &a, const TRegion &b)
{
	return 						// a && b true if both are non-empty
	 ( !EmptyRgn(a.fRgn->fRgnH) || !EmptyRgn(b.fRgn->fRgnH) );
}

void TRegion::DetachSharedRgn(void)
{
    if (fRgn->fCount > 1)           // if this TRegion uses a shared RegionData,
    {                               // create a copy, and disconnect from the
        RegionData *tmpRgn = fRgn;  // shared RegionData object, otherwise, 
        --fRgn->fCount;             // leave this TRegion's data alone
        fRgn = new RegionData;
        CopyRgn(tmpRgn->fRgnH, fRgn->fRgnH);
    }
}
