/********************************************************* DEFINITION
	DATE:	9/17/93
	AUTHOR: Eric R. Ros
	
	CLASS:  CPPZone411
	
	SUPERCLASS: CPPPeriodicTask
	
		This class lets you look up zones on the network using
		either phase 1 or phase 2 appletalk
	
********************************************************************/

#include <CPPZone411.h>
#include <CPPMaBell.h>
#include <CPPTaskManager.h>
#include <CPPStringList.h>
#include <StringTools.h>
#include <MemoryTools.h>
#include <ToolboxTools.h>
#include <ConvertTools.h>

/*-----------------------------------------------------------------*/
/*----------------------- INTERNAL CONSTANTS ----------------------*/
/*-----------------------------------------------------------------*/

#define		kPhaseTwo		53		/* version 53 or later has phase 2 Atalk */
#define		kSync			FALSE	
#define		kAsync			TRUE
#define 	kBridgeSocketID 6		/* socket used to ask bridge for zone list */
#define 	_GetZoneList	8

/*-----------------------------------------------------------------*/

	CPPZone411::CPPZone411 (CPPTaskManager *TaskManager,
						    CPPMaBell *MaBell,
						    long minPeriod, Boolean deleteWhenDone) :
						  	CPPPeriodicTask (TaskManager, minPeriod,
						  				   deleteWhenDone)
	{
		OSErr	ErrCode;
		
		FoundList = new CPPStringList ();
		
		this->MaBell = MaBell;
		this->AtalkVersionNumber = MaBell->GetAtalkVersion();
		if ((this->ZIPReferenceNumber = 
			 MaBell->GetDriverNumber (kSelectXPP)) == kDriverNotOpen)
	  	  ErrorAlert (kDriverNotOpen, "\pCan't do zone lookup");
		
		// make sure all our pointers are initialized
		this->myBDS = NULL;
		this->myATP = NULL;
		this->myXPB = NULL;
		this->BridgeAddress = NULL;
		this->returnBuffer = NULL;
		
		// find out whether we are connected to a network or not
		this->SetBridgeAddress();
	}
	
/*-----------------------------------------------------------------*/

	CPPZone411::~CPPZone411(void)
	{
		// there are no outstanding lookup calls; zonelookups are
		// done in sections synchronously
	
		// Clear the list of zones
		FoundList->DeleteItems(TRUE);
		delete this->FoundList;
		
		NukePtr(this->returnBuffer);
		NukePtr(this->BridgeAddress);
		NukePtr(this->myATP);
		NukePtr(this->myBDS);
		NukePtr(this->myXPB);
	}
	
/*-----------------------------------------------------------------*/

	char *CPPZone411::ClassName (void)
	{
		return "CPPZone411";
	}

/*-----------------------------------------------------------------*/

	Boolean	CPPZone411::NthZone (long whichItem, Boolean getCopy, StringPtr *ZoneName)
	/* return the specified zone from the list of found zones; */
	/* return FALSE if they ask for a zone which is greater than */
	/* the number of zones retrieved so far */
	/* if getCopy is TRUE, return a copy of the StringPtr */
	{
		StringPtr	TempString;
		
		*ZoneName = NULL;
		
		if (this->FoundList->GetNumItems() < whichItem)
		  return FALSE;
		else
		  {
		    *ZoneName = (*FoundList)[whichItem];
		    if (getCopy)
		      *ZoneName = (StringPtr)(Ptr2Ptr((Ptr)(*ZoneName)));
		    return TRUE;
		  }
	}

/*-----------------------------------------------------------------*/

	CPPStringList	*CPPZone411::GetFoundList (void)
	/* return the list where the zones are stored */
	{
		return this->FoundList;
	}

/*-----------------------------------------------------------------*/

	long CPPZone411::NumZonesFound (Boolean *isDone)
	/* return the number of zones found so far; return whether */
	/* the task has completed in the 'hasCompleted' variable */
	{
		*isDone = this->hasCompleted;
		return this->FoundList->GetNumItems();
	}

	
/*-----------------------------------------------------------------*/

	void CPPZone411::StartZoneLookup (CompletionProc DoProc)
	/* this routine sets all the initial values for the lookup task */
	/* and then sends it to the manager to be queued for execution */
	{
		// exit if this task is still involved in a lookup, or 
		// we have no manager
		if ((!this->hasCompleted) || (!this->ourManager))
		  return;
		
		// otherwise, set up the call
		this->SetCompletionProc(DoProc);
		this->hasCompleted = FALSE;				// note that we are not done
		this->FoundList->DeleteItems(TRUE);		// delete all known zones
		
		// now set up the param block calls
		if (this->AtalkVersionNumber >= kPhaseTwo)
		  StartPhase2Lookup ();
		else
		  StartPhase1Lookup ();
		
		// finally, add it to the periodic task manager queue
		if (this->callResult == noErr)
		  this->ourManager->AddPeriodicTask(this);	// now add the task
	}

/*-----------------------------------------------------------------*/

	void CPPZone411::DoPeriodicAction (void)
	/* if the lookup call has completed, extract the zone names and */
	/* execute another lookup call, unless we have gotten all the zones */
	/* in which case, set the hasCompleted flag */
	{
		Boolean	isDone;
		
		// call the inherited method to update frequency count
		CPPPeriodicTask::DoPeriodicAction();
		
		// do our periodic task
		if (this->AtalkVersionNumber >= kPhaseTwo)
		  Phase2PeriodicAction ();
		else
		  Phase1PeriodicAction ();
		
		// if the lookup is completed and no zones were returned,
		// append the default '*' to represent the zone we are in
		if (this->NumZonesFound(&isDone) == 0)
		  if (isDone)
		    this->FoundList->InsertInOrder (String2String("\p*"));		
	}

/*-----------------------------------------------------------------*/

	void CPPZone411::DoCompletedAction (void)
	{
		/* dispose of memory allocated during lookup */
	  	NukePtr(this->returnBuffer);
		if (this->AtalkVersionNumber >= kPhaseTwo)
		  NukePtr(this->myXPB);
		else
	  	  {
	  	    NukePtr(this->myATP);
	  	    NukePtr(this->myBDS);
	  	  }

		CPPPeriodicTask::DoCompletedAction();
	}

/*-----------------------------------------------------------------*/

	StringPtr CPPZone411::GetOurZoneName(OSErr *ErrCode)
	{
		if (this->AtalkVersionNumber < kPhaseTwo)
		  return GetZoneNamePhase1 (ErrCode);
		else
		  return GetZoneNamePhase2 (ErrCode);
	}
	
/*-----------------------------------------------------------------*/
/*------------------------ PRIVATE MEMEBERS -----------------------*/
/*-----------------------------------------------------------------*/

	void CPPZone411::SetBridgeAddress ()
	{
		OSErr	ErrCode;
		short	OurNodeID = 0, BridgeNodeID = 0;
		short	OurNetID = 0;
		
		NukePtr(this->BridgeAddress);
		this->BridgeAddress = (AddrBlockPtr)NewPtr(sizeof(AddrBlock));
		
		// the full address of the bridge will be the node # returned
		// by GetBridgeAddress, our own network number, and a predefined
		// socket number
		ErrCode = GetNodeAddress(&OurNodeID, &OurNetID);
		BridgeNodeID = GetBridgeAddress();
		
		if (OurNetID != 0 && BridgeNodeID != 0) 
		  {
			this->BridgeAddress->aNet = OurNetID;
			this->BridgeAddress->aNode = BridgeNodeID;
			this->BridgeAddress->aSocket = kBridgeSocketID;
			this->HaveNetwork = TRUE;
		  }
		else
		  this->HaveNetwork = FALSE;
	}

/*-----------------------------------------------------------------*/

	void CPPZone411::ProcessZoneList (short HowManyReturned)
	// this routine takes the zone names returned by a PBControl call and
	// moves them into the StringArray 
	{
		Ptr		Traverse = this->returnBuffer;
		short	i;
		
		// at this point, this->returnBuffer points at the data
		// and HowManyReturned has the # of entries in the buffer
		for (i=0; i< HowManyReturned; i++)
		  {
		  	this->FoundList->InsertInOrder((StringPtr)Data2PString(Traverse + 1, *Traverse));
		  	Traverse += (*Traverse) + 1;
		  }
		
	}

/*-----------------------------------------------------------------*/
/*------------------------- PHASE 2 LOOKUP ------------------------*/
/*-----------------------------------------------------------------*/

	void CPPZone411::StartPhase2Lookup ()
	{
		// allocate the buffer which the zone names will be returned into
		
		this->myXPB = (XCallPtr)NewPtr(sizeof(XCallParam));
		if ((this->callResult = MemError()) == noErr)
		  {
			this->returnBuffer = NewPtrClear(578);
			if ((this->callResult = MemError()) == noErr)
			  {
				// Set this so that DoPeriodic Action will not abort
				this->callResult = 0;
				
				// set up the xpb block for the first call
				myXPB->zipInfoField[0] = 0;	
				myXPB->zipInfoField[1] = 0;	 
				myXPB->zipLastFlag = 0;
		
				myXPB->ioRefNum = this->ZIPReferenceNumber;
				myXPB->csCode = xCall;
				myXPB->xppSubCode = zipGetZoneList;
				myXPB->xppTimeout = 3;
				myXPB->xppRetry = 4;
				myXPB->zipBuffPtr = (Ptr)this->returnBuffer;
				return;
			  }
			else
			  NukePtr(this->myXPB);
		  }

		// these lines will only execute if the memory allocation failed
	  	this->hasCompleted = TRUE;
	  	return;
	}

/*-----------------------------------------------------------------*/

	void CPPZone411::Phase2PeriodicAction ()
	{				
		// make the lookup call and, if it succeeds, add the zone names to the list
		if ((this->myXPB->zipLastFlag == 0) && (this->callResult == noErr))
		  {
		  	this->callResult = PBControl((ParmBlkPtr)this->myXPB, kSync);
		  	if (this->callResult == noErr)
		  	  this->ProcessZoneList (this->myXPB->zipNumZones);
		  }
	
		// if for some reason the call fails, or we reach the last zone
		// name, clean up after ourselves and note that we are done
		if ((this->myXPB->zipLastFlag != 0) || (this->callResult != noErr))
		  {
		    NukePtr(this->returnBuffer);
			this->hasCompleted = TRUE;
		  }
	}

/*-----------------------------------------------------------------*/
/*------------------------- PHASE 1 LOOKUP ------------------------*/
/*-----------------------------------------------------------------*/
	
	void CPPZone411::StartPhase1Lookup ()
	{
		// if we are not connected to a network, exit now.
		if (this->HaveNetwork)
		  {
			this->myATP = (ATPPBPtr)NewPtr(sizeof(ATPParamBlock));
			if ((this->callResult = MemError()) == noErr)
			  {
				this->returnBuffer = NewPtr(578); /* size of BDS */
				if ((this->callResult = MemError()) == noErr)
				  {
					this->myBDS = (BDSPtr)NewPtr(sizeof(BDSElement));
					if ((this->callResult = MemError()) == noErr)
					  {
						this->StartFromZoneNum = 1;
						
						
						// do some initial setup of the BDS and lookup block
						this->myATP->ATPioCompletion = 0L; 
						this->myBDS->buffSize = 578;
						this->myBDS->buffPtr = this->returnBuffer;
						this->myBDS->dataSize = 0;
						this->myBDS->userBytes = 0;
						
						return;	// let the periodic task do all the calling
					  }
				  }
			  }
		  }
		  
		// these lines will only execute if we don't have a network
		// or the memory allocation failed
		NukePtr(this->myATP);
		NukePtr(this->myBDS);
		NukePtr(this->returnBuffer);

	    this->hasCompleted = TRUE;
	    return;
	}

/*-----------------------------------------------------------------*/

	void CPPZone411::Phase1PeriodicAction ()
	{
		short	CallType = _GetZoneList;
		short	NumZonesReceived;
		long	UserData;
		Boolean	GotAllZones;
		
	  	switch (this->myATP->ATPioResult)  {
	  	
	  		case 1 :		// still busy getting names
	  			break;

	  		case noErr :
	  			// if the call has completed successfully, find out how
	  			// many zones were returned and add them to the list;
	  			// this will also tell us whether we have all the zones or not
				UserData = this->myBDS->userBytes;
				BlockMove((Ptr) &UserData, (Ptr) &GotAllZones, 1); 
				BlockMove((Ptr)&UserData + 2, (Ptr) &NumZonesReceived, 2);
	  			
	  			// process the zones
	  			if (NumZonesReceived > 0)
	  			  ProcessZoneList (NumZonesReceived);
	  			
	  			// if we've gotten all the zones, exit now;
	  			if (GotAllZones)
	  			  {
	  			  	this->hasCompleted = TRUE;
	  			  	return;
	  			  }
	  			
	  			// we haven't gotten all the zones, so queue up another call
			  	this->StartFromZoneNum += NumZonesReceived;
			  	
			  	// let the call know how much space we have for replies
				this->myBDS->buffSize = 578;
				this->myBDS->buffPtr = this->returnBuffer;
				this->myBDS->dataSize = 0;
				this->myBDS->userBytes = 0;
				
				// fill in the call parameters;
				this->myATP->ATPatpFlags = 0;
				this->myATP->ATPioCompletion = 0L; 
				this->myATP->ATPuserData = 0;
				this->myATP->ATPaddrBlock = *this->BridgeAddress;
				this->myATP->ATPreqLength = 0;
				this->myATP->ATPreqPointer = 0L;
				this->myATP->ATPbdsPointer = (Ptr)this->myBDS;
				this->myATP->ATPnumOfBuffs = 1;
				this->myATP->ATPtimeOutVal = 4;
				this->myATP->ATPretryCount = 4;
	
				// fidget with the userdata fields to get all the information
				// correctly aligned
				BlockMove((Ptr) &CallType + 1, (Ptr) &UserData, 1L);
				BlockMove((Ptr) &this->StartFromZoneNum, (Ptr)&UserData + 2, 2L);
				// fill in the userdata field of the ATP record
				this->myATP->ATPuserData = UserData;
	
				// make the call asynchronously
				this->callResult = PSendRequest(this->myATP, kAsync);
	  			break;
	  			
	  		default :		// if an error occurred, return it
				this->callResult = myATP->ATPioResult;
				hasCompleted = TRUE;
	  			break;
	  	}	// switch

	}
	
/*-----------------------------------------------------------------*/
/*----------------------- ZONE NAME LOOKUP ------------------------*/
/*-----------------------------------------------------------------*/
	
	StringPtr CPPZone411::GetZoneNamePhase1(OSErr *ErrCode)
	/* if we don't have phase 2, we have to use ATP to get */
	/* the name of our zone */
	{
		char			myZoneName[32];
		BDSElement 		myZoneBDS;
		ATPParamBlock 	LookupRec;
		
		*ErrCode = noErr;
		
		// if there is no network, return the default '*'
		if (!this->HaveNetwork)
		  return "\p*";
		
		// tell the call how big the return buffer is
		myZoneBDS.buffSize = 33;
		myZoneBDS.buffPtr = (Ptr)myZoneName;
		myZoneBDS.dataSize = 0;
		myZoneBDS.userBytes = 0;
	
		// set up the lookup call
		LookupRec.ATPatpFlags = 0;
		LookupRec.ATPioCompletion = 0L; 
		LookupRec.ATPuserData = 0;	/* ATP user data */
		LookupRec.ATPaddrBlock = *this->BridgeAddress;
		LookupRec.ATPreqLength = 0;
		LookupRec.ATPreqPointer = 0L;
		LookupRec.ATPbdsPointer = (Ptr)(&myZoneBDS);
		LookupRec.ATPnumOfBuffs = 1;
		LookupRec.ATPtimeOutVal = 3;
		LookupRec.ATPretryCount = 3;
		LookupRec.ATPuserData = 0x07000000;		/* GetLocalZone = 7 */
		
		// make the lookup call
		*ErrCode = PSendRequest(&LookupRec, false);
		if (*ErrCode != noErr)
		  return "\p*";
		else
		  return (StringPtr)Data2PString (&myZoneName[1], (short)myZoneName[0]);
	}

/*-----------------------------------------------------------------*/

	StringPtr CPPZone411::GetZoneNamePhase2(OSErr *ErrCode)
	{
		XPPParamBlock	MyXPPPB;
		char			myZoneName[32];

		*ErrCode = noErr;
		
		// if there is no network, return the default '*' zone name
		if (!this->HaveNetwork)
		  return "\p*";
								  
		/* set the fields of the XPP record */
		MyXPPPB.XCALL.xppTimeout = 3;
		MyXPPPB.XCALL.xppRetry = 4;
		MyXPPPB.XCALL.zipBuffPtr = myZoneName;
		MyXPPPB.XCALL.zipInfoField[0] = 0;
		MyXPPPB.XCALL.zipInfoField[1] = 0;
		
		*ErrCode = GetMyZone(&MyXPPPB, FALSE);
		if (*ErrCode != noErr)
		  return "\p*";
		else
		  return (StringPtr)Data2PString(&myZoneName[1], (short)myZoneName[0]);
	}


