/***************************************************** IMPLEMENTATION
	DATE:	10/4/93
	AUTHOR: Eric R. Ros

	CLASS:  CPPDArray
	
	SUPERCLASS: CPPObject
	
		This C++ class manages a dynamic array of arbitrarily
		sized objects
	
********************************************************************/

#include <CPPDArray.h>
#include <String.h>
#include <MathTools.h>
#include <ToolboxTools.h>

/*-----------------------------------------------------------------*/
/*---------------------------- MACROS -----------------------------*/
/*-----------------------------------------------------------------*/

	#define OFFSET(i) ((i-1) * sizeof(Ptr))
	#define	ITEMPTR(i) ((Ptr)((*listItems) + ((i-1) * this->itemSize)))

/*-----------------------------------------------------------------*/
/*------------------------ PUBLIC METHODS -------------------------*/
/*-----------------------------------------------------------------*/

	CPPDArray::CPPDArray (short objectSize)
	/* Class constructor */
	{
		this->numSlots = 0;
		this->numItems = 0;
		this->itemSize = objectSize;
		
		this->listItems = NewHandleClear(sizeof(Ptr));
	}

/*-----------------------------------------------------------------*/

	CPPDArray::~CPPDArray (void)
	/* Class destructor */
	{
		if (this->listItems)
		  DisposHandle(this->listItems);
	}

/*-----------------------------------------------------------------*/

	char *CPPDArray::ClassName (void)
	{
		return "CPPDArray";
	}
	
/*-----------------------------------------------------------------*/

	void	CPPDArray::AppendItem (Ptr NewItem)
	{
		CPPDArray::InsertItem (NewItem, this->numItems+1);
	}

/*-----------------------------------------------------------------*/

	void	CPPDArray::InsertItem (Ptr NewItem, long insertAt)
	/* insert an item in the list at the specified position.  if */
	/* insertAt is larger than the number of items, it is appended */
	/* to the end of the list */
	{
		if (this->numItems >= this->numSlots)
			Resize( this->numSlots + 5);
	
		// if the item goes inbetween two existing elements, move
		// the elements on the right down by 1 slot, otherwise, note
		// that it should go in the first unused slot in the list
		if (insertAt <= this->numItems) 
		  BlockMove( ITEMPTR(insertAt), ITEMPTR( insertAt + 1),
					(this->numItems - insertAt + 1) * this->itemSize);
		else
		  insertAt = this->numItems + 1;	
		
		// insert the item and increment the count of items in the list
		if (NewItem != NULL)
		  BlockMove (NewItem, ITEMPTR(insertAt), this->itemSize);
		
		this->numItems++;
		
		// let everyone talking to us know we have changed
		BroadcastMessage (kItemAdded, (void *)insertAt);
	}

/*-----------------------------------------------------------------*/

	void	CPPDArray::DeleteItem (long deleteWhich, Boolean doDispose)
	/* remove the data for the 'deleteWhich'th item from the list */
	{	
		Ptr TempPtr;
		
		if (IN_RANGE(deleteWhich, 1, this->numItems))
		  {
			// move the items to the right of the deleted element over
			// one to the left; this effectively deletes the item
			if (deleteWhich <= this->numItems)		
			  BlockMove( ITEMPTR( deleteWhich+1), ITEMPTR( deleteWhich),
								(this->numItems - deleteWhich + 1) * this->itemSize);
		
			// if we now have more free slots than we should, resize the
			// array
			if (this->numSlots > this->numItems + 5)
			  Resize( this->numSlots-5);
			
			// note that we have removed an item
			this->numItems--;
			
			// let everyone talking to us know we have changed
			BroadcastMessage (kItemDeleted, (void *)deleteWhich);
		  }
		  
	}

/*-----------------------------------------------------------------*/

	void	CPPDArray::DeleteItems (Boolean doDispose)
	/* remove all objects from the list; */
	/* NOTE: doDispose is irrelevant here, but used in later subclasses */
	{		
		for (long i = GetNumItems(); i > 0; i--)
		  CPPDArray::DeleteItem (i, doDispose);
	}

/*-----------------------------------------------------------------*/

	void CPPDArray::Get (void *dataPtr, long getWhich)
	/* assuming that dataPtr points to a buffer large enough to hold */
	/* a single element of the list, return the data in the item */
	/* specified by getWhich */
	{		
		if (IN_RANGE(getWhich, 1, this->numItems) && (dataPtr != NULL))
		  BlockMove (ITEMPTR(getWhich), dataPtr, this->itemSize);
	}

/*-----------------------------------------------------------------*/

	void *CPPDArray::operator[] (long getWhich)
	/* return a pointer to a copy of the data in position 'getWhich' */
	{
		Ptr	TempPtr = NULL;
		
		if (IN_RANGE(getWhich, 1, this->numItems))
		  {
		  	TempPtr = NewPtrClear (this->itemSize);
		  	if (TempPtr)
		  	  BlockMove (ITEMPTR(getWhich), TempPtr, this->itemSize);
		  }
		  
		return (void *)TempPtr;
	}
	
/*-----------------------------------------------------------------*/

	void	CPPDArray::SetItem (Ptr NewItem, long setWhich, Boolean deleteOld)
	/* Replace the data at position setWhich */
	{
		if (IN_RANGE(setWhich, 1, this->numItems) && (NewItem != NULL))
		  {
		    BlockMove (NewItem, ITEMPTR(setWhich), this->itemSize);
		    
			// let everyone talking to us know we have changed
			BroadcastMessage (kItemChanged, (void *)setWhich);
		  }
	}
	
/*-----------------------------------------------------------------*/

	void	CPPDArray::Swap (long Index1, long Index2)
	/* if both indices are valid, this routine swaps the data */
	/* at those indices */
	{
		Ptr Item1 = NewPtrClear(this->itemSize), 
			Item2 = NewPtrClear(this->itemSize);
		
		if (IN_RANGE(Index1, 0, this->numItems) &&
			IN_RANGE(Index2, 0, this->numItems))
		  {
		  	Get (Item1, Index1);
		  	Get (Item2, Index2);
		  	
		  	CPPDArray::SetItem (Item2, Index1, FALSE);
		  	CPPDArray::SetItem (Item1, Index2, FALSE);
		  	
		  	if (Item1) DisposPtr(Item1);
		  	if (Item2) DisposPtr(Item2);
		  }
	}
	
/*-----------------------------------------------------------------*/
	
	long	CPPDArray::GetNumItems (void)
	{
		return this->numItems;
	}

/*-----------------------------------------------------------------*/

	long	CPPDArray::FindIndex (Ptr ItemToFind)
	/* search for any occurrance of the data pointed to by ItemToFind */
	/* inside the list; return the index of the first one found */
	{
		for (long i = 1; i <= numItems; i++)
		  if (memcmp (ITEMPTR(i), ItemToFind, this->itemSize) == 0)
		    return i;
		return 0;
	}


/*-----------------------------------------------------------------*/
/*---------------------- PROTECTED METHODS ------------------------*/
/*-----------------------------------------------------------------*/

	void	CPPDArray::Clear (long clearWhich)
	/* set the contents of item 'clearWhich' to be blank */
	{
		Ptr		TempPtr;
		Ptr		Nada = NewPtrClear(this->itemSize);
		
		if (IN_RANGE (clearWhich, 1, this->numItems))
		  CPPDArray::SetItem (Nada, clearWhich, FALSE);
		if (Nada)
		  DisposPtr(Nada);
	}

/*-----------------------------------------------------------------*/
/*----------------------- PRIVATE METHODS -------------------------*/
/*-----------------------------------------------------------------*/

	void	CPPDArray::Resize (long numSlots)
	/* this routine resizes the 'items' handle so that it contains */
	/* the specified number of entries */
	{
		Handle	TempHandle = NULL;
		OSErr	ErrCode;
		
		// use the 'sethandlesize' call to resize the block of memory
		SetHandleSize(this->listItems, numSlots * this->itemSize);
		if (MemError() != noErr)
		  {	
		  	// if you can't resize, allocate a new block and copy the
		  	// contents of listItems into it.
		  	TempHandle = NewHandle(numSlots * this->itemSize);
		  	if ((ErrCode = MemError()) != noErr)
		  	  {
		  	  	ErrorAlert (ErrCode, "\pCould not resize the list");
		  	  	return;
		 	  }
		  	
		  	BlockMove (*this->listItems, *TempHandle, numSlots * this->itemSize);
		  	DisposeHandle(this->listItems);
		  	this->listItems = TempHandle;
		  }

		this->numSlots = numSlots;
	}

