
/* symbolize.c */
/* Copyright (c) 1995 Gustav Larsson */

/*
 * Three areas of this program have been optimized: memory
 * allocation, file I/O, and the symbol parse/convert
 * algorithms. Memory allocation and file I/O together
 * typically use up 70-80% of the total execution time.
 * Tuning these areas is tricky because so much is out
 * of our control. Factors such as the state of the disk
 * cache and the heap can affect overall execution time
 * by 10%, completely swamping many optimizations in the
 * parse/convert algorithms. Still, I have heavily
 * optimized the parse/convert algorithms because that
 * was the fun part of writing the program.
 *
 * Limitations:
 * - This program will not tolerate blank lines in the
 *   input file or symbol file (a final carriage-return
 *   is okay).
 * - Symbols names longer than about 2K may sometimes
 *   cause the output buffer to be overrun, trashing memory.
 *   This seems like a more-than-reasonable limit, even for
 *   mangled C++ names.
 */

#pragma options(pack_enums) /* required by <Memory.h> */

#include <stdio.h>
#include <Memory.h>

/* ====================================================== */

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

#define CEILING(value,n) (n*(((value)+n-1)/n))

/*
 * There is a substantial performance hit if a write
 * exceeds the disk cache. Large buffers also take longer
 * to allocate via NewPtr. Any buffer size between 8K and
 * 16K seems to produce comparable results. The actual
 * buffer size here is 16K, to allow the contents to run
 * over the 14K threshold.
 */

#define OUTPUT_BUFSIZE (16*1024)
#define OUTPUT_THRESHOLD (14*1024)

/* ====================================================== */

/*
 * ishex[] indicates which ASCII characters are valid
 * hex digits (0-9, A-F, a-f). There are entries for all
 * 256 character codes since anything could be in the
 * input file.
 */

static uchar ishex[] = {
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 00-0F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 10-1F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 20-2F */
  1,1,1,1,1,1,1,1, 1,1,0,0,0,0,0,0,   /* 30-3F */
  0,1,1,1,1,1,1,0, 0,0,0,0,0,0,0,0,   /* 40-4F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 50-5F */
  0,1,1,1,1,1,1,0, 0,0,0,0,0,0,0,0,   /* 60-6F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 70-7F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 80-8F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* 90-9F */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* A0-AF */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* B0-BF */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* C0-CF */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* D0-DF */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,   /* E0-EF */
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0    /* F0-FF */
};

/*
 * hexFromChar[] converts a single hex character to its
 * binary value. Since this table is only used the character
 * is known to be a hex digit, we can end the table at 'f'.
 */

static uchar hexFromChar[] = {
  0, 0, 0, 0, 0, 0, 0, 0,  0,0,0,0,0,0,0,0,   /* 00-0F */
  0, 0, 0, 0, 0, 0, 0, 0,  0,0,0,0,0,0,0,0,   /* 10-1F */
  0, 0, 0, 0, 0, 0, 0, 0,  0,0,0,0,0,0,0,0,   /* 20-2F */
  0, 1, 2, 3, 4, 5, 6, 7,  8,9,0,0,0,0,0,0,   /* 30-3F */
  0,10,11,12,13,14,15, 0,  0,0,0,0,0,0,0,0,   /* 40-4F */
  0, 0, 0, 0, 0, 0, 0, 0,  0,0,0,0,0,0,0,0,   /* 50-5F */
  0,10,11,12,13,14,15                         /* 60-66 */
};

/*
 * charsFromInt[] converts a binary value (0..99) to a pair
 * of decimal digits. The ULTOA macro converts two digits
 * at a time.
 */

static ushort charsFromInt[] = {
  '00','01','02','03','04','05','06','07','08','09',
  '10','11','12','13','14','15','16','17','18','19',
  '20','21','22','23','24','25','26','27','28','29',
  '30','31','32','33','34','35','36','37','38','39',
  '40','41','42','43','44','45','46','47','48','49',
  '50','51','52','53','54','55','56','57','58','59',
  '60','61','62','63','64','65','66','67','68','69',
  '70','71','72','73','74','75','76','77','78','79',
  '80','81','82','83','84','85','86','87','88','89',
  '90','91','92','93','94','95','96','97','98','99'
};

/* ====================================================== */

void Symbolize( FILE *inputFile, FILE *symbolFile,
                FILE *outputFile, unsigned short symLength);

static unsigned short parseSymbols( char *buffer,
                char **name, ulong *value );

static void convert( char *inputBuffer, char *outputBuffer,
             char **symbolName, ulong *symbolValue,
             ushort numSymbols, ushort symLength,
             size_t inputLength, FILE* outputFile );

/* ====================================================== */

void Symbolize( FILE *inputFile, FILE *symbolFile,
                FILE *outputFile, unsigned short symLength)
{
  char *memory, *inputBuffer, *symbolBuffer, *outputBuffer;
  char **symbolName;
  size_t inputMax, symbolFileMax, len;
  ushort maxSymbols, numSymbols;
  ulong *symbolValue;

  /*
   * Compute buffer sizes, making them as tight as possible.
   * Round up for alignment.
   *
   *    Up to 9 characters can be added to the
   *    input buffer ("\r00000000"). See convert().
   *
   *    Up to 2 characters can be added to the
   *    symbol file buffer ("\r\0").
   *
   *    The shortest possible line in the symbol file
   *    is 11 characters ("12345678 A\r"). Also, reserve
   *    room for one extra symbol since the last one
   *    gets duplicated; see parseSymbols().
   */

  inputMax = CEILING( inputFile->len + 9, 4 );
  symbolFileMax = CEILING( symbolFile->len + 2, 4);
  maxSymbols = symbolFile->len / 11 + 1;

  /*
   * Grab all the memory in one call to NewPtr, then
   * parcel it out. This one call usually takes about
   * 50% of the total execution time, at least on my
   * machine.
   */

  memory = NewPtr(
            inputMax + symbolFileMax + OUTPUT_BUFSIZE +
            (sizeof(char*) + sizeof(ulong*)) * maxSymbols );

  if ( !memory )
    return; /* at least the machine won't crash */

  inputBuffer = memory;
  symbolBuffer = inputBuffer + inputMax;
  outputBuffer = symbolBuffer + symbolFileMax;
  symbolName = (char **)( outputBuffer + OUTPUT_BUFSIZE );
  symbolValue = (ulong *)( symbolName + maxSymbols );

  /*
   * Put files into binary mode. This can reduce execution
   * time by 20-40% for this program. Setting files to
   * binary mode should really be done at fopen().
   */

  inputFile->binary =
  symbolFile->binary =
  outputFile->binary = 1;

  /*
   * Read the symbol file. Add a final '\r' if necessary.
   * Parse it, filling in symbolName[] and symbolValue[].
   */

  len = fread( symbolBuffer, 1, symbolFileMax, symbolFile );
  if ( symbolBuffer[ len-1 ] != '\r' )
    symbolBuffer[ len++ ] = '\r';
  symbolBuffer[ len ] = '\0';

  numSymbols = parseSymbols( symbolBuffer, symbolName,
                                              symbolValue );
                                              
  /*
   * Read the input file. Add a file '\r' if necessary.
   * Convert() will write the output file.
   */

  len = fread( inputBuffer, 1, inputMax, inputFile );
  if ( inputBuffer[ len-1 ] != '\r' )
    inputBuffer[ len++ ] = '\r';

  convert( inputBuffer, outputBuffer,
           symbolName, symbolValue,
           numSymbols, symLength,
           len, outputFile );

  /*
   * Deallocate memory.
   */

  DisposePtr( memory );
}

/* ====================================================== */

/*
 * I had a great deal of trouble getting Think C to store
 * pointer variables in registers. An address register
 * would sometimes remain allocated even after it went out
 * of scope (this didn't happen with data registers). In
 * desperation, I created a single char* register variable
 * that gets passed into various macros. When repeatedly
 * accessing an array, the first access below is slightly
 * faster:
 *
 *    int array[10];
 *    register int *parray = array;
 *    foo = parray[i];  <-- slightly faster
 *    foo = array[i];
 *
 * Here are some casts for the shared address register:
 */

#define TEMP_UC ((uchar*)temp)
#define TEMP_US ((ushort*)temp)
#define TEMP_UL ((ulong*)temp)

/*
 * XTOUL converts ASCII hex (X) to an unsigned long (UL).
 * Ptr gets bumped past the 8 hex digits.
 */

#define XTOUL(ptr,result,temp)  \
{                               \
  register ulong _value;        \
                                \
  temp = (char*) hexFromChar;   \
  _value = TEMP_UC[ *ptr++ ];   /* nybble 7 (high) */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 6 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 5 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 4 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 3 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 2 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 1 */ \
  _value <<= 4;                 \
  _value += TEMP_UC[ *ptr++ ];  /* nybble 0 (low) */ \
  result = _value;              \
}

/*
 * ULTOA converts an unsigned long (UL) to ASCII decimal (A)
 * two digits at a time. Buffer should be an array of 5
 * ushorts and will be filled with the right-justified
 * ASCII digits (2 chars per ushort). Length will be set
 * to the number of digits.
 *
 * This code is optimized for the common case of just a
 * few digits. / and % are inlined for 16-bit operands but
 * generate a function call for 32-bit operands (except on
 * a 68020 or better).
 */

#define TEN_TO_THE_4th 10000
#define TEN_TO_THE_5th 100000
#define TEN_TO_THE_6th 1000000
#define TEN_TO_THE_7th 10000000
#define TEN_TO_THE_8th 100000000
#define TEN_TO_THE_9th 1000000000

#define ULTOA(value,buffer,length,temp)           \
{                                                 \
  temp = (char*) charsFromInt;                    \
                                                  \
  if ( value < 100 )                              \
  {                                               \
    register ushort sval = value;                 \
    buffer[4] = TEMP_US[ sval ];                  \
    length = ( sval < 10 ? 1 : 2 );               \
  }                                               \
  else if ( value < TEN_TO_THE_4th )              \
  {                                               \
    register ushort sval = value;                 \
    buffer[4] = TEMP_US[ sval%100 ];              \
    buffer[3] = TEMP_US[ sval/100 ];              \
    length = ( sval < 1000 ? 3 : 4 );             \
  }                                               \
  else if ( value < TEN_TO_THE_6th )              \
  {                                               \
    register ushort sval = value % 10000;         \
    buffer[4] = TEMP_US[ sval%100 ];              \
    buffer[3] = TEMP_US[ sval/100 ];              \
    buffer[2] = TEMP_US[ value/10000 ];           \
    length = ( value < TEN_TO_THE_5th ? 5 : 6 );  \
  }                                               \
  else if ( value < TEN_TO_THE_8th )              \
  {                                               \
    register ushort sval = value % 10000;         \
    buffer[4] = TEMP_US[ sval%100 ];              \
    buffer[3] = TEMP_US[ sval/100 ];              \
    sval = value / 10000;                         \
    buffer[2] = TEMP_US[ sval%100 ];              \
    buffer[1] = TEMP_US[ sval/100 ];              \
    length = ( value < TEN_TO_THE_7th ? 7 : 8 );  \
  }                                               \
  else                                            \
  {                                               \
    register ushort sval = value % 10000;         \
    buffer[4] = TEMP_US[ sval%100 ];              \
    buffer[3] = TEMP_US[ sval/100 ];              \
    sval = (value / 10000) % 10000;               \
    buffer[2] = TEMP_US[ sval%100 ];              \
    buffer[1] = TEMP_US[ sval/100 ];              \
    buffer[0] = TEMP_US[ value/TEN_TO_THE_8th ];  \
    length = ( value < TEN_TO_THE_9th ? 9 : 10 ); \
  }                                               \
}

/*
 * LOOKUP performs a binary search of valueTable[]
 * and returns an index such that:
 *
 *   valueTable[index] <= value < valueTable[index+1]
 *
 * valueTable[numSyms-1] and valueTable[numSyms] should
 * both equal FFFFFFFF, so that index will be numSyms-1
 * in this case.
 */

#define LOOKUP(value,index,valueTable,numSyms,temp) \
{                                   \
  register ulong _lo, _hi;          \
                                    \
  temp = (char*) symbolValue;       \
  _lo = 0;                          \
  _hi = numSymbols;                 \
  while ( _lo+1 != _hi )            \
  {                                 \
    index = (_lo + _hi) >> 1;       \
    if ( value < TEMP_UL[index] )   \
      _hi = index;                  \
    else                            \
      _lo = index;                  \
  }                                 \
  index = _lo;                      \
}

/*
 * OUTPUT_SYMBOL writes "symbol+offset" into the output
 * buffer. outPtr is bumped past the string.  Remember
 * that a symbol name is terminated with a '\r', not '\0'.
 */

#define OUTPUT_SYMBOL(offset,symName,outPtr,temp) \
{                                               \
  register ulong _offset;                       \
                                                \
  /* Copy symbol name */                        \
  {                                             \
    temp = symName;                             \
    while ( ( *outPtr++ = *temp++ ) != '\r' ) ; \
  }                                             \
                                                \
/* We've copied the \r into the output buffer.  \
 * If offset is nonzero, replace \r with +      \
 * and and output the offset (decimal).         \
 * If offset is zero, back up one char.         \
 */                                             \
  _offset = offset;                             \
  if ( _offset )                                \
  {                                             \
    ushort _buffer[5], _bufLen;                 \
    ULTOA(_offset,_buffer,_bufLen,temp)         \
    {                                           \
      temp = ((char*) &_buffer[5]) - _bufLen;   \
      *(outPtr-1) = '+';                        \
      while ( _bufLen-- )                       \
        *outPtr++ = *temp++;                    \
    }                                           \
  }                                             \
  else outPtr--;  /* offset is zero */          \
}

/*
 * OUTPUT_ADDRESS write "[symbol+offset  ]" into the
 * output buffer. outPtr is bumped past the string.
 * There will be exactly symLen characters between "["
 * and "]". Truncate the symbol name or pad with spaces
 * as necessary.
 *
 * _count is the number of characters remaining.
 * _trunc is the number of characters remaining in
 * the symbol name (possibly truncated). It gets reused
 * in the second for loop to hold the number of digits
 * remaining in the decimal offset.
 */

#define OUTPUT_ADDRESS(offset,symName,symLen,outPtr,temp) \
{                                                   \
  register ulong _offset = offset;                  \
  if ( _offset )                                    \
  {                                                 \
    ushort _buffer[5], _bufLen;                     \
    ULTOA(_offset,_buffer,_bufLen,temp)             \
    {                                               \
      register int _count, _trunc;                  \
      *outPtr++ = '[';                              \
      for ( temp = symName, _count = symLen,        \
                _trunc = _count-_bufLen-1;          \
            _trunc && *temp != '\r';                \
            _count--, _trunc-- )                    \
        *outPtr++ = *temp++;                        \
                                                    \
      *outPtr++ = '+';                              \
      _count--;                                     \
      for ( temp = ((char*) &_buffer[5]) - _bufLen, \
                _trunc = _bufLen;                   \
            _trunc;                                 \
            _count--, _trunc-- )                    \
        *outPtr++ = *temp++;                        \
      while ( _count-- )                            \
        *outPtr++ = ' ';                            \
      *outPtr++ = ']';                              \
    }                                               \
  }                                                 \
  else                                              \
  {                                                 \
    register int _count;                            \
    *outPtr++ = '[';                                \
    for ( temp = symName, _count = symLen;          \
          _count && *temp != '\r';                  \
          _count-- )                                \
      *outPtr++ = *temp++;                          \
    while ( _count-- )                              \
      *outPtr++ = ' ';                              \
    *outPtr++ = ']';                                \
  }                                                 \
}

/* ====================================================== */

/*
 * parseSymbols() parses the symbol file. Each element of
 * name[] gets a pointer to a symbol name. Each element of
 * value[] gets the value of a symbol. Note that each
 * symbol name is terminated by '\r', not '\0'.
 */

static unsigned short parseSymbols( char *buffer,
                                    char **name,
                                    ulong *value )
{
  register char *ptr;
  register char *temp;
  char **nextName;
  ulong *nextValue;
  ushort numSymbols;

  nextName = name;    /* where to save info for */
  nextValue = value;  /* the next symbol */
  numSymbols = 0;

  ptr = buffer;
  while ( *ptr )
  {
    /* The first 8 characters must be a hex number */
    XTOUL(ptr,*nextValue,temp)

    /* Skip whitespace between value and symbol */
    while ( *ptr++ == ' ' ) ;

    /* Save pointer to start of symbol */
    *nextName = ptr-1;

    /* Find start of next line */
    while ( *ptr++ != '\r' ) ;

    nextName++;
    nextValue++;
    numSymbols++;
  }

  /*
   * Duplicate last symbol so the LOOKUP macro
   * finds FFFFFFFF correctly.
   */
  *nextValue = *(nextValue-1);

  return numSymbols;
}

/* ====================================================== */

/*
 * convert() converts 8 digit hex values in the input file.
 * It is also responsible for writing the output bufffer to
 * the output file.
 */

static void convert(
            char *inputBuffer, char *outputBuffer,
            char **symbolName, ulong *symbolValue,
            ushort numSymbols, ushort symLength,
            size_t inputLength, FILE *outputFile )
{
  register char *inPtr, *outPtr;
  register char *temp;
  ulong addrValue, nextValue, *nextValuePointer;
  char *addrName, **nextName, *endPtr, *writeThreshold;

  /*
   * inPtr and outPtr are both register variables. The
   * third address register variable is temp.
   */

  inPtr = inputBuffer;
  outPtr = outputBuffer;

  /*
   * Stop when inPtr equals endPtr. Flush the output buffer
   * to disk when outPtr exceeds writeThreshold.
   */

  endPtr = inPtr + inputLength;
  writeThreshold = outputBuffer + OUTPUT_THRESHOLD;

  /*
   * Add eight ASCII '0's to end of input file. The search
   * algorithm expects eigth hex digits after each \r, even
   * at the end of the file.
   */

  {
    temp = endPtr;
    *temp++ = '0'; *temp++ = '0';
    *temp++ = '0'; *temp++ = '0';
    *temp++ = '0'; *temp++ = '0';
    *temp++ = '0'; *temp++ = '0';
  }

  /* Force lookup first time through. */
  addrValue = 0xFFFFFFFF;
  nextValue = 0;    /* in case first address is FFFFFFFF */
  nextValuePointer = symbolValue-1;

  /* Loop once per input line */
  while ( inPtr != endPtr )
  {

    /*
     * Assume that the first eight characters of each
     * line are all hex digits. Convert this value to
     * [symbol+offset] form. A line will usually have the
     * same symbol as the previous line, or sometimes
     * the next higher symbol. Thus, we check for these
     * two cases first and do a binary search only as
     * a last resort.
     */

    {
      register ulong address;
      XTOUL(inPtr,address,temp)
      if ( address < addrValue )
        goto lookup;  /* address going backward (unusual) */
      else if ( address >= nextValue )
      {
        /* Address doesn't match current symbol. */
        /* Check if it matches the next symbol.  */
        addrValue = nextValue;
        nextValue = *(++nextValuePointer);
        if ( address < nextValue )
          addrName = *nextName++; /* yes, it matches next */
        else
        {
          /* Doesn't match current or next symbol, */
          /* so do a full binary search.           */
          register ulong index;
        lookup:
          LOOKUP(address,index,symbolValue,numSymbols,temp)
          addrName = symbolName[index];
          nextName = symbolName + index + 1;
          nextValuePointer = symbolValue + index + 1;
          nextValue = *nextValuePointer;
          addrValue = *(nextValuePointer - 1);
        }
      }
      OUTPUT_ADDRESS(address-addrValue,addrName,
                     symLength,outPtr,temp)
    }

    /*
     * Scan the rest of the line for 8 digit hex numbers.
     * It is often possible to jump ahead many characters
     * when we find a non-hex digit. We check seven
     * characters ahead for a hex digit, then six characters
     * ahead, etc. If we find a non-hex digit we know the
     * intervening characters can't possibly be an 8 digit
     * number.
     *
     * It turns out that if there is a \r anywhere in the
     * next eight characters, it will be the first non-hex
     * character we encounter. This happens because the
     * first eight characters of a line are hex digits, and
     * we are looking ahead at most eight characters. Thus,
     * if we start beyond a \r, we will see hex digits all
     * the way back to the \r. There can't be a second \r
     * hiding earlier in the line since each line has at
     * least eight characters; another whole line wouldn't
     * fit into the remaining characters. When we jump to
     * copy8..copy1, we only need to check for \r at copy1.
     */

    /* Loop until end of line */
    while ( *inPtr != '\r' )
    {
      {
        temp = inPtr+7;
        if ( !ishex[ *temp-- ] ) goto copy8;
        if ( !ishex[ *temp-- ] ) goto copy7;
        if ( !ishex[ *temp-- ] ) goto copy6;
        if ( !ishex[ *temp-- ] ) goto copy5;
        if ( !ishex[ *temp-- ] ) goto copy4;
        if ( !ishex[ *temp-- ] ) goto copy3;
        if ( !ishex[ *temp-- ] ) goto copy2;
        if ( !ishex[ *temp ] ) goto copy1;
      }

      {
        /* Found 8 hex digits. Convert to binary */
        /* and ouput in "symbol+offset" form.    */

        register ulong value, index;
        XTOUL(inPtr,value,temp)
        LOOKUP(value,index,symbolValue,numSymbols,temp)
        OUTPUT_SYMBOL(value-symbolValue[index],
                      symbolName[index],outPtr,temp)
        continue;
      }

      /* Didn't have 8 hex digits.        */
      /* Copy to output and check for \r. */

      copy8:  *outPtr++ = *inPtr++;
      copy7:  *outPtr++ = *inPtr++;
      copy6:  *outPtr++ = *inPtr++;
      copy5:  *outPtr++ = *inPtr++;
      copy4:  *outPtr++ = *inPtr++;
      copy3:  *outPtr++ = *inPtr++;
      copy2:  *outPtr++ = *inPtr++;
      copy1:  if ( *inPtr != '\r' )
                *outPtr++ = *inPtr++;
              else break; /* exit the inner while loop */
    }

    /* Now we're at the end of the line. Copy the \r, */
    /* then flush the output buffer if we've run past */
    /* the threshold.                                 */

    *outPtr++ = '\r';
    inPtr++;

    if ( outPtr >= writeThreshold )
    {
      fwrite(outputBuffer,1,outPtr-outputBuffer,outputFile);
      outPtr = outputBuffer;
    }
  }

  /* We have reached the end of the input file. */
  /* Flush the rest of the output buffer.       */

  if ( outPtr != outputBuffer )
    fwrite(outputBuffer,1,outPtr-outputBuffer,outputFile);
}

