/***************************************************** IMPLEMENTATION
	DATE:	10/24/93
	AUTHOR: Eric R. Ros

	CLASS:  CPPTreeNode
	
	SUPERCLASS: CPPObjectList
	
		This C++ class manages a node in a hierarchical tree
	
********************************************************************/


#include <CPPTreeNode.h>
#include <MathTools.h>
#include <CPPTree.h>

extern	Rect	kEmptyRect;

/*-----------------------------------------------------------------*/
/*------------------------ APPLY ROUTINES -------------------------*/
/*-----------------------------------------------------------------*/

	void	SetRoot (CPPTreeNode *theNode, long param)
	/* set the value of 'Root' to the passed-in value */
	{
		if (theNode)
		  theNode->Root = (CPPTree *)param;
	}
	
/*-----------------------------------------------------------------*/

	void	CountChildren (CPPTreeNode *theNode, long param)
	/* add 1 to the value in param every time we are called */
	{
		(*((long *)param))++;
	}

/*-----------------------------------------------------------------*/
/*------------------------ PUBLIC METHODS -------------------------*/
/*-----------------------------------------------------------------*/

	CPPTreeNode::CPPTreeNode (CPPObject *NodeData, CPPTree *BelongsTo, Boolean becomeOwner)
	{
		this->nodeData = NodeData;
		this->ownsData = becomeOwner;
		this->Root = BelongsTo;
		this->Parent = NULL;
	}

/*-----------------------------------------------------------------*/

	CPPTreeNode::~CPPTreeNode (void)
	{
		if (this->ownsData && this->nodeData)
		  delete this->nodeData;
		  
		if (this->Parent)
		  {	// remove item from parent's list
		  	Parent->ReceiveMessage(this, kPrepareDelete, (void*)this);
		  	this->Parent->DeleteItem (this, FALSE);
		  	// tell the parent it was deleted
		    Parent->ReceiveMessage(this, kChildDeleted, (void *)this);
		  }
	}
			
/*-----------------------------------------------------------------*/

	Boolean	CPPTreeNode::Member (char *className)
	{
		if (strcmp(className, CPPTreeNode::ClassName()) == 0)
		  return TRUE;
		else
		  return CPPObjectList::Member(className);
	}

/*-----------------------------------------------------------------*/

	char 	*CPPTreeNode::ClassName (void)
	{
		return "CPPTreeNode";
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::TellParent (short Message, void *data)
	/* Pass a message along to your parent; if you have no */
	/* parent, pass the message to yourself */
	{
		if (this->Parent)
		  this->Parent->ReceiveMessage (this, Message, data);
		else
		  this->ReceiveMessage (this, Message, data);
	}

/*-----------------------------------------------------------------*/

	CPPObject *CPPTreeNode::Clone(void)
	{
		return NULL;
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::AddNode (CPPTreeNode *NewNode)
	/* add a child to this node */
	{
		InsertNode (NewNode, this->numItems+1);
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::InsertNode (CPPTreeNode *NewNode, long insertWhere)
	/* add a child to this node */
	{
		if (NewNode)
		  {
			NewNode->Parent = this;
			ApplyToFamily (NewNode, SetRoot, (long)this->Root);
			InsertItem (NewNode, insertWhere);
			
			this->ReceiveMessage (this, kNodeAdded, (void *)NewNode);
		  }
	}

/*-----------------------------------------------------------------*/

	CPPTreeNode	*CPPTreeNode::NthChild (long whichChild)
	{
		return (CPPTreeNode *)CPPObjectList::operator[](whichChild);
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::RemoveChild (CPPTreeNode *theChild)
	/* disconnect a child from its parent without affecting */
	/* its subtree at all */
	{
		long	itsIndex = this->FindIndex(theChild);
		
		if (itsIndex)
		  {
		  	ReceiveMessage(this, kPrepareRemove, (void*)theChild);
		  	DeleteItem (theChild, FALSE);
		  	ReceiveMessage (this, kChildRemoved, (void *)theChild);
		  }
	}

/*-----------------------------------------------------------------*/

	Boolean CPPTreeNode::InsertParent (CPPTreeNode *NewNode)
	/* insert a new node before the current node */
	{	
		CPPTreeNode	*ParentNode = this->Parent;
		long	insertionPoint;
		
		if (NewNode && ParentNode)
		  {
	  	  	// remove us from our old position
	  	  	insertionPoint = ParentNode->FindIndex(this);
	  	  	ParentNode->RemoveChild (this);
	  	  	
	  	  	// add the new parent in our old position
	  	  	NewNode->Parent = ParentNode;
	  	  	ParentNode->InsertItem (NewNode, insertionPoint);
	  	  	
	  	  	// Make the new parent our parent
	  	  	NewNode->AddNode (this);
	  	  	return TRUE;
		  }
		else
		  {
		  	this->Root->SetTopMostNode (NewNode);
		  	AddNode (this);
		  	return TRUE;
		  }
	}

/*-----------------------------------------------------------------*/

	CPPObject	*CPPTreeNode::GetNodeData (void)
	{
		return this->nodeData;
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::SetNodeData (CPPObject *NewData, Boolean becomeOwner)
	{
		if (this->ownsData && this->nodeData)
		  delete this->nodeData;
		this->nodeData = NewData;
		this->ownsData = becomeOwner;
		ReceiveMessage (this, kNodeChangedData, this);
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::SwitchParents (CPPTreeNode *OldParent, CPPTreeNode *NewParent)
	/* move a node from one parent to another */
	{
		if (OldParent == NewParent) return;
		
		OldParent->RemoveChild (this);
		
		NewParent->AddNode (this);
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::ReceiveMessage (CPPGossipMonger *toldBy, 
										 short reason, void* info)
	/* respond to the special messages which the tree nodes send */
	{
		switch (reason) {
		
			case kChildDeleted	:

				break;
				
			case kNodeAdded	:
				
				break;
			
			case kChildRemoved	:	
				
				break;
		}	
	}

/*-----------------------------------------------------------------*/

	void	CPPTreeNode::GetFamilyBounds (Rect *bounds)
	{
		*bounds = kEmptyRect;
	}

/*-----------------------------------------------------------------*/

	short	CPPTreeNode::FamilyDepth (short currentDepth)
	/* return the maximum depth of the current subtree */
	{
		CPPTreeNode	*theChild;
		short	MaxDepth = currentDepth+1;
		
		if (this->numItems)
		  {
			for (long i = 1; i <= this->numItems; i++)
			  {
			  	theChild = NthChild(i);
			  	if (theChild)
			  	  MaxDepth = Max(theChild->FamilyDepth(currentDepth + 1), MaxDepth);
			  }
		  }
		  
		return MaxDepth;
	}

/*-----------------------------------------------------------------*/

	CPPObjectList	*CPPTreeNode::BuildSelectionList (void)
	/* Return a list of the selected nodes in the family; */
	/* if connected nodes are selected, they are returned as a */
	/* whole family rather than as individual nodes */
	{
		return NULL;
	}
	
/*-----------------------------------------------------------------*/
/*--------------------- PUBLIC CLASS METHODS ----------------------*/
/*-----------------------------------------------------------------*/

	void ApplyToFamily (CPPTreeNode *familyHead, 
						ApplyProc theProc, long param)
	/* apply the specified routine to each element in a family */
	{
		CPPTreeNode	*TempNode;
		
		if (familyHead && theProc)
		  {
		  	for (long i = 1; i <= familyHead->GetNumItems(); i++)
		  	  if ((TempNode = familyHead->NthChild (i)) != NULL)
		  	    ApplyToFamily (TempNode, theProc, param);
		  		
		  	(*theProc)(familyHead, param);
		  }
	}

/*-----------------------------------------------------------------*/

	Boolean BApplyToFamily (CPPTreeNode *familyHead, 
							BApplyProc theProc, long param)
	/* apply the specified routine to each element in a family */
	{
		CPPTreeNode	*TempNode;
		Boolean	BTemp = FALSE;
		
		if (familyHead && theProc)
		  {
		  	for (long i = 1; i <= familyHead->GetNumItems(); i++)
		  	  if ((TempNode = familyHead->NthChild (i)) != NULL)
		  	    BTemp = BApplyToFamily (TempNode, theProc, param) || BTemp;
		  
		  	BTemp = (*theProc)(TempNode, param) || BTemp;
		  }
		  
		return BTemp;
	}

/*-----------------------------------------------------------------*/

	long	CPPTreeNode::FamilySize (CPPTreeNode *theNode)
	/* return the number of nodes in the family whose head is theHead */
	/* NOTE:  The family head is included in the count */
	{
		long	count = 0;
		
		if (theNode)
		  ApplyToFamily (theNode, CountChildren, (long)(&count));
		
		return count;
	}

/*-----------------------------------------------------------------*/

	void CPPTreeNode::DeleteFamily (CPPTreeNode *theNode)
	/* delete the entire family whose head is theNode */
	{
		CPPTreeNode	*ParentNode;
		
		if (theNode && theNode->Parent)
		  {
		  	// accomplish the deletion by removing the entire
		  	// family from the tree, then deleting them all
		  	// secretly
		  	ParentNode = theNode->Parent;
		  	ParentNode->RemoveChild (theNode);
		  	theNode->ClearFamily (theNode);
		  }
	}

/*-----------------------------------------------------------------*/

	void CPPTreeNode::ClearFamily (CPPTreeNode *theNode)
	/* recursively delete all of the nodes belonging to a */
	/* family */
	{
		CPPTreeNode *tempNode, *ParentNode;
		long		numItems = theNode->GetNumItems();
		
		if (theNode)
		  {
		    for (long i = 1; i <= numItems; i++)
		      {
		 		tempNode = theNode->NthChild(1);
				if (tempNode)
				  {	// remove the item from the list w/o deleting it
				  	theNode->DeleteItem(tempNode, FALSE);
				  	// have it clear its children
				  	ClearFamily(tempNode);
				  }
			  }
		   
		   // delete the node once its children are all cleared
		   theNode->Parent = NULL;
		   delete theNode;
		  }
	}

/*-----------------------------------------------------------------*/

	Boolean CPPTreeNode::DeleteNode (CPPTreeNode *theNode)
	/* If the node has a parent, copy its children into its parents */
	/* list, delete the node and return TRUE.  Otherwise return FALSE */
	{
		CPPTreeNode	*ParentNode, *ChildNode;
		long		insertionPoint, itsIndex;
		long		numChildren = theNode->GetNumItems();
		
		if (theNode && theNode->Parent)
		  {
		  	ParentNode = theNode->Parent;
		  	// erase the deleted node and its children
		  	ParentNode->ReceiveMessage(theNode, kPrepareDelete, (void*)theNode);
		  	itsIndex = insertionPoint = ParentNode->FindIndex(theNode);
		  	
		  	// remove the node from it's parents list of children
			ParentNode->DeleteItem (theNode, FALSE);

		  	// have the parent adopt its children
		  	for (long i = 1; i <= numChildren; i++)
		  	  {
		  	  	ChildNode = theNode->NthChild(i);
		  	  	ChildNode->Parent = ParentNode;
				ParentNode->InsertItem (ChildNode, insertionPoint++);		  	    
		  	  }
			
			// redraw the tree without the deleted nodes 	
			ParentNode->ReceiveMessage (theNode, kChildDeleted, (void *)theNode);

			// finally delete the node
			theNode->Parent = NULL;
			delete theNode;
						
			return TRUE;
		  }
		else
		  return FALSE;
	}

