#include <OSUtils.h>
#include <Memory.h>
#include <Files.h>

/* OVERVIEW  market.c -- by Xan Gregg
I cache all symbol names in memory and as many blocks
of trade data from disk as possible are cached in the rest
of the allowed memory.  For each symbol I keep a "pointer"
(block# and entry#) to the most recent trade involving that
symbol.  Each trade points to the previous trade of that
symbol.

The header includes an array of 26K entries in which each
entry points to a list of symbols.  All symbols with the
same first three letters are on the same list.  So the array
of symbol lists is indexed by the first three letters.

The trade data are stored on disk as a summary of the trade
information passed to NewTrade().  The time is stored in
a long, and the volume is stored as a daily cummulative
value.  The data is accumulated into a 8K block which
is written to disk when it gets full.

I know that there will be under 500,000 trades (since
the trade is at most 10MB and each Trade struct is 22
bytes).  And I have been told there will be about 10,000
symbols and at most 16K.  So that puts each symbol at
an average of 50 trades each.  That's good for my approach
of indexing by the symbol name.  However, since I iterate
through the list of trades in reverse order, I am at a
disadvantage if a lot of older data is requested.

*/

typedef unsigned char uchar;
typedef unsigned short	ushort;
typedef unsigned long ulong;

typedef struct TimeStamp {
	uchar			yearsFrom1900;
	uchar			month;
	uchar			day;
	uchar			hour;
	uchar			minute;
	uchar			second;
} TimeStamp;

typedef char Str7[8];

typedef struct Trade {
	Str7			symbol;
	TimeStamp		time;
	Fixed			price;
	ulong			numShares;
} Trade;

void *InitTradeDatabase(ulong maxRAM);
void NewTrade(void *privateDataPtr, Trade trade);
Fixed PriceIs(void *privateDataPtr, Str7 symbol, TimeStamp time);
ulong VolumeIs(void *privateDataPtr, Str7 symbol, TimeStamp time);

#define kBlockSize  8192L
#define kBlockEntries   (kBlockSize/16)
#define kFileSize   (8*1024L*1024L)
                /* enough to store 500,000 trades */
#define kNumBlocks  1024L

#define kNumSymbolIndices (26*32*32L)
#define kSecondsPerDay  (60L*60L*24)

typedef struct SymbolEntry SymbolEntry;

struct SymbolEntry
 {
    long        restOfSymbol;   /* last 4 chars */
    SymbolEntry *next;          /* with same 1st 3 chars */
    short       latestBlockNum; /* location on disk */
    short       latestEntryNum; /* location with block */
 };

/* 16-byte struct stored on disk for each trade */
typedef struct
 {
    ulong   seconds;    /* seconds since 1904 */
    Fixed   price;      /* same as from Trade record */
    ulong   totalShares; /* current and previous */
    short   prevBlockNum;
    short   prevEntryNum;
 } TradeSummary;

typedef struct
 {
    TradeSummary    data[kBlockEntries];
    short   blockNum;   /* where it is on disk */
    short   filler;
 } CachedBlock;

typedef struct
 {
    CachedBlock *blockLocation[kNumBlocks]; /* 4KB */
    Ptr         symbolDataEnd;
    CachedBlock *lowestBlockP;
    CachedBlock *highestBlockP;
    CachedBlock *oldestBlockP;
    CachedBlock *currentOutBlockP;
    short       currentEntryNum;
    short       refNum;
    ParamBlockRec   pb;
    ushort      symbolIndex[kNumSymbolIndices]; /* 26KB */
    SymbolEntry symbols[1]; /* variable length */
 } Header;

/* the above struct takes less than 32K, and the
symbol data takes less than 192K (assuming a max of
16K symbols), so the cached blocks can use everything
after the first 224K */

#define gBlockLocation      dataP->blockLocation
#define gSymbolDataEnd      dataP->symbolDataEnd
#define gLowestBlockP       dataP->lowestBlockP
#define gHighestBlockP      dataP->highestBlockP
#define gOldestBlockP       dataP->oldestBlockP
#define gCurrentOutBlockP   dataP->currentOutBlockP
#define gCurrentEntryNum    dataP->currentEntryNum
#define gRefNum             dataP->refNum
#define gPB                 dataP->pb
#define gSymbolIndex        dataP->symbolIndex
#define gSymbols            dataP->symbols


void *InitTradeDatabase(ulong maxRAM)
{
    Header      *dataP;
    CachedBlock *blockP;

    maxRAM = maxRAM & -4L;  /* make sure end is aligned */
    dataP = (Header *) NewPtrClear(maxRAM);
    if (dataP == 0)
        DebugStr("\p NO MEM");

    /* skip the first symbol as 0 is not a valid index */
    gSymbolDataEnd = (Ptr) (gSymbols + 1);

    /* Initialize memory blocks at end of data area. */
    blockP = (CachedBlock *) (((Ptr) dataP) + maxRAM);
    gCurrentOutBlockP = blockP - 1;
    /* leaving 224K for headers and symbols */
    while ((Ptr) (blockP - 1) > ((Ptr) dataP) + 224*1024L)
     {
        blockP--;
        blockP->blockNum = -1;
     }
    gLowestBlockP = blockP;
    gHighestBlockP = gCurrentOutBlockP;
    gOldestBlockP = gHighestBlockP - 1;
    gCurrentOutBlockP->blockNum = 0;
    gBlockLocation[0] = gCurrentOutBlockP;

    { /* create swap file */
    OSErr   err;
    long    size;
    err = Create("\pXansSwapFile", 0, 'xxxx', 'DATA');
    if (err)
     {
        FSDelete("\pXansSwapFile", 0);
        err = Create("\pXansSwapFile", 0, 'xxxx', 'DATA');
        if (err)
            DebugStr("\p NO FILE");
     }
    FSOpen("\pXansSwapFile", 0, &gRefNum);
    if (gRefNum == 0)
        DebugStr("\p NO FILE");
    size = kFileSize;
    err = AllocContig(gRefNum, &size);
    size = kFileSize;
    if (err)
        err = Allocate(gRefNum, &size);
    if (err)
        DebugStr("\p NO FILE");
    SetEOF(gRefNum, kFileSize);
    }
    return (void *) dataP;
}


/* Take a chance and use OS trap here.  If assembly
were allowed could at least cache the address of the
code and call it with a function pointer instead of
through a trap, or we could just write this whole
routine is assembly.
*/
ulong timeToSeconds(TimeStamp *timeP)
{
    DateTimeRec dateTime;
    ulong       seconds;

    dateTime.year = 1900 + timeP->yearsFrom1900;
    dateTime.month = timeP->month;
    dateTime.day = timeP->day;
    dateTime.hour = timeP->hour;
    dateTime.minute = timeP->minute;
    dateTime.second = timeP->second;
    Date2Secs(&dateTime, &seconds);
    return seconds;
}


/* returns 0 if the symbol was not found */
SymbolEntry *findSymbolEntry(Header *dataP, Str7 symbol)
{
    register long           rest;
    register SymbolEntry    *symbolP;
    long        memRest;
    short       keyIndex;
    char        secondChar;
    char        thirdChar;
    char        *p;
    short       len = symbol[0];

    /* figure the keyIndex (first three letters) */
    if (len < 2)
     {
        secondChar = 0;
        thirdChar = 0;
     }
    else
     {
        secondChar = symbol[2] - '@';
        if (len != 2)
            thirdChar = symbol[3] - '@';
        else
            thirdChar = 0;
     }
    keyIndex = ((symbol[1] - 'A') << 10)
                        | (secondChar << 5) | thirdChar;

    /* and the 'rest' (last four letters) */
    memRest = 0;
    p = (char *) &memRest;
    while (len > 3)
        *p++ = symbol[len--];

    rest = memRest;
    if (gSymbolIndex[keyIndex] == 0)
        symbolP = 0;
    else
     {
        symbolP = &gSymbols[gSymbolIndex[keyIndex]];
        while ((symbolP != 0)
                 && (symbolP->restOfSymbol != rest))
            symbolP = symbolP->next;
     }
    return symbolP;
}



/* like findSymbol, but inserts it if necessary */
SymbolEntry *findInsertSymbolEntry(Header *dataP,
                                        Str7 symbol)
{
    register long           rest;
    register SymbolEntry    *symbolP;
    long        memRest;
    short       keyIndex;
    char        secondChar;
    char        thirdChar;
    char        *p;
    short       len = symbol[0];

    /* figure the keyIndex (first three letters) */
    if (len < 2)
     {
        secondChar = 0;
        thirdChar = 0;
     }
    else
     {
        secondChar = symbol[2] - '@';
        if (len != 2)
            thirdChar = symbol[3] - '@';
        else
            thirdChar = 0;
     }
    keyIndex = ((symbol[1] - 'A') << 10)
                     | (secondChar << 5) | thirdChar;

    /* and the 'rest' (last four letters) */
    memRest = 0;
    p = (char *) &memRest;
    while (len > 3)
        *p++ = symbol[len--];

    rest = memRest;
    if (gSymbolIndex[keyIndex] == 0)
     {  /* this is the first symbol with this key */
        symbolP = (SymbolEntry *) gSymbolDataEnd;
        gSymbolIndex[keyIndex] =
                            (ushort) (symbolP - gSymbols);
        gSymbolDataEnd += sizeof(SymbolEntry);
        symbolP->restOfSymbol = rest;
        symbolP->next = 0;
        symbolP->latestBlockNum = -1;
     }
    else
     {  /* search the list for our last four letters */
        symbolP = &gSymbols[gSymbolIndex[keyIndex]];
        while ((symbolP->next != 0) &&
                    (symbolP->restOfSymbol != rest))
            symbolP = symbolP->next;
        if (symbolP->restOfSymbol != rest)
         {  /* not found so add it */
            symbolP->next = (SymbolEntry *) gSymbolDataEnd;
            symbolP = symbolP->next;
            gSymbolDataEnd += sizeof(SymbolEntry);
            symbolP->restOfSymbol = rest;
            symbolP->next = 0;
            symbolP->latestBlockNum = -1;
         }
     }
    return symbolP;
}

/* returns a pointer to a block in memory.  If the
block is not already in memory, it is loaded from disk. */
CachedBlock *getBlock(Header *dataP, short blockNum)
{
    CachedBlock     *blockP;
    ParamBlockRec   readPB;

    blockP = gBlockLocation[blockNum];
    if (blockP == 0)
     {  /* need to load it from disk */
        if (gOldestBlockP->blockNum >= 0)
            gBlockLocation[gOldestBlockP->blockNum] = 0;
        gOldestBlockP->blockNum = blockNum;
        readPB.ioParam.ioCompletion = 0;
        readPB.ioParam.ioRefNum = gRefNum;
        readPB.ioParam.ioBuffer = (Ptr) gOldestBlockP->data;
        readPB.ioParam.ioReqCount = kBlockSize;
        readPB.ioParam.ioPosMode = fsFromStart
                             | (1 << 5); /* don't cache */
        readPB.ioParam.ioPosOffset = ((long) blockNum)
                                         * kBlockSize;
        PBRead(&readPB, true);
        gBlockLocation[blockNum] = gOldestBlockP;

        /* find the next oldest block */
        if (gOldestBlockP == gLowestBlockP)
            gOldestBlockP = gHighestBlockP;
        else
            gOldestBlockP--;
        if (gOldestBlockP == gCurrentOutBlockP)
            if (gOldestBlockP == gLowestBlockP)
                gOldestBlockP = gHighestBlockP;
            else
                gOldestBlockP--;
        if ((gPB.ioParam.ioResult > 0)
                && (((Ptr) gOldestBlockP->data)
                         == gPB.ioParam.ioBuffer))
            if (gOldestBlockP == gLowestBlockP)
                gOldestBlockP = gHighestBlockP;
            else
                gOldestBlockP--;
        if (gOldestBlockP == gCurrentOutBlockP)
            if (gOldestBlockP == gLowestBlockP)
                gOldestBlockP = gHighestBlockP;
            else
                gOldestBlockP--;
        /* wait for read to complete */
        while (readPB.ioParam.ioResult > 0)
            ;
     }
    return blockP;
}


void writeCurrentBlock(Header *dataP)
{
    CachedBlock *p;

    /* make sure any previous write is done */
    while (gPB.ioParam.ioResult > 0)
        ;
    gPB.ioParam.ioRefNum = gRefNum;
    gPB.ioParam.ioBuffer = (Ptr) gCurrentOutBlockP->data;
    gPB.ioParam.ioReqCount = kBlockSize;
    gPB.ioParam.ioPosMode = fsFromStart;
    gPB.ioParam.ioPosOffset =
        ((long) gCurrentOutBlockP->blockNum) * kBlockSize;
    PBWrite(&gPB, true);

    /* find the next oldest block */
    p = gOldestBlockP;
    if (gOldestBlockP == gLowestBlockP)
        gOldestBlockP = gHighestBlockP;
    else
        gOldestBlockP--;
    if (gOldestBlockP == gCurrentOutBlockP)
        if (gOldestBlockP == gLowestBlockP)
            gOldestBlockP = gHighestBlockP;
        else
            gOldestBlockP--;
    gCurrentOutBlockP = p;
}


void NewTrade(void *privateP, Trade trade)
{
    Header          *dataP = (Header *) privateP;
    ulong           seconds;
    TradeSummary    *tradeP;
    SymbolEntry     *symbolP;

    seconds = timeToSeconds(&trade.time);

    if (gCurrentEntryNum == kBlockEntries)
     {  /* current block is full, so write it out */
        short   curBlockNum = gCurrentOutBlockP->blockNum;

        writeCurrentBlock(dataP);
        gCurrentEntryNum = 0;
        if (gCurrentOutBlockP->blockNum >= 0)
            gBlockLocation[gCurrentOutBlockP->blockNum] = 0;
        gCurrentOutBlockP->blockNum = curBlockNum + 1;
        gBlockLocation[gCurrentOutBlockP->blockNum] =
                                     gCurrentOutBlockP;
     }
    symbolP = findInsertSymbolEntry(dataP, trade.symbol);
    tradeP = &gCurrentOutBlockP->data[gCurrentEntryNum];
    tradeP->seconds = seconds;
    tradeP->price = trade.price;
    tradeP->totalShares = trade.numShares;
    if (symbolP->latestBlockNum >= 0)
     {
        CachedBlock *blockP;
        TradeSummary *prevTradeP;
        ulong       todaySeconds;

        todaySeconds = seconds - (seconds % kSecondsPerDay);
        blockP = getBlock(dataP, symbolP->latestBlockNum);
        prevTradeP = &blockP->data[symbolP->latestEntryNum];
        if (prevTradeP->seconds >= todaySeconds)
            tradeP->totalShares += prevTradeP->totalShares;
     }
    tradeP->prevBlockNum = symbolP->latestBlockNum;
    tradeP->prevEntryNum = symbolP->latestEntryNum;
    symbolP->latestBlockNum = gCurrentOutBlockP->blockNum;
    symbolP->latestEntryNum = gCurrentEntryNum;
    gCurrentEntryNum++;
}


Fixed PriceIs(void *privateP, Str7 symbol, TimeStamp time)
{
    Header      *dataP = (Header *) privateP;
    ulong       seconds;
    SymbolEntry *symbolP;

    seconds = timeToSeconds(&time);
    symbolP = findSymbolEntry(dataP, symbol);
    if (symbolP)
     {
        CachedBlock     *blockP;
        TradeSummary    *tradeP;

        /* look for the most recent trade that is before
            or on the requested time */
        blockP = getBlock(dataP, symbolP->latestBlockNum);
        tradeP = &blockP->data[symbolP->latestEntryNum];
        while (tradeP->seconds > seconds)
         {
            if (tradeP->prevBlockNum < 0)
                return 0;
            blockP = getBlock(dataP, tradeP->prevBlockNum);
            tradeP = &blockP->data[tradeP->prevEntryNum];
         }
        return tradeP->price;
     }
    else
        return 0;
}


ulong VolumeIs(void *privateP, Str7 symbol, TimeStamp time)
{
    Header      *dataP = (Header *) privateP;
    ulong       seconds;
    SymbolEntry *symbolP;
    ulong       todaySeconds;

    seconds = timeToSeconds(&time);
    todaySeconds = seconds - (seconds % kSecondsPerDay);
    symbolP = findSymbolEntry(dataP, symbol);
    if (symbolP)
     {
        CachedBlock     *blockP;
        TradeSummary    *tradeP;

        /* look for the most recent trade that is before
            the requested time and on the same day */
        blockP = getBlock(dataP, symbolP->latestBlockNum);
        tradeP = &blockP->data[symbolP->latestEntryNum];
        while (tradeP->seconds >= seconds)
         {
            if (tradeP->prevBlockNum < 0)
                return 0;
            blockP = getBlock(dataP, tradeP->prevBlockNum);
            tradeP = &blockP->data[tradeP->prevEntryNum];
         }
        if (tradeP->seconds < todaySeconds)
            return 0;
        else
            return tradeP->totalShares;
     }
    else
        return 0;

}
