/* =========================================================
	ChessMoveEnumerator.c
	Copyright 1995 Gary Beith

	Objective: To determine all legal moves (excluding
	castling and en passant captures) available for a
	given chess position. It is NOT required to single
	out intelligent moves!
========================================================= */

/* ---------------------------------------------------------
	Items required by the problem statement
--------------------------------------------------------- */

#pragma options ( mc68020, !mc68881, require_protos )
#pragma options ( pack_enums, align_arrays )

typedef enum { rowA = 0, rowB, rowC, rowD, rowE, rowF, rowG,
						   rowH } Row;
typedef enum { col1 = 0, col2, col3, col4, col5, col6, col7,
						   col8 } Col;
typedef enum { whiteSide = 0, blackSide } Side;
typedef enum { king = 0, queen, rook, bishop, knight, pawn }
						   ChessPiece;

typedef struct Square {
		Row							row;
		Col							col;
} Square;

typedef struct PiecePosition {
		Square					sq;
		Side						side;
		ChessPiece			piece;
} PiecePosition;

typedef struct ChessMove {
		Square					fromSq;
		Square					toSq;
		Boolean					moveIsCapture;
		Boolean					opponentPlacedInCheck;
} ChessMove;

short /* numberOfMoves */ LegalChessMoves(
		PiecePosition		piecePositionArray[],
		short						numberOfPieces,
		Side						sideToMove,
		ChessMove				legalMoveArray[],
		void						*privateDataPtr
);

void InitChess(void *privateDataPtr);


/* =========================================================
	From here on, I'm on my own :-)
========================================================= */

#define kMaxChessPiece				pawn

typedef Square *SquarePtr;
typedef PiecePosition *PiecePosPtr;
typedef ChessMove *ChessMovePtr;

typedef enum {
		stmPawnSq0 = 0, stmPawnSq1, stmPieceSq0, stmPieceSq1,
		offBoardSq0, offBoardSq1, emptySq0, emptySq1,
		oppPieceSq0, oppPieceSq1, oppPawnSq0, oppPawnSq1
} MySquare, *MySquarePtr;

/* ---------------------------------------------------------
	The attack flag is used to denote squares near the King
	known to be under attack, thus preventing the King from
	moving to them.  State "0" for each square type represents
	attack flag not set, state "1" represents flag set.  The
	attack mask is used to clear the flag when testing for
	empty squares.
--------------------------------------------------------- */
#define kAttackFlag						0x01
#define kAttackMask						0x0e

typedef char MySquareIncr, *MySqIncrPtr;
typedef unsigned char MyAxis, *MyAxisPtr;
typedef short MySquareDiff, *MySqDiffPtr;

typedef struct {
		MySqIncrPtr			sqIncrArrayPtr;
		MyAxisPtr				axisArrayPtr;
		short						numberOfAxes;
} MyPieceTable;

typedef struct {
		MySquarePtr			sqPtr;
		MyPieceTable		pcTable;
		Square					fromSq;
		MyAxis					discAxis, pinAxis;
} MyPiece, *MyPiecePtr, **MyPieceHandle;

typedef struct {
		MyPieceHandle		pcHandle;
		MyPieceTable		pcTable;
		MySquare				sqType;
} MyPieceInfo, *MyPcInfoPtr;

/* ---------------------------------------------------------
	My internal square numbering scheme:
						 ** Opponent **
	[175] -   -   - .... -   - [185]
	  -   -   -   - .... -   -   -   -   -   -   - [174]
	  -   - (145) - .... - (152) -   -   -   -   -   -
				 < six more rows here >
	  -   - ( 33) - .... - ( 40) -   -   -   -   -   -
	[ 15] -   -   - .... -   -   -   -   -   -   -   -
	    [  0] -   - .... -   -   -   -   -   -   - [ 14]
					 ** Side to move **

	Numbers in parentheses indicate corners of the actual
	chess board.  Some reasons for choosing this scheme:
	1.	A border of at least two off-board squares is
			provided around the board to facilitate the
			determination of valid Knight moves.
	2.	Adjacent squares in a column have an absolute index
			difference of 16 (efficient power of 2).  A difference
			of at least 15 was required so that index values for
			squares in the same row would always differ by less
			than those for squares in different rows.
	3.	The perspective of the side to move is used rather
			than that of either White or Black exclusively.  This
			enables the identical move enumeration algorithm to
			be used for either side - differences such as pawn
			movement direction are accounted for when converting
			from and to the problem statement's board format.
--------------------------------------------------------- */
#define kLowerLeftSqIndex			33
#define kMin4thRankSqIndex		81
#define kMin5thRankSqIndex		97
#define kMin8thRankSqIndex		145
#define kUpperRightSqIndex		152
#define kMaxSqIndex						185

/*	Maximum square index differential on the board =
		kUpperRightSqIndex - kLowerLeftSqIndex */
#define kMaxSqDiff						119

/* ---------------------------------------------------------
	For purposes of this program, "long move" pieces are
	defined as consisting of Queens, Rooks, and Bishops.
	These pieces may move as many unobstructed spaces in
	each of their permitted directions as are available on
	the board.  The maximum counts given for long move pieces
	and Knights allow for the potential for all eight Pawns
	to promote to such pieces.
--------------------------------------------------------- */
#define kMaxLongMovePieces		13
#define kMaxKnights						10
#define kMaxPawns							8
					
typedef struct {
		MyPiecePtr			endSTMKingPtr, endOppKingPtr,
										endSTMLMPiecePtr, endOppLMPiecePtr,
										endSTMKnightPtr, endOppKnightPtr,
										endSTMPawnPtr;
		MySquare				sqArray[kMaxSqIndex + 1];
} MyChessBoard;

typedef struct {
		MyChessBoard		chessBoard, emptyBoard;
		MyPiece					stmKing, oppKing,
										stmLMPieceArray[kMaxLongMovePieces],
										oppLMPieceArray[kMaxLongMovePieces],
										stmKnightArray[kMaxKnights],
										oppKnightArray[kMaxKnights],
										stmPawnArray[kMaxPawns];
		MyPiecePtr			pcPtrArray[kMaxSqIndex + 1];
		Square					whiteSqArray[kMaxSqIndex + 1],
										blackSqArray[kMaxSqIndex + 1];
		SquarePtr				stmSqArrayPtr;
		MyPieceInfo			stmPcInfoArray[kMaxChessPiece + 1],
										oppPcInfoArray[kMaxChessPiece + 1];
		MySquareIncr		queenSqIncrArray[2 * kMaxSqDiff + 1],
										rookSqIncrArray[2 * kMaxSqDiff + 1],
										bishopSqIncrArray[2 * kMaxSqDiff + 1];
		Boolean					knightMoveArray[2 * kMaxSqDiff + 1];
		MyAxis					queenAxisArray[4], rookAxisArray[2],
										bishopAxisArray[2];
		MySquareDiff		knightSqDiffArray[8],
										kingSqDiffArray[8];
} MyData, *MyDataPtr;

/* ---------------------------------------------------------
	Local function prototypes
--------------------------------------------------------- */

void SetUpMyChessBoard(MyDataPtr dataPtr,
											 PiecePosition piecePositionArray[],
											 short numberOfPieces,
											 Side sideToMove);

short FindChecksAndPins(MyDataPtr dataPtr,
												MySquarePtr *checkSqHandle,
												MySquareIncr *checkSqIncrPtr);

void EnumerateNonKingMoves(MyDataPtr dataPtr,
													 ChessMovePtr *moveHandle);

void EnumerateNonKingCapturesOfCheckingPiece
														 (MyDataPtr dataPtr,
														  MySquarePtr checkSqPtr,
															ChessMovePtr *moveHandle);

void EnumerateInterpositionMoves(MyDataPtr dataPtr,
																 MySquarePtr checkSqPtr,
																 MySquareIncr checkSqIncr,
																 ChessMovePtr *moveHandle);

void EnumerateKingMoves(MyDataPtr dataPtr,
												ChessMovePtr *moveHandle);


/* =========================================================
	Timed legal move enumeration routine
========================================================= */

short LegalChessMoves(PiecePosition piecePositionArray[],
											short numberOfPieces,
											Side sideToMove,
											ChessMove legalMoveArray[],
											void *privateDataPtr)
{
	MyDataPtr				dataPtr = privateDataPtr;
	short						numberOfChecks;
	MySquarePtr			checkSqPtr;
	MySquareIncr		checkSqIncr;
	ChessMovePtr		movePtr = legalMoveArray;

	SetUpMyChessBoard(dataPtr, piecePositionArray,
										numberOfPieces, sideToMove);

	numberOfChecks = FindChecksAndPins(dataPtr, &checkSqPtr,
																		 &checkSqIncr);

	if (numberOfChecks == 0)
	{
		EnumerateNonKingMoves(dataPtr, &movePtr);
	}
	else if (numberOfChecks == 1)
	{
		EnumerateNonKingCapturesOfCheckingPiece(dataPtr,
																						checkSqPtr,
																						&movePtr);
		if (checkSqIncr)
		{
			EnumerateInterpositionMoves(dataPtr, checkSqPtr,
																	checkSqIncr, &movePtr);
		}
	}
	/* else if (numberOfChecks == 2)
			 Only King moves are possible (if any exist). */

	EnumerateKingMoves(dataPtr, &movePtr);

	return (movePtr - legalMoveArray); /* Number of moves */
}


/* ------------------------------------------------------ */

void SetUpMyChessBoard(MyDataPtr dataPtr,
											 PiecePosition piecePositionArray[],
											 short numberOfPieces,
											 Side sideToMove)
{
	PiecePosPtr			posPtr, endPosPtr;
	short						sqIndex;
	MyPcInfoPtr			pcInfoPtr;
	MySquarePtr			sqPtr;
	MyPiecePtr			pcPtr;

	dataPtr->chessBoard = dataPtr->emptyBoard;

	posPtr = piecePositionArray;
	endPosPtr = posPtr + numberOfPieces;

	if (sideToMove)	/* == blackSide */
	{
		dataPtr->stmSqArrayPtr = dataPtr->blackSqArray;
		do
		{
			sqIndex = kUpperRightSqIndex -
								((posPtr->sq.row << 4) + posPtr->sq.col);

			pcInfoPtr = (posPtr->side) ?
									dataPtr->stmPcInfoArray + posPtr->piece :
									dataPtr->oppPcInfoArray + posPtr->piece;

			sqPtr	= dataPtr->chessBoard.sqArray + sqIndex;
			if ((*sqPtr = pcInfoPtr->sqType) != oppPawnSq0)
			{
				dataPtr->pcPtrArray[sqIndex] =
						pcPtr = (*pcInfoPtr->pcHandle)++;

				pcPtr->sqPtr = sqPtr;
				pcPtr->pcTable = pcInfoPtr->pcTable;
				pcPtr->fromSq = posPtr->sq;
				pcPtr->discAxis = pcPtr->pinAxis = 0;
			}
		}
		while (++posPtr != endPosPtr);
	}
	else
	{
		dataPtr->stmSqArrayPtr = dataPtr->whiteSqArray;
		do
		{
			sqIndex = kLowerLeftSqIndex +
								((posPtr->sq.row << 4) + posPtr->sq.col);

			pcInfoPtr = (posPtr->side) ?
									dataPtr->oppPcInfoArray + posPtr->piece :
									dataPtr->stmPcInfoArray + posPtr->piece;

			sqPtr	= dataPtr->chessBoard.sqArray + sqIndex;
			if ((*sqPtr = pcInfoPtr->sqType) != oppPawnSq0)
			{
				dataPtr->pcPtrArray[sqIndex] =
						pcPtr = (*pcInfoPtr->pcHandle)++;

				pcPtr->sqPtr = sqPtr;
				pcPtr->pcTable = pcInfoPtr->pcTable;
				pcPtr->fromSq = posPtr->sq;
				pcPtr->discAxis = pcPtr->pinAxis = 0;
			}
		}
		while (++posPtr != endPosPtr);
	}
}


/* ------------------------------------------------------ */

short FindChecksAndPins(MyDataPtr dataPtr,
												MySquarePtr *checkSqHandle,
												MySquareIncr *checkSqIncrPtr)
{
	MySquarePtr			kingSqPtr, sqPtr;
	MyPiecePtr 			endPcPtr, pcPtr;
	short						sqIncr, stmPcSqIndex, numberOfChecks = 0;

	/* -------------------------------------------------------
		Since I must report all moves which place the opponent
		in check, I first scan for possible discovered checks.
		These occur when only one piece (which must belong to
		the side to move) stands between one of the mover's
		long move pieces and the enemy King, and that piece
		is moved off of the line of attack.  Because I am only
		required to analyze legal positions, I know that at
		least one piece will exist on any such line of attack.
	------------------------------------------------------- */
	kingSqPtr = dataPtr->oppKing.sqPtr;

	endPcPtr = dataPtr->chessBoard.endSTMLMPiecePtr;
	for (pcPtr = dataPtr->stmLMPieceArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		sqPtr = pcPtr->sqPtr;
		sqIncr = pcPtr->pcTable.sqIncrArrayPtr
						 [kingSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
			continue;			/* King not "in line" with this piece */

		sqPtr += sqIncr;
		stmPcSqIndex = 0;
		while (*sqPtr <= emptySq1)	/* Not opponent's piece */
		{
			if (*sqPtr < emptySq0)		/* Is mover's piece */
			{
				if (stmPcSqIndex)
					break;		/* More than one piece in its path */

				stmPcSqIndex = sqPtr - dataPtr->chessBoard.sqArray;
			}

			if ((sqPtr += sqIncr) == kingSqPtr)
			{
				/* Record line of attack for future reference */
				dataPtr->pcPtrArray[stmPcSqIndex]->discAxis =
						(sqIncr < 0) ? -sqIncr : +sqIncr;
				break;
			}
		}
	}

	/* -------------------------------------------------------
		A similar analysis determines if any pieces are
		"pinned," i.e., they stand in the way of an attack on
		the mover's King by an enemy piece.  A pinned piece can
		move only along the line of attack (if such a move is
		permitted for the piece; in particular, a pinned Knight
		is never able to move).
	------------------------------------------------------- */
	kingSqPtr = dataPtr->stmKing.sqPtr;

	endPcPtr = dataPtr->chessBoard.endOppLMPiecePtr;
	for (pcPtr = dataPtr->oppLMPieceArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		sqPtr = pcPtr->sqPtr;
		sqIncr = pcPtr->pcTable.sqIncrArrayPtr
						 [kingSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
			continue;

		if ((sqPtr += sqIncr) == kingSqPtr)
		{
			/* ---------------------------------------------------
				Piece is adjacent to King.  The King may capture it
				if it is not protected by another piece.  However,
				no interposition moves are possible.
			--------------------------------------------------- */
			sqPtr[+sqIncr] |= kAttackFlag;
			*checkSqHandle = pcPtr->sqPtr;
			*checkSqIncrPtr = 0;
			numberOfChecks++;
			continue;
		}

		stmPcSqIndex = 0;
		while (*sqPtr <= emptySq1)
		{
			if (*sqPtr < emptySq0)
			{
				if (stmPcSqIndex)
					break;

				stmPcSqIndex = sqPtr - dataPtr->chessBoard.sqArray;
			}

			if ((sqPtr += sqIncr) == kingSqPtr)
			{
				if (stmPcSqIndex) /* Pinned piece was found */
				{
					dataPtr->pcPtrArray[stmPcSqIndex]->pinAxis =
							(sqIncr < 0) ? -sqIncr : +sqIncr;
				}
				else
				{
					/* -----------------------------------------------
						King is in check "from a distance."  If the King
						moves, it must leave the line of attack, which
						is recorded so that interposition moves can be
						determined later on.
					----------------------------------------------- */
					sqPtr[+sqIncr] |= kAttackFlag;
					sqPtr[-sqIncr] |= kAttackFlag;
					*checkSqHandle = pcPtr->sqPtr;
					*checkSqIncrPtr = sqIncr;
					numberOfChecks++;
				}

				break;
			}
		}
	}

	/* -------------------------------------------------------
		The maximum number of checks possible is two, only one
		of which can come from a Pawn or Knight.
	------------------------------------------------------- */
	if (numberOfChecks < 2)
	{
		/* -----------------------------------------------------
			Per ANSI standard, a sequence point occurs after the
			evaluation of each operand of the Logical Or and
			Logical And operators.  Therefore, I can validly rely
			on the value of sqPtr being set in exactly the order
			specified in the next expression (and others like it).
			That is not generally true of most other operators.
		----------------------------------------------------- */
		if ((*(sqPtr = kingSqPtr + 15) >= oppPawnSq0) ||
				(*(sqPtr += 2) >= oppPawnSq0))
		{
			*checkSqHandle = sqPtr;
			*checkSqIncrPtr = 0;
			numberOfChecks++;
		}
		else
		{
			endPcPtr = dataPtr->chessBoard.endOppKnightPtr;
			for (pcPtr = dataPtr->oppKnightArray;
					 pcPtr < endPcPtr; pcPtr++)
			{
				sqPtr = pcPtr->sqPtr;
				if (dataPtr->knightMoveArray
						[kingSqPtr - sqPtr + kMaxSqDiff])
				{
					*checkSqHandle = sqPtr;
					*checkSqIncrPtr = 0;
					numberOfChecks++;
					break;
				}
			}
		}
	}

	return numberOfChecks;
}


/* ------------------------------------------------------ */

void EnumerateNonKingMoves(MyDataPtr dataPtr,
													 ChessMovePtr *moveHandle)
{
	MySquarePtr			oppKingSqPtr, toSqPtr, sqPtr;
	MyPiecePtr			endPcPtr, pcPtr;
	MyAxisPtr				axisPtr, endAxisPtr;
	short						toSqIncr, sqIncr, toSqIndex;
	ChessMovePtr		movePtr = *moveHandle;
	MySqDiffPtr			sqDiffPtr, endSqDiffPtr;

	oppKingSqPtr = dataPtr->oppKing.sqPtr;

	endPcPtr = dataPtr->chessBoard.endSTMLMPiecePtr;
	for (pcPtr = dataPtr->stmLMPieceArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		/* -----------------------------------------------------
			The move table for long move pieces consists of a set
			of axes of movement for the particular type of piece
			(i.e., Queen, Rook, or Bishop).  The positive and
			negative values of each axis are used as square
			increments, beginning at the piece's current square
			pointer, and continuing until further movement is
			blocked by another piece or by the end ofthe board.
			In the event the blockage is caused by an oppenent's
			piece, one additional move results; that of capturing
			the piece by moving onto its square.
		----------------------------------------------------- */
		axisPtr = pcPtr->pcTable.axisArrayPtr;
		endAxisPtr = axisPtr + pcPtr->pcTable.numberOfAxes;
		do
		{
			if (pcPtr->pinAxis && (pcPtr->pinAxis != *axisPtr))
				continue; /* Piece is pinned on some other axis */

			toSqIncr = *axisPtr;
			do
			{
				toSqPtr = pcPtr->sqPtr;
				do
				{
					if (*(toSqPtr += toSqIncr) < emptySq0)
						break; /* Square off-board or holds own piece */

					movePtr->fromSq = pcPtr->fromSq;
					movePtr->toSq =
							dataPtr->stmSqArrayPtr
							[toSqPtr - dataPtr->chessBoard.sqArray];
					movePtr->moveIsCapture = (*toSqPtr > emptySq1);

					/* -----------------------------------------------
						If a long move piece has a discovered check
						axis, it is not able to move along that axis.
						Otherwise, it would be checking the enemy King,
						and the position would not be legal with this
						side to move.  Therefore, any move of such a
						piece will result in a discovered check.
					----------------------------------------------- */
					if (pcPtr->discAxis)
					{
						(movePtr++)->opponentPlacedInCheck = true;
						continue;
					}

					sqIncr = pcPtr->pcTable.sqIncrArrayPtr
									 [oppKingSqPtr - toSqPtr + kMaxSqDiff];
					if (sqIncr == 0)
					{
						(movePtr++)->opponentPlacedInCheck = false;
						continue;
					}

					/* -----------------------------------------------
						Scan along the path to the enemy King from the
						target square for this move.  If the first
						occupied square is that of the King, a check
						will result from the move; otherwise, no check.
					----------------------------------------------- */
					sqPtr = toSqPtr;
					do
					{
						sqPtr += sqIncr;
					}
					while ((*sqPtr & kAttackMask) == emptySq0);

					(movePtr++)->opponentPlacedInCheck =
							(sqPtr == oppKingSqPtr);
				}
				while (*toSqPtr <= emptySq1);
			}
			while ((toSqIncr = -toSqIncr) < 0);
		}
		while (++axisPtr != endAxisPtr);
	}

	endPcPtr = dataPtr->chessBoard.endSTMKnightPtr;
	for (pcPtr = dataPtr->stmKnightArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		if (pcPtr->pinAxis)
			continue;

		toSqPtr = pcPtr->sqPtr;

		/* -----------------------------------------------------
			Unlike long move pieces, each move of a Knight may be
			evaluated independently from all of its other moves.
			Thus, the move table consists of a set of relative
			square differentials, each added in turn to the
			previous move's square pointer, rather than basing all
			of the moves on the Knight's current square pointer.
		----------------------------------------------------- */
		sqDiffPtr = dataPtr->knightSqDiffArray;
		endSqDiffPtr = sqDiffPtr + 8;
		do
		{
			if (*(toSqPtr += *sqDiffPtr) >= emptySq0)
			{
				movePtr->fromSq = pcPtr->fromSq;
				movePtr->toSq =
						dataPtr->stmSqArrayPtr
						[toSqPtr - dataPtr->chessBoard.sqArray];
				movePtr->moveIsCapture = (*toSqPtr > emptySq1);
				(movePtr++)->opponentPlacedInCheck =
						(pcPtr->discAxis ||
						 dataPtr->knightMoveArray
						 [oppKingSqPtr - toSqPtr + kMaxSqDiff]);
			}
		}
		while (++sqDiffPtr != endSqDiffPtr);
	}

	endPcPtr = dataPtr->chessBoard.endSTMPawnPtr;
	for (pcPtr = dataPtr->stmPawnArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		/* -----------------------------------------------------
			Pawns require two different move evaluations; one for
			captures, and a different one for non-capture moves.
		----------------------------------------------------- */
		for (toSqIncr = 15; toSqIncr <= 17; toSqIncr += 2)
		{
			if ((pcPtr->pinAxis &&
					 (pcPtr->pinAxis != toSqIncr)) ||
					(*(toSqPtr = pcPtr->sqPtr + toSqIncr) <=
					 emptySq1))
			{
				continue; /* Pawn pinned or no capture available */
			}

			toSqIndex = toSqPtr - dataPtr->chessBoard.sqArray;

			movePtr->fromSq = pcPtr->fromSq;
			movePtr->toSq = dataPtr->stmSqArrayPtr[toSqIndex];
			movePtr->moveIsCapture = true;

			if (pcPtr->discAxis)
			{
				(movePtr++)->opponentPlacedInCheck = true;
				continue;
			}

			if (toSqIndex < kMin8thRankSqIndex)
			{
				(movePtr++)->opponentPlacedInCheck =
						(((toSqPtr + 15) == oppKingSqPtr) ||
						 ((toSqPtr + 17) == oppKingSqPtr));
				continue;
			}

			/* ---------------------------------------------------
				Pawn promotion brings up two interesting points.
				One, the value of "opponentPlacedInCheck" becomes
				ambiguous, since the Pawn can promote to a Queen,
				Rook, Bishop, or Knight, and the squares that it
				attacks will vary accordingly.  Since a Pawn most
				often promotes to a Queen, I have chosen to set
				"opponentPlacedInCheck" on that basis.  The second
				point concerns the "ghost image" of the Pawn after
				it moves.  In other cases, it does not matter that
				the piece being moved is still recorded as being on
				its original square, since a line of attack to the
				enemy King which goes through that square must be
				blocked anyway; otherwise, the piece would have
				checked the King from its original square, and the
				position would be illegal.  In the case of a
				promoted Pawn, however, this assumption does not
				hold true.  Therefore, I compare the square pointer
				with the Pawn's original position; if they match, I
				treat the square as though it were empty.
			--------------------------------------------------- */
			sqIncr = dataPtr->queenSqIncrArray
							 [oppKingSqPtr - toSqPtr + kMaxSqDiff];
			if (sqIncr == 0)
			{
				(movePtr++)->opponentPlacedInCheck = false;
				continue;
			}

			sqPtr = toSqPtr;
			do
			{
				sqPtr += sqIncr;
			}
			while (((*sqPtr & kAttackMask) == emptySq0) ||
						 (sqPtr == pcPtr->sqPtr));

			(movePtr++)->opponentPlacedInCheck =
					(sqPtr == oppKingSqPtr);
		}

		if (pcPtr->pinAxis & 0x0f) /* i.e., axis not 0 or 16 */
			continue;

		/* -----------------------------------------------------
			Ordinarily, the Pawn advances only one square, but it
			may continue to a second square if it is not blocked
			and has not yet reached the fourth rank.
		----------------------------------------------------- */
		toSqPtr = pcPtr->sqPtr;
		do
		{
			if ((*(toSqPtr += 16) & kAttackMask) != emptySq0)
				break;

			toSqIndex = toSqPtr - dataPtr->chessBoard.sqArray;

			movePtr->fromSq = pcPtr->fromSq;
			movePtr->toSq = dataPtr->stmSqArrayPtr[toSqIndex];
			movePtr->moveIsCapture = false;

			if (pcPtr->discAxis & 0x0f)
			{
				(movePtr++)->opponentPlacedInCheck = true;
				continue;
			}

			if (toSqIndex < kMin8thRankSqIndex)
			{
				(movePtr++)->opponentPlacedInCheck =
						(((toSqPtr + 15) == oppKingSqPtr) ||
						 ((toSqPtr + 17) == oppKingSqPtr));
				continue;
			}

			sqIncr = dataPtr->queenSqIncrArray
							 [oppKingSqPtr - toSqPtr + kMaxSqDiff];
			if (sqIncr == 0)
			{
				(movePtr++)->opponentPlacedInCheck = false;
				break;
			}

			sqPtr = toSqPtr;
			do
			{
				sqPtr += sqIncr;
			}
			while (((*sqPtr & kAttackMask) == emptySq0) ||
						 (sqPtr == pcPtr->sqPtr));

			(movePtr++)->opponentPlacedInCheck =
					(sqPtr == oppKingSqPtr);
			break;
		}
		while (toSqIndex < kMin4thRankSqIndex);
	}

	*moveHandle = movePtr;
}


/* ------------------------------------------------------ */

void EnumerateNonKingCapturesOfCheckingPiece
														 (MyDataPtr dataPtr,
														  MySquarePtr checkSqPtr,
															ChessMovePtr *moveHandle)
{
	MySquarePtr			oppKingSqPtr, sqPtr;
	short						checkSqIndex, sqIncr, fromSqIncr;
	Square					checkSq;
	MyPiecePtr			endPcPtr, pcPtr;
	ChessMovePtr		movePtr = *moveHandle;

	oppKingSqPtr = dataPtr->oppKing.sqPtr;	

	checkSqIndex = checkSqPtr - dataPtr->chessBoard.sqArray;
	checkSq = dataPtr->stmSqArrayPtr[checkSqIndex];

	endPcPtr = dataPtr->chessBoard.endSTMLMPiecePtr;
	for (pcPtr = dataPtr->stmLMPieceArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		if (pcPtr->pinAxis)
			continue;

		sqPtr = pcPtr->sqPtr;
		sqIncr = pcPtr->pcTable.sqIncrArrayPtr
						 [checkSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
			continue;

		do
		{
			sqPtr += sqIncr;
		}
		while((*sqPtr & kAttackMask) == emptySq0);

		if (sqPtr != checkSqPtr)
			continue;

		movePtr->fromSq = pcPtr->fromSq;
		movePtr->toSq = checkSq;
		movePtr->moveIsCapture = true;

		if (pcPtr->discAxis)
		{
			(movePtr++)->opponentPlacedInCheck = true;
			continue;
		}

		sqIncr = pcPtr->pcTable.sqIncrArrayPtr
						 [oppKingSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
		{
			(movePtr++)->opponentPlacedInCheck = false;
			continue;
		}

		do
		{
			sqPtr += sqIncr;
		}
		while ((*sqPtr & kAttackMask) == emptySq0);

		(movePtr++)->opponentPlacedInCheck =
				(sqPtr == oppKingSqPtr);
	}

	endPcPtr = dataPtr->chessBoard.endSTMKnightPtr;
	for (pcPtr = dataPtr->stmKnightArray;
			 pcPtr < endPcPtr; pcPtr++)
	{
		if (pcPtr->pinAxis == 0)
		{
			if (dataPtr->knightMoveArray
					[checkSqPtr - pcPtr->sqPtr + kMaxSqDiff])
			{
				movePtr->fromSq = pcPtr->fromSq;
				movePtr->toSq = checkSq;
				movePtr->moveIsCapture = true;
				(movePtr++)->opponentPlacedInCheck =
						(pcPtr->discAxis ||
						 dataPtr->knightMoveArray
						 [oppKingSqPtr - checkSqPtr + kMaxSqDiff]);
			}
		}
	}

	for (fromSqIncr = 15; fromSqIncr <= 17; fromSqIncr += 2)
	{
		if (*(sqPtr = checkSqPtr - fromSqIncr) > stmPawnSq1)
			continue; /* Square does not contain mover's Pawn */

		pcPtr = dataPtr->pcPtrArray
						[sqPtr - dataPtr->chessBoard.sqArray];
		if (pcPtr->pinAxis)
			continue;

		movePtr->fromSq = pcPtr->fromSq;
		movePtr->toSq = checkSq;
		movePtr->moveIsCapture = true;

		if (pcPtr->discAxis)
		{
			(movePtr++)->opponentPlacedInCheck = true;
			continue;
		}

		sqPtr = checkSqPtr;
		if (checkSqIndex < kMin8thRankSqIndex)
		{
			(movePtr++)->opponentPlacedInCheck =
					(((sqPtr + 15) == oppKingSqPtr) ||
					 ((sqPtr + 17) == oppKingSqPtr));
			continue;
		}

		sqIncr = dataPtr->queenSqIncrArray
						 [oppKingSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
		{
			(movePtr++)->opponentPlacedInCheck = false;
			continue;
		}

		do
		{
			sqPtr += sqIncr;
		}
		while (((*sqPtr & kAttackMask) == emptySq0) ||
					 (sqPtr == pcPtr->sqPtr));

		(movePtr++)->opponentPlacedInCheck =
				(sqPtr == oppKingSqPtr);
	}

	*moveHandle = movePtr;
}


/* ------------------------------------------------------ */

void EnumerateInterpositionMoves(MyDataPtr dataPtr,
																 MySquarePtr checkSqPtr,
																 MySquareIncr checkSqIncr,
																 ChessMovePtr *moveHandle)
{
	MySquarePtr			oppKingSqPtr, toSqPtr, sqPtr;
	short						toSqIndex, sqIncr;
	Square					toSq;
	MyPiecePtr			endPcPtr, pcPtr;
	ChessMovePtr		movePtr = *moveHandle;

	oppKingSqPtr = dataPtr->oppKing.sqPtr;	

	/* -------------------------------------------------------
		This "monster" do-loop repeats for each square which
		lies between the checking piece and the mover's King.
	------------------------------------------------------- */
	toSqPtr = checkSqPtr + checkSqIncr;
	do
	{
		toSqIndex = toSqPtr - dataPtr->chessBoard.sqArray;
		toSq = dataPtr->stmSqArrayPtr[toSqIndex];

		endPcPtr = dataPtr->chessBoard.endSTMLMPiecePtr;
		for (pcPtr = dataPtr->stmLMPieceArray;
				 pcPtr < endPcPtr; pcPtr++)
		{
			if (pcPtr->pinAxis)
				continue;

			sqPtr = pcPtr->sqPtr;
			sqIncr = pcPtr->pcTable.sqIncrArrayPtr
							 [toSqPtr - sqPtr + kMaxSqDiff];
			if (sqIncr == 0)
				continue;

			while ((sqPtr += sqIncr) != toSqPtr)
			{
				if ((*sqPtr & kAttackMask) != emptySq0)
					goto NextLMPiece; /* Break while, continue for */
			}

			movePtr->fromSq = pcPtr->fromSq;
			movePtr->toSq = toSq;
			movePtr->moveIsCapture = false;
	
			if (pcPtr->discAxis)
			{
				(movePtr++)->opponentPlacedInCheck = true;
				continue;
			}
	
			sqIncr = pcPtr->pcTable.sqIncrArrayPtr
							 [oppKingSqPtr - sqPtr + kMaxSqDiff];
			if (sqIncr == 0)
			{
				(movePtr++)->opponentPlacedInCheck = false;
				continue;
			}
	
			do
			{
				sqPtr += sqIncr;
			}
			while ((*sqPtr & kAttackMask) == emptySq0);
	
			(movePtr++)->opponentPlacedInCheck =
					(sqPtr == oppKingSqPtr);

		NextLMPiece:
			;
		}
	
		endPcPtr = dataPtr->chessBoard.endSTMKnightPtr;
		for (pcPtr = dataPtr->stmKnightArray;
				 pcPtr < endPcPtr; pcPtr++)
		{
			if (pcPtr->pinAxis == 0)
			{
				if (dataPtr->knightMoveArray
						[toSqPtr - pcPtr->sqPtr + kMaxSqDiff])
				{
					movePtr->fromSq = pcPtr->fromSq;
					movePtr->toSq = toSq;
					movePtr->moveIsCapture = false;
					(movePtr++)->opponentPlacedInCheck =
							(pcPtr->discAxis ||
							 dataPtr->knightMoveArray
							 [oppKingSqPtr - toSqPtr + kMaxSqDiff]);
				}
			}
		}

		/* -----------------------------------------------------
			Pawn interposition is not possible if the square in
			the same column in the previous row does not contain
			a mover's Pawn, AND any one of the following is true:
			1.	The square where interposition is to take place is
					not on the fourth rank,
			2.	The square in the same column on the third rank is
					not empty,
			3.	The square in the same column on the second rank
					does not contain a mover's Pawn.
		----------------------------------------------------- */
		if ((*(sqPtr = toSqPtr - 16) > stmPawnSq1) &&
				((toSqIndex >= kMin5thRankSqIndex) ||
				 (toSqIndex < kMin4thRankSqIndex) ||
				 ((*sqPtr & kAttackMask) != emptySq0) ||
				 (*(sqPtr -= 16) > stmPawnSq1)))
		{
			continue;
		}

		pcPtr = dataPtr->pcPtrArray
						[sqPtr - dataPtr->chessBoard.sqArray];
		if (pcPtr->pinAxis)
			continue;

		movePtr->fromSq = pcPtr->fromSq;
		movePtr->toSq = toSq;
		movePtr->moveIsCapture = false;

		if (pcPtr->discAxis & 0x0f)
		{
			(movePtr++)->opponentPlacedInCheck = true;
			continue;
		}

		sqPtr = toSqPtr;
		if (toSqIndex < kMin8thRankSqIndex)
		{
			(movePtr++)->opponentPlacedInCheck =
					(((sqPtr + 15) == oppKingSqPtr) ||
					 ((sqPtr + 17) == oppKingSqPtr));
			continue;
		}

		sqIncr = dataPtr->queenSqIncrArray
						 [oppKingSqPtr - sqPtr + kMaxSqDiff];
		if (sqIncr == 0)
		{
			(movePtr++)->opponentPlacedInCheck = false;
			continue;
		}

		do
		{
			sqPtr += sqIncr;
		}
		while (((*sqPtr & kAttackMask) == emptySq0) ||
					 (sqPtr == pcPtr->sqPtr));

		(movePtr++)->opponentPlacedInCheck =
				(sqPtr == oppKingSqPtr);
	}
	while (*(toSqPtr += checkSqIncr) >= emptySq0);

	*moveHandle = movePtr;
}


/* ------------------------------------------------------ */

void EnumerateKingMoves(MyDataPtr dataPtr,
												ChessMovePtr *moveHandle)
{
	MySquarePtr			stmKingSqPtr, sqPtr, toSqPtr;
	short						discAxis, sqIncr;
	MySqDiffPtr			sqDiffPtr, endSqDiffPtr;
	MyPiecePtr			endPcPtr, pcPtr;
	ChessMovePtr		movePtr = *moveHandle;

	stmKingSqPtr = dataPtr->stmKing.sqPtr;
	discAxis = dataPtr->stmKing.discAxis;

	sqPtr = dataPtr->oppKing.sqPtr;
	sqIncr = stmKingSqPtr - sqPtr;
	if ((sqIncr >= -34) && (sqIncr <= +34))
	{
		/* ----------------------------------------------------
			The opposing King might be within range of squares
			to which this King could move, so set the attack flag
			for all squares the opposing King attacks.  This will
			include some squares to which this King cannot move -
			it did not seem worth the trouble to code a fancy
			algorithm which ensures that only squares to which
			this King can move have the attack flag set.  In most
			normal chess positions (excluding the endgame), the
			two Kings are at opposite ends of the board, so this
			code won't be executed very often anyway.
		---------------------------------------------------- */
		sqDiffPtr = dataPtr->kingSqDiffArray;
		endSqDiffPtr = sqDiffPtr + 8;
		do
		{
			*(sqPtr += *sqDiffPtr) |= kAttackFlag;
		}
		while (++sqDiffPtr != endSqDiffPtr);
	}

	toSqPtr = stmKingSqPtr;

	sqDiffPtr = dataPtr->kingSqDiffArray;
	endSqDiffPtr = sqDiffPtr + 8;
	do
	{
		/* -----------------------------------------------------
			The King cannot move to this square if:
			1.	It is off-board or occupied by mover's own piece,
			2.	It is flagged as known to be under attack,
			3.	It is within capture range of an enemy Pawn.
		----------------------------------------------------- */
		if ((*(toSqPtr += *sqDiffPtr) < emptySq0) ||
				(*toSqPtr & kAttackFlag) ||
				(*(sqPtr = toSqPtr + 15) >= oppPawnSq0) ||
				(*(sqPtr += 2) >= oppPawnSq0))
		{
			continue;
		}

		/* -----------------------------------------------------
			The square is also blocked if a long move piece or
			Knight attacks the square.
		----------------------------------------------------- */
		endPcPtr = dataPtr->chessBoard.endOppLMPiecePtr;
		for (pcPtr = dataPtr->oppLMPieceArray;
			   pcPtr < endPcPtr; pcPtr++)
		{
			sqPtr = pcPtr->sqPtr;
			sqIncr = pcPtr->pcTable.sqIncrArrayPtr
							 [toSqPtr - sqPtr + kMaxSqDiff];
			if (sqIncr)
			{
				do
				{
					if ((sqPtr += sqIncr) == toSqPtr)
						goto NextKingMove;

					/* -----------------------------------------------
						Set an "audit trail" of attacked squares for
						this piece.  Most of these attack flags will be
						useless, but they cost very little to set, and
						they make the test for empty squares easier.
						If one does happen to fall in a square next to
						the mover's King, it will save time by removing
						that square from consideration immediately.
					----------------------------------------------- */
					*sqPtr |= kAttackFlag;
				}
				while (*sqPtr == emptySq1);
			}
		}

		endPcPtr = dataPtr->chessBoard.endOppKnightPtr;
		for (pcPtr = dataPtr->oppKnightArray;
			   pcPtr < endPcPtr; pcPtr++)
		{
			if (dataPtr->knightMoveArray
					[toSqPtr - pcPtr->sqPtr + kMaxSqDiff])
			{
				goto NextKingMove;
			}
		}

		movePtr->fromSq = dataPtr->stmKing.fromSq;
		movePtr->toSq =
				dataPtr->stmSqArrayPtr
				[toSqPtr - dataPtr->chessBoard.sqArray];
		movePtr->moveIsCapture = (*toSqPtr > emptySq1);

		/* -----------------------------------------------------
			The only checks which can result from a King move are
			discovered checks.
		----------------------------------------------------- */
		if (discAxis == 0)
		{
			(movePtr++)->opponentPlacedInCheck = false;
		}
		else
		{
			sqIncr = toSqPtr - stmKingSqPtr;
			(movePtr++)->opponentPlacedInCheck =
					((discAxis != sqIncr) && (discAxis != -sqIncr));
		}

	NextKingMove:
		;
	}
	while (++sqDiffPtr != endSqDiffPtr);

	*moveHandle = movePtr;
}


/* =========================================================
	Untimed initialization routine
========================================================= */

void InitChess(void *privateDataPtr)
{
	MyDataPtr				dataPtr = privateDataPtr;
	MySquarePtr			sqPtr;
	int							index, sqRow, sqCol;
	SquarePtr				whiteSqPtr, blackSqPtr;
	MySqIncrPtr			posQueenSqIncrPtr, negQueenSqIncrPtr,
									posRookSqIncrPtr, negRookSqIncrPtr,
									posBishopSqIncrPtr, negBishopSqIncrPtr;
	Boolean					*posKnightMovePtr, *negKnightMovePtr;

	/* -------------------------------------------------------
		Quoting from the problem statement: "The privateDataPtr
		parameter will be the same pointer provided to
		InitChess."  Therefore, I can set pointers based on its
		value in my private data structure.  This would not be
		possible if the parameter were guaranteed only to point
		to a copy of the same data, but not necessarily to have
		the same value.
	------------------------------------------------------- */
	dataPtr->emptyBoard.endSTMKingPtr = &dataPtr->stmKing;
	dataPtr->emptyBoard.endOppKingPtr = &dataPtr->oppKing;
	dataPtr->emptyBoard.endSTMLMPiecePtr =
			dataPtr->stmLMPieceArray;
	dataPtr->emptyBoard.endOppLMPiecePtr =
			dataPtr->oppLMPieceArray;
	dataPtr->emptyBoard.endSTMKnightPtr =
			dataPtr->stmKnightArray;
	dataPtr->emptyBoard.endOppKnightPtr =
			dataPtr->oppKnightArray;
	dataPtr->emptyBoard.endSTMPawnPtr =
			dataPtr->stmPawnArray;

	/* -------------------------------------------------------
		Initialize the empty chess board structure and the
		square coordinate arrays.
	------------------------------------------------------- */
	sqPtr = dataPtr->emptyBoard.sqArray;
	for (index = 0; index <= kMaxSqIndex; index++)
	{
		*sqPtr++ = offBoardSq0;
	}

	sqPtr = dataPtr->emptyBoard.sqArray + kLowerLeftSqIndex;
	whiteSqPtr = dataPtr->whiteSqArray + kLowerLeftSqIndex;
	blackSqPtr = dataPtr->blackSqArray + kUpperRightSqIndex;
	for (sqRow = 0; sqRow < 8; sqRow++)
	{
		for (sqCol = 0; sqCol < 8; sqCol++)
		{
			*sqPtr++	= emptySq0;
			whiteSqPtr->row = sqRow;
			blackSqPtr->row = sqRow;
			(whiteSqPtr++)->col = sqCol;
			(blackSqPtr--)->col = sqCol;
		}
		sqPtr += 8; whiteSqPtr += 8; blackSqPtr -= 8;
	}

	/* -------------------------------------------------------
		Initialize the piece information arrays.
	------------------------------------------------------- */
	dataPtr->stmPcInfoArray[king].pcHandle =
			&dataPtr->chessBoard.endSTMKingPtr;
	dataPtr->stmPcInfoArray[king].sqType = stmPieceSq0;

	dataPtr->stmPcInfoArray[queen].pcHandle =
			&dataPtr->chessBoard.endSTMLMPiecePtr;
	dataPtr->stmPcInfoArray[queen].pcTable.sqIncrArrayPtr =
			dataPtr->queenSqIncrArray;
	dataPtr->stmPcInfoArray[queen].pcTable.axisArrayPtr =
			dataPtr->queenAxisArray;
	dataPtr->stmPcInfoArray[queen].pcTable.numberOfAxes = 4;
	dataPtr->stmPcInfoArray[queen].sqType = stmPieceSq0;

	dataPtr->stmPcInfoArray[rook].pcHandle =
			&dataPtr->chessBoard.endSTMLMPiecePtr;
	dataPtr->stmPcInfoArray[rook].pcTable.sqIncrArrayPtr =
			dataPtr->rookSqIncrArray;
	dataPtr->stmPcInfoArray[rook].pcTable.axisArrayPtr =
			dataPtr->rookAxisArray;
	dataPtr->stmPcInfoArray[rook].pcTable.numberOfAxes = 2;
	dataPtr->stmPcInfoArray[rook].sqType = stmPieceSq0;

	dataPtr->stmPcInfoArray[bishop].pcHandle =
			&dataPtr->chessBoard.endSTMLMPiecePtr;
	dataPtr->stmPcInfoArray[bishop].pcTable.sqIncrArrayPtr =
			dataPtr->bishopSqIncrArray;
	dataPtr->stmPcInfoArray[bishop].pcTable.axisArrayPtr =
			dataPtr->bishopAxisArray;
	dataPtr->stmPcInfoArray[bishop].pcTable.numberOfAxes = 2;
	dataPtr->stmPcInfoArray[bishop].sqType = stmPieceSq0;

	dataPtr->stmPcInfoArray[knight].sqType = stmPieceSq0;
	dataPtr->stmPcInfoArray[knight].pcHandle =
			&dataPtr->chessBoard.endSTMKnightPtr;

	dataPtr->stmPcInfoArray[pawn].pcHandle =
			&dataPtr->chessBoard.endSTMPawnPtr;
	dataPtr->stmPcInfoArray[pawn].sqType = stmPawnSq0;

	dataPtr->oppPcInfoArray[king].pcHandle =
			&dataPtr->chessBoard.endOppKingPtr;
	dataPtr->oppPcInfoArray[king].sqType = oppPieceSq0;

	dataPtr->oppPcInfoArray[queen].pcHandle =
			&dataPtr->chessBoard.endOppLMPiecePtr;
	dataPtr->oppPcInfoArray[queen].pcTable.sqIncrArrayPtr =
			dataPtr->queenSqIncrArray;
	dataPtr->oppPcInfoArray[queen].pcTable.axisArrayPtr =
			dataPtr->queenAxisArray;
	dataPtr->oppPcInfoArray[queen].pcTable.numberOfAxes = 4;
	dataPtr->oppPcInfoArray[queen].sqType = oppPieceSq0;

	dataPtr->oppPcInfoArray[rook].pcHandle =
			&dataPtr->chessBoard.endOppLMPiecePtr;
	dataPtr->oppPcInfoArray[rook].pcTable.sqIncrArrayPtr =
			dataPtr->rookSqIncrArray;
	dataPtr->oppPcInfoArray[rook].pcTable.axisArrayPtr =
			dataPtr->rookAxisArray;
	dataPtr->oppPcInfoArray[rook].pcTable.numberOfAxes = 2;
	dataPtr->oppPcInfoArray[rook].sqType = oppPieceSq0;

	dataPtr->oppPcInfoArray[bishop].pcHandle =
			&dataPtr->chessBoard.endOppLMPiecePtr;
	dataPtr->oppPcInfoArray[bishop].pcTable.sqIncrArrayPtr =
			dataPtr->bishopSqIncrArray;
	dataPtr->oppPcInfoArray[bishop].pcTable.axisArrayPtr =
			dataPtr->bishopAxisArray;
	dataPtr->oppPcInfoArray[bishop].pcTable.numberOfAxes = 2;
	dataPtr->oppPcInfoArray[bishop].sqType = oppPieceSq0;

	dataPtr->oppPcInfoArray[knight].pcHandle =
			&dataPtr->chessBoard.endOppKnightPtr;
	dataPtr->oppPcInfoArray[knight].sqType = oppPieceSq0;

	dataPtr->oppPcInfoArray[pawn].sqType = oppPawnSq0;

	/* -------------------------------------------------------
		Initialize the piece movement look-up tables.
	------------------------------------------------------- */
	posQueenSqIncrPtr = negQueenSqIncrPtr =
			dataPtr->queenSqIncrArray + kMaxSqDiff;
	posRookSqIncrPtr = negRookSqIncrPtr =
			dataPtr->rookSqIncrArray + kMaxSqDiff;
	posBishopSqIncrPtr = negBishopSqIncrPtr =
			dataPtr->bishopSqIncrArray + kMaxSqDiff;
	posKnightMovePtr = negKnightMovePtr =
			dataPtr->knightMoveArray + kMaxSqDiff;

	for (sqRow = 0; sqRow < 8; sqRow++)
	{
		*posQueenSqIncrPtr++	= 0;
		*negQueenSqIncrPtr--	= 0;
		*posRookSqIncrPtr++		= 0;
		*negRookSqIncrPtr--		= 0;
		*posBishopSqIncrPtr++	= 0;
		*negBishopSqIncrPtr--	= 0;
		*posKnightMovePtr++		= false;
		*negKnightMovePtr--		= false;

		for (sqCol = sqRow ? -7 : +1; sqCol < 8; sqCol++)
		{
			if (sqRow == 0)
			{
				*posQueenSqIncrPtr++	= + 1;
				*negQueenSqIncrPtr--	= - 1;
				*posRookSqIncrPtr++		= + 1;
				*negRookSqIncrPtr--		= - 1;
				*posBishopSqIncrPtr++	=   0;
				*negBishopSqIncrPtr--	=   0;
				*posKnightMovePtr++		= false;
				*negKnightMovePtr--		= false;
			}
			else if (sqRow + sqCol == 0)
			{
				*posQueenSqIncrPtr++	= +15;
				*negQueenSqIncrPtr--	= -15;
				*posRookSqIncrPtr++		=   0;
				*negRookSqIncrPtr--		=   0;
				*posBishopSqIncrPtr++	= +15;
				*negBishopSqIncrPtr--	= -15;
				*posKnightMovePtr++		= false;
				*negKnightMovePtr--		= false;
			}
			else if (sqCol == 0)
			{
				*posQueenSqIncrPtr++	= +16;
				*negQueenSqIncrPtr--	= -16;
				*posRookSqIncrPtr++		= +16;
				*negRookSqIncrPtr--		= -16;
				*posBishopSqIncrPtr++	=   0;
				*negBishopSqIncrPtr--	=   0;
				*posKnightMovePtr++		= false;
				*negKnightMovePtr--		= false;
			}
			else if (sqRow == sqCol)
			{
				*posQueenSqIncrPtr++	= +17;
				*negQueenSqIncrPtr--	= -17;
				*posRookSqIncrPtr++		=   0;
				*negRookSqIncrPtr--		=   0;
				*posBishopSqIncrPtr++	= +17;
				*negBishopSqIncrPtr--	= -17;
				*posKnightMovePtr++		= false;
				*negKnightMovePtr--		= false;
			}
			else
			{
				*posQueenSqIncrPtr++	= 0;
				*negQueenSqIncrPtr--	= 0;
				*posRookSqIncrPtr++		= 0;
				*negRookSqIncrPtr--		= 0;
				*posBishopSqIncrPtr++	= 0;
				*negBishopSqIncrPtr--	= 0;

				/*	A Knight moves only to those squares within two
						rows and columns that a Queen cannot reach. */
				if ((sqRow < 3) &&
					(sqCol > -3) && (sqCol < 3))
				{
					*posKnightMovePtr++	= true;
					*negKnightMovePtr--	= true;
				}
				else
				{
					*posKnightMovePtr++	= false;
					*negKnightMovePtr--	= false;
				}
			}
		}
	}

	dataPtr->queenAxisArray[0]		=  1;
	dataPtr->queenAxisArray[1]		= 15;
	dataPtr->queenAxisArray[2]		= 16;
	dataPtr->queenAxisArray[3]		= 17;

	dataPtr->rookAxisArray[0]			=  1;
	dataPtr->rookAxisArray[1]			= 16;

	dataPtr->bishopAxisArray[0]		= 15;
	dataPtr->bishopAxisArray[1]		= 17;
	
	dataPtr->knightSqDiffArray[0]	= -33;
	dataPtr->knightSqDiffArray[1]	= + 2;
	dataPtr->knightSqDiffArray[2]	= +13;
	dataPtr->knightSqDiffArray[3]	= + 4;
	dataPtr->knightSqDiffArray[4]	= +28;
	dataPtr->knightSqDiffArray[5]	= + 4;
	dataPtr->knightSqDiffArray[6]	= +13;
	dataPtr->knightSqDiffArray[7]	= + 2;

	dataPtr->kingSqDiffArray[0]		= -17;
	dataPtr->kingSqDiffArray[1]		= + 1;
	dataPtr->kingSqDiffArray[2]		= + 1;
	dataPtr->kingSqDiffArray[3]		= +14;
	dataPtr->kingSqDiffArray[4]		= + 2;
	dataPtr->kingSqDiffArray[5]		= +14;
	dataPtr->kingSqDiffArray[6]		= + 1;
	dataPtr->kingSqDiffArray[7]		= + 1;
}



