/*  This algorithm is based on the following idea.  Assuming
  that we are going to have to check many rows or columns
  which don't contain any non-white pixels, it is faster
  to combine all of the pixels in a row or column and look
  at the end result than it is to check each pixel
  individually.  This is done by ORing entire rows or
  columns together for 1-bit and 8-bit deep pixel maps and
  ANDing entire rows or columns together for 32-bit deep
  pixel maps.  The two different methods are necessary
  because for 1-bit and 8-bit pixel maps, white is
  represented by zeros and for 32-bit pixel maps, white
  is represented by ones.

  The mask tables below are used with 1-bit and 8-bit deep
  pixel maps.  They are needed when the left or right side
  of the selection rectangle is not word aligned.
*/

long LeftMask1[32] =
   {0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
    0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
    0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
    0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
    0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
    0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
    0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
    0x0000000F, 0x00000007, 0x00000003, 0x00000001};

long RightMask1[32] =
   {0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
    0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
    0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
    0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
    0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
    0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
    0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
    0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE, 0xFFFFFFFF};

long LeftMask8[4] =
   {0xFFFFFFFF, 0x00FFFFFF, 0x0000FFFF, 0x000000FF};

long RightMask8[4] =
   {0xFF000000, 0xFFFF0000, 0xFFFFFF00, 0xFFFFFFFF};

long DirectWhite = 0x00FFFFFF;  // Value of white pixel
                                // in 32-bit map.

void EnclosingBounds(PixMapHandle pm,
   Rect selection, Rect *enclosingRect)
{
   PixMapPtr   map;
   long        pixelSize, rowBytes, accumulator,
               leftMask, rightMask, baseAddr,
               leftSide, rightSide, topSide, bottomSide,
               numWholeWords, needLeftMask, needRightMask,
               i, j, k, l, m;

   map = *pm;

/*  Compute position of selection rectangle relative to
  upper-left corner of pixel map.  */

   leftSide = selection.left - map->bounds.left;
   rightSide = selection.right - map->bounds.left;
   topSide = selection.top - map->bounds.top;
   bottomSide = selection.bottom - map->bounds.top;

/*  Check validity of selection rectangle.  */

   if ((rightSide <= leftSide) || (bottomSide <= topSide))
   {
      enclosingRect->left = enclosingRect->right =
         enclosingRect->top = enclosingRect->bottom = 0;
      return;
   }

/*  Determine characteristics of pixel map.  */

   rowBytes = map->rowBytes;
   if (rowBytes >= 0) pixelSize = 1;  // BitMap
   else pixelSize = map->pixelSize;   // PixelMap
   rowBytes &= 0x3FFF;                // Strip flags
   baseAddr = (long) map->baseAddr;

/*  Handle 1-bit and 8-bit deep pixel maps with same chunk
  of code.  32-bit deep pixel map handled separately.  */

   if (pixelSize != 32)
   {

/*  Move baseAddr over to the first column of the selection
  rectangle, still keeping it word aligned.  Then determine
  what masks are needed for leftmost and rightmost words
  in the selection and how many whole words there are in
  between.  */

      if (pixelSize == 1)
      {
         baseAddr += (leftSide >> 5) << 2;
         leftMask = LeftMask1[leftSide & 0x1F];
         rightMask = RightMask1[(rightSide - 1) & 0x1F];
         numWholeWords = (rightSide >> 5) -
            ((leftSide + 31) >> 5);
      }
      else
      {
         baseAddr += leftSide & 0xFFFC;
         leftMask = LeftMask8[leftSide & 3];
         rightMask = RightMask8[(rightSide - 1) & 3];
         numWholeWords = (rightSide >> 2) -
            ((leftSide + 3) >> 2);
      }

/*  Set flags indicating what masks are in use.  If the
  left and right boundaries of the selection fall within
  the same word, then take the intersection of the left
  and right masks and only consider one column of words.  */

      needLeftMask = (leftMask + 1 != 0);
      needRightMask = (rightMask + 1 != 0);
      if (numWholeWords < 0)
      {
         leftMask &= rightMask;
         needRightMask = 0;
      }

/*  Find first row with a non-white pixel by ORing the
  whole row together and checking for a non-zero result.  */

      j = topSide;
      accumulator = 0;
      m = baseAddr + j * rowBytes;  // Top-left corner
      do
      {
         k = m;
         if (needLeftMask)
         {
            accumulator |= (*(long *) k) & leftMask;
            k += 4;
         }
         for (i = 0; i < numWholeWords; i++)
         {
            accumulator |= *(long *) k;
            k += 4;
         }
         if (needRightMask)
         {
            accumulator |= (*(long *) k) & rightMask;
         }
         if (accumulator != 0) break;
         m += rowBytes;
      } while (++j < bottomSide);
      if (j == bottomSide)    // Whole selection is white
      {
         enclosingRect->left = enclosingRect->right =
            enclosingRect->top = enclosingRect->bottom = 0;
         return;
      }
                topSide = j;

/*  Find last row with a non-white pixel.  */

      j = bottomSide - 1;
      accumulator = 0;
      m = baseAddr + j * rowBytes;  // Bottom-left corner
      do
      {
         k = m;
         if (needLeftMask)
         {
            accumulator |= (*(long *) k) & leftMask;
            k += 4;
         }
         for (i = 0; i < numWholeWords; i++)
         {
            accumulator |= *(long *) k;
            k += 4;
         }
         if (needRightMask)
         {
            accumulator |= (*(long *) k) & rightMask;
         }
         if (accumulator != 0) break;
         m -= rowBytes;
      } while (--j >= topSide);
      bottomSide = j + 1;

/*  Find leftmost column containing a non-white pixel.  */

      accumulator = 0;
      m = baseAddr + topSide * rowBytes;
      l = 0;
      if (needLeftMask)
      {
         k = m;
         j = topSide;
         do
         {
            accumulator |= (*(long *) k) & leftMask;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != 0) goto leftFound;
         l += 4;
      }
      for (i = 0; i < numWholeWords; i++)
      {
         k = m + l;
         j = topSide;
         do
         {
            accumulator |= *(long *) k;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != 0) goto leftFound;
         l += 4;
      }
      if (needRightMask)
      {
         k = m + l;
         j = topSide;
         do
         {
            accumulator |= (*(long *) k) & rightMask;
            k += rowBytes;
         } while (++j < bottomSide);
      }

/*  When we get to here, we have narrowed down the left-
  most non-white to the word.  The value in the accumulator
  will tell us the exact column of the pixel.  We then
  move baseAddr over to the last column of the selection
  rectangle (word aligned).  */

leftFound:
      if (pixelSize == 1)
      {
         leftSide = (leftSide & 0xFFFFFFE0) + (l << 3);
         while (accumulator >= 0)
         {
            leftSide++;
            accumulator <<= 1;
         }
         baseAddr = (long) map->baseAddr +
            (((rightSide - 1) >> 5) << 2);
      }
      else
      {
         leftSide = (leftSide & 0xFFFFFFFC) + l;
         while ((accumulator & 0xFF000000) == 0)
         {
            leftSide++;
            accumulator <<= 8;
         }
         baseAddr = (long) map->baseAddr +
            ((rightSide - 1) & 0xFFFC);
      }

/*  Find rightmost column containing a non-white pixel.  */

      accumulator = 0;
      m = baseAddr + topSide * rowBytes;
      l = 0;
      if (needRightMask)
      {
         k = m;
         j = topSide;
         do
         {
            accumulator |= (*(long *) k) & rightMask;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != 0) goto rightFound;
         l += 4;
      }
      for (i = 0; i < numWholeWords; i++)
      {
         k = m - l;
         j = topSide;
         do
         {
            accumulator |= *(long *) k;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != 0) goto rightFound;
         l += 4;
      }
      if (needLeftMask)
      {
         k = m - l;
         j = topSide;
         do
         {
            accumulator |= (*(long *) k) & leftMask;
            k += rowBytes;
         } while (++j < bottomSide);
      }
rightFound:
      if (pixelSize == 1)
      {
         rightSide = ((rightSide + 31) & 0xFFFFFFE0) - (l << 3);
         while ((accumulator & 1) == 0)
         {
            rightSide--;
            accumulator >>= 1;
         }
      }
      else
      {
         rightSide = ((rightSide + 3) & 0xFFFFFFFC) - l;
         while ((accumulator & 0x000000FF) == 0)
         {
            rightSide--;
            accumulator >>= 8;
         }
      }
   }

/*  Now for the code which handles 32-bit deep pixel maps.
  For direct pixels white is ones, unlike indexed pixels
  where white is zeros.  We will use the same technique,
  but we will have to AND the rows and columns together.
  We don't have to worry about left and right masks -- in
  32-bit deep pixel maps every pixel is word aligned.  */

   else
   {
      baseAddr += leftSide << 2;
      numWholeWords = rightSide - leftSide;

/*  Find first row.  */

      j = topSide;
      accumulator = DirectWhite;
      m = baseAddr + j * rowBytes;
      do
      {
         k = m;
         i = 0;
         do
         {
            accumulator &= *(long *) k;
            k += 4;
         } while (++i < numWholeWords);
         if (accumulator != DirectWhite) break;
         m += rowBytes;
      } while (++j < bottomSide);
      if (j == bottomSide)       // All white pixels
      {
         enclosingRect->left = enclosingRect->right =
            enclosingRect->top = enclosingRect->bottom = 0;
         return;
      }
      topSide = j;

/*  Find last row.  */

      j = bottomSide - 1;
      accumulator = DirectWhite;
      m = baseAddr + j * rowBytes;
      do
      {
         k = m;
         i = 0;
         do
         {
            accumulator &= *(long *) k;
            k += 4;
         } while (++i < numWholeWords);
         if (accumulator != DirectWhite) break;
         m -= rowBytes;
      } while (--j >= topSide);
      bottomSide = j + 1;

/*  Find leftmost column.  */

      accumulator = DirectWhite;
      m = baseAddr + topSide * rowBytes;
      l = 0;
      i = 0;
      do
      {
         k = m + l;
         j = topSide;
         do
         {
            accumulator &= *(long *) k;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != DirectWhite) break;
         l += 4;
      } while (++i < numWholeWords);
      leftSide += l >> 2;

/*  Find rightmost column.  */

      baseAddr = (long) map->baseAddr +
         (rightSide << 2) - 4;
      accumulator = DirectWhite;
      m = baseAddr + topSide * rowBytes;
      l = 0;
      i = 0;
      do
      {
         k = m - l;
         j = topSide;
         do
         {
            accumulator &= *(long *) k;
            k += rowBytes;
         } while (++j < bottomSide);
         if (accumulator != DirectWhite) break;
         l += 4;
      } while (++i < numWholeWords);
      rightSide -= l >> 2;
   }

/*  Return enclosing rectangle in the pixel map's
  local coordinates.  */

   enclosingRect->left = leftSide + map->bounds.left;
   enclosingRect->right = rightSide + map->bounds.left;
   enclosingRect->top = topSide + map->bounds.top;
   enclosingRect->bottom = bottomSide + map->bounds.top;
}
