/***************************************************************************
	File:		SMTPSenderDoc.cp
	Version:	1.0
	Purpose:	A Document representing a piece of mail
	Author: 	Christopher Haupt (cfh)
				Copyright  1996, Christopher Haupt
	
	Notes:		An SMTP state machine and PowerPlant Event-Loop
				Async client wrapper. Implements a bare-bones 
				SMTP. Notice that it doesn't handle all cases and
				doesn't check for extended return codes or use
				ESMTP commands.
				
	Revision History:
	960201.cfh 	New file
***************************************************************************/

#include "SMTPSenderDoc.h"
#include "SMTPConstants.h"
#include "SMTPDelEPThreadProc.h"

#include <LEndpoint.h>
#include <LInternetMapper.h>
#include <UInternetSupport.h>
#include <UNetworkMessages.h>
#include <UNetworkFactory.h>
#include <LString.h>

const ResIDT	WIND_MailEditor	=	8001;

#pragma mark --- SMTPSenderDoc ---
// ---------------------------------------------------------------------------
//		 ctor/dtor
// ---------------------------------------------------------------------------
//	
SMTPSenderDoc::SMTPSenderDoc(LCommander *inSuper) :
	LSingleDoc(inSuper)
{
	mEndpoint	= nil;
	mStatusPane	= nil;
	mBody		= nil;		
	mCurPos		= 1;
	mMachineReplyState = eStarting;
	mMachineState	= eConnecting;	
	BuildSessionWindow();
}

SMTPSenderDoc::~SMTPSenderDoc()
{
	// tell notifiers to lay off
	ClientIsClosing();

	// close endpoint if any
	if (mEndpoint)
	{
		// try to defer the deletion of the endpoint
		LSimpleThread *aThread = new LSimpleThread((ThreadProc) DeleteEndpointObject, (void*) mEndpoint);
		if (aThread)
			aThread->Resume();	
		else
			delete mEndpoint;
		mEndpoint = nil;
	}
}

// ---------------------------------------------------------------------------
//		 BuildSessionWindow
// ---------------------------------------------------------------------------
//	
void
SMTPSenderDoc::BuildSessionWindow()
{
	// Create window for our document
	mWindow = LWindow::CreateWindow(WIND_MailEditor, this);
	ThrowIfMemFail_(mWindow);
	mWindow->SetSuperCommander(this);	// because the default arrangement
										// used by createwindow doesn't stick
										
	// set up the control listener relationship
	UReanimator::LinkListenerToControls(this,mWindow,WIND_MailEditor);
	
	// get the status window
	mStatusPane = (LCaption *) mWindow->FindPaneByID('STAT');
	
	// add attachments
	AddAttachment(new LUndoer);
	
	mWindow->Show();

}

// ---------------------------------------------------------------------------
//		 AllowSubRemoval
// ---------------------------------------------------------------------------
//	If user tries to close the main window, interpret this as a request
//	to close the connection. The window is not immediately closed; rather
//	it is closed when the connection actually goes away.

Boolean
SMTPSenderDoc::AllowSubRemoval(
	LCommander *inSub)
{
	if (inSub == mWindow && IsIdle() == false) {
		Disconnect();
		return false;
	}
	else
		return LSingleDoc::AllowSubRemoval(inSub);
}

// ---------------------------------------------------------------------------
//		 HandleAsyncMessage
// ---------------------------------------------------------------------------

void
SMTPSenderDoc::HandleAsyncMessage(
	const LAsyncMessage& inMessage)
{
	switch (inMessage.GetMessageType()) {
		case T_DISCONNECT:
		case T_ORDREL:
			mEndpoint->AcceptDisconnect();
			// fall thru
		case T_DISCONNECTCOMPLETE:
			delete this;
			break;

		case T_BINDCOMPLETE:
			if (noErr == inMessage.GetResultCode())
			{
				LInternetDNSAddress remoteAddress(mSMTPHost, kSMTPPort);
				mEndpoint->Connect(remoteAddress);
			}
			break;
			
		case T_CONNECT:
			if (noErr == inMessage.GetResultCode())
				mEndpoint->AutoReceive();
			break;

		case T_DATA:
		case T_EXDATA:
			LDataArrived* data = (LDataArrived*) &inMessage;
			if (data->GetDataSize())
				RunMachine((char *) data->GetDataBuffer(), data->GetDataSize());
			break;
	}
}

// ---------------------------------------------------------------------------
//		 Connect
// ---------------------------------------------------------------------------
//	Create the SMTP interpreter thread and let it do the work of connecting.

void
SMTPSenderDoc::Connect()
{
	mEndpoint = UNetworkFactory::CreateTCPEndpoint(UNetworkFactory::Asynchronous(this));
	ThrowIfNil_(mEndpoint);

	// Initialization: Bind to any local port, then connect to the remote host.
	LInternetIPAddress address(0, 0);
	mEndpoint->Bind(address);
}

// ---------------------------------------------------------------------------
//		 Disconnect
// ---------------------------------------------------------------------------

void
SMTPSenderDoc::Disconnect()
{
	if (mEndpoint)
		mEndpoint->Disconnect(true);
}

// ---------------------------------------------------------------------------
//		 IsIdle
// ---------------------------------------------------------------------------

Boolean
SMTPSenderDoc::IsIdle()
{
	if (mEndpoint == nil)
		return true;
	EEndpointState netState = mEndpoint->GetState();
	return ((netState == endpoint_Idle) || (netState == endpoint_Unbound));
}

// ---------------------------------------------------------------------------
//		 SendMailMessage
// ---------------------------------------------------------------------------
//	Prepare mail message for sending, initiate thread
void
SMTPSenderDoc::SendMailMessage(void)
{
	
	LEditField *aField = nil;
	
	::SetCursor(*::GetCursor(watchCursor));
		
	aField = (LEditField *) mWindow->FindPaneByID('TO  ');
	if (aField)
		aField->GetDescriptor(mTo);

	aField = (LEditField *) mWindow->FindPaneByID('SUBJ');
	if (aField)
		aField->GetDescriptor(mSubject);

	aField = (LEditField *) mWindow->FindPaneByID('FROM');
	if (aField)
		aField->GetDescriptor(mReturnAddress);

	aField = (LEditField *) mWindow->FindPaneByID('HOST');
	if (aField)
		aField->GetDescriptor(mSMTPHost);
	
	LTextEdit *aBodyPane = (LTextEdit *) mWindow->FindPaneByID('BODY');
	if (aBodyPane)
		mBody = aBodyPane->GetTextHandle();

	if (mTo.Length() && mSMTPHost.Length() && mReturnAddress.Length())
	{	
		
		// get it started
		if (mStatusPane) mStatusPane->SetDescriptor("\pConnecting");
		Connect();
	}
	else
		if (mStatusPane) mStatusPane->SetDescriptor("\pPlease fill in all fields!");
		
	::SetCursor(&UQDGlobals::GetQDGlobals()->arrow);	
}


// ---------------------------------------------------------------------------
//		 ListenToMessage
// ---------------------------------------------------------------------------
//	Respond to commands
void
SMTPSenderDoc::ListenToMessage(MessageT inMessage, void * /*ioParam*/)
{
	switch (inMessage) {
		case 'SEND':
		// check to see if the mail system has been inits
		SendMailMessage();
			
		break;
			
		case 'STOP':
			// stop the thread
			if (mStatusPane) mStatusPane->SetDescriptor("\pAborting send");
			if (mEndpoint)
			{
				SendRSET();
				Disconnect();
			}
		break;
	}
}

// ===========================================================================
#pragma mark -
// ---------------------------------------------------------------------------
//		 Run
// ---------------------------------------------------------------------------
// A kinda skanky state machine for SMTP. I know, I know, but it was a hack
// thrown together for the talk in an hour or so. The better way to do this
// is as a nice table driven machine that covers all return codes, etc. That's
// what I use in my production code, inquire within :)
void
SMTPSenderDoc::RunMachine(char *inDataBuffer, Uint32 inDataSize)
{
	Uint32	thePosition = 0;
	Boolean done = false;
	
	while (!done && thePosition < inDataSize) 
		if (ParseReply(inDataBuffer, inDataSize, thePosition))
		{
		
			switch (mLastCode) {
				case 251:	// ok, but non-local user
				case 250:	// success
					switch (mMachineState) {
						case eGreetingLong:
						case eGreetingShort:
							SendMAIL();
							mMachineState = eMailSender;
							break;
							
						case eMailSender:
							if (SendNext())
								mMachineState = eMailDestination;	
							else
								mMachineState = eMailInitiateData;							
							break;
							
						case eMailDestination:
							if (!SendNext())
								mMachineState = eMailInitiateData;
							break;
								
						case eMailInitiateData:
							SendDATA();
							mMachineState = eMailBody;
							break;
							
						case eQuitting:
							SendQUIT();
							mMachineState = eDisconnecting;
							
							break;
							
						case eDisconnecting:
							break;
						
						default:
							Assert_(false);
							break;
					}
					break;
					
				case 354:	// intermediate data 
					if (eMailBody == mMachineState)
					{
						SendHeader();
						SendBody();
						mMachineState = eQuitting;
					}
					else
						Assert_(false);
					break;

				case 550:	// unknown recipient
					if (eMailDestination == mMachineState)
					{
						if (!SendNext())
							mMachineState = eMailInitiateData;
					}
					else
					{
						SendDATA();
						mMachineState = eMailBody;					
					}
					break;
					
				case 500:	// unknown command
					if (eGreetingLong == mMachineState) // try short help
					{
						SendHELO(true);
						mMachineState = eGreetingShort;
					}
					break;	
					
				case 220:	// connection established
					if (eConnecting == mMachineState)
					{
						SendHELO(false);
						mMachineState = eGreetingLong;
					}
					// ignore other phases getting this...
					break;
					
				case 421:	// service not avail, going down
				case 221:	// closing after msg transmission
					mMachineState = eDisconnecting;
					break;
					
				default:
					Assert_(false);	// what did i miss?
					SendRSET();
					mMachineState = eQuitting;
					break;
			}
		}
}

// ---------------------------------------------------------------------------
//		 ParseReply
// ---------------------------------------------------------------------------
//
Boolean
SMTPSenderDoc::ParseReply(char *inBuffer, Uint32 inBufferLen, Uint32& inPos)
{
	Boolean	done = false;
	char	theChar;
	
	
	if (mMachineReplyState == eStarting)
	{
		statusBuffer[0] = 0;
		mMachineReplyState = eGetStatusMode;
	}
			
	while (!done && inPos < inBufferLen) {
		
		theChar = (char) *(inBuffer + inPos);
		++inPos;
		
		switch (mMachineReplyState) {
		
			case eGetStatusMode:
				++statusBuffer[0];
				statusBuffer[statusBuffer[0]] = theChar;
				if (3 == statusBuffer[0])
				{
					mMachineReplyState = eSkipJunkMode;
					::StringToNum((StringPtr) statusBuffer, &mLastCode);

				}
				break;
			
			case eSkipJunkMode:
				if ('\r' == theChar)
					mMachineReplyState = eEOLMode;
				break;
					
			case eEOLMode:
				if ('\n' == theChar)
				{
					done = true;
					mMachineReplyState = eStarting;
				}
				else
					mMachineReplyState = eSkipJunkMode;
				break;				
		}
	} // while
		
	return done;
}


#pragma mark ---SMTP Protocol Functions---
// ---------------------------------------------------------------------------
//		 SendHELO
// ---------------------------------------------------------------------------
// start session (step 0)
void
SMTPSenderDoc::SendHELO(Boolean inUseShort)
{
	Int32 status = 0;
	LStr255 param("HELO ");
	
	// need to get the IP address for this Endpoint
	if (inUseShort)
	{
		param[0] = 4;
	}
	else
	{
		LNetworkAddress *myAddress = (LNetworkAddress *) mEndpoint->GetLocalAddress(nil);
		if (myAddress)
		{
			Str255 tmpStr;
			param += myAddress->GetDescriptor(tmpStr);
			delete myAddress;
		}
	}
	
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
		
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendQUIT
// ---------------------------------------------------------------------------
//
void 
SMTPSenderDoc::SendQUIT(void)
{
	LStr255 param("QUIT");
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendRSET
// ---------------------------------------------------------------------------
//
void 
SMTPSenderDoc::SendRSET(void)
{
	LStr255 param("RSET");
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendNOOP
// ---------------------------------------------------------------------------
//
void 
SMTPSenderDoc::SendNOOP(void)
{
	LStr255 param("NOOP");
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendMAIL
// ---------------------------------------------------------------------------
// initiates mail session (step 1)
void 
SMTPSenderDoc::SendMAIL(void)
{
	LStr255 param("MAIL FROM:");
	
	// now attach the reverse path
	param += '<';
	param += mReturnAddress;
	param += '>';
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendRCPT
// ---------------------------------------------------------------------------
// sends one mail recipient (step 2) (may be repeated)
void 
SMTPSenderDoc::SendRCPT(LStr255 &inRecipient)
{
	LStr255 param("RCPT TO:");
	
	// now attach the reverse path
	param += '<';
	param += inRecipient;
	param += '>';
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendNext
// ---------------------------------------------------------------------------
// parses the mTo data member, sending to each recipient
Boolean 
SMTPSenderDoc::SendNext(void)
{
	if (mTo.Length())
	{
		// multiple destinations are separated by commas (,)
		LStr255 tmpTo;
		Uint8 thePos = mTo.Find(",", 1, mCurPos);
		if (0 == thePos)
			thePos = mTo.Length();
		
		// now extract the string
		tmpTo = mTo(mCurPos, thePos);
		mCurPos = thePos + 1; // set next pos to be 1 beyond comma
				
		// remove tailing comma if any
		if (tmpTo[tmpTo[0]] == ',')
			tmpTo[0] = tmpTo[0] - 1;
			
		// strip trailing white space if any
		while (tmpTo[0] && tmpTo[tmpTo[0]] == ' ')
			--tmpTo[0];
			
		// strip leading white space
		Uint8 i = 1;
		Uint8 j = 0;
		while (i <= tmpTo[0] && tmpTo[i] == ' ')
		{
			++i; ++j;
		}
		if (j)
			tmpTo = tmpTo.Remove(1, j);

		SendRCPT(tmpTo);
		
		return (mTo.Length() > mCurPos);
	}
	else
		return false;
}

// ---------------------------------------------------------------------------
//		 SendDATA
// ---------------------------------------------------------------------------
// this is the "preamble" before the header and body
void 
SMTPSenderDoc::SendDATA(void)
{
	LStr255 param("DATA");
	if (mStatusPane)
		mStatusPane->SetDescriptor((StringPtr) param);
	param += kCRLF;
	Int32 theSize = param.Length();
	mEndpoint->Send(&param[1], theSize);
}

// ---------------------------------------------------------------------------
//		 SendBody
// ---------------------------------------------------------------------------
// parse the body buffer and send line by line, breaking at <CR>
// make sure that and periods (.) at the start of a line are double
// encoded.
void 
SMTPSenderDoc::SendBody(void)
{
	Int32 dataLen 	= ::GetHandleSize(mBody);
	Int32 theMaxSize= 1024 * 2;// 2K at a time
	Int32 theSize 	= theMaxSize;	
	Int32 thePos 	= 0;
	StPointerBlock	buffer(theSize + 4);
	StHandleLocker 	lock(mBody);
	Ptr	  	dataPtr = *mBody;
	Ptr		sendPtr = buffer.mPtr;
	
	if (mStatusPane)
		mStatusPane->SetDescriptor("\pSending Body");
	
	// add linefeeds to the end of \r
	// checks for <>.<>
	Int32 theBufferSize;
	Boolean startOfLine = true;
	
	while(thePos < dataLen)
	{
		if (dataLen - thePos < theSize)
			theSize = dataLen - thePos;
			
		// scan this send segment and byte stuff leading periods
		// and add linefeeds as necessary
		theBufferSize 	= 0;
		sendPtr = buffer.mPtr;
		
		for (Int32 i = 0; i < theSize && theBufferSize < theMaxSize; i++)
		{
			*sendPtr++ = *dataPtr;
			++theBufferSize;
			
			if (startOfLine)
			{	
				if ('.' == *dataPtr)
				{
					*sendPtr++ = '.';
					++theBufferSize;
				}
				startOfLine = false;	
			}
			
			if ('\r' == *dataPtr)
			{
				*sendPtr++ = '\n';
				++theBufferSize;
				startOfLine = true;
			}
			++dataPtr;
			++thePos;
		}

		Int32 tmpBufferSize = theBufferSize;
		mEndpoint->Send(buffer.mPtr, tmpBufferSize);
		while (kOTFlowErr == tmpBufferSize)
		{
			tmpBufferSize = theBufferSize;
			mEndpoint->Send(buffer.mPtr, tmpBufferSize);
		}
		
	}
	
	// send EOM
	LStr255	sendString = "";
	sendString = kEOM;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);
	
	if (mStatusPane)
		mStatusPane->SetDescriptor("\pMessage sent, waiting for confirmation");

}

// ---------------------------------------------------------------------------
//		 SendHeader
// ---------------------------------------------------------------------------
// date,  to, xmailer, reply-to, from, subj
//
void 
SMTPSenderDoc::SendHeader(void)
{
	Int32 	theSize;
	LStr255 tmpString = "";
	LStr255 sendString = "";
// date
	MakeMailDate(tmpString);
	sendString.Assign(kMBDateKey, kMBDateKeyLen);
	sendString += ' ';
	sendString += tmpString;
	sendString += kCRLF;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);
// to
	sendString.Assign(kMBToKey, kMBToKeyLen);
	sendString += ' ';
	sendString += mTo;
	sendString += kCRLF;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);
// xmailer
	sendString.Assign(kMBXMailerKey, kMBXMailerKeyLen);
	sendString += ' ';
	sendString.Append(kMonetNameString,kMonetNameStringLen);
	sendString += kCRLF;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);
// from
	sendString.Assign(kMBFromKey, kMBFromKeyLen);
	sendString += ' ';
	sendString += mReturnAddress;
	sendString += kCRLF;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);
// subject
	sendString.Assign(kMBSubjectKey, kMBSubjectKeyLen);
	sendString += ' ';
	sendString += mSubject;
	sendString += kCRLF;
	theSize = sendString.Length();
	mEndpoint->Send(&sendString[1], theSize);

// end-of-header is an empty line
	theSize = kMBEOHKeyLen;
	mEndpoint->Send(kMBEndOfHeaderKey, theSize);

//
	if (mStatusPane)
		mStatusPane->SetDescriptor("\pSending Header");
}

// ---------------------------------------------------------------------------
//		 MakeMailDate
// ---------------------------------------------------------------------------
// Spiffy RFC compliant date string builder, non-optimized
void
SMTPSenderDoc::MakeMailDate(LStr255 &outDate)
{
	DateTimeRec theDT;
	MachineLocation theML;
	
	::GetTime(&theDT);
	::ReadLocation(&theML);
	
	// day section
	outDate.Assign(kDay[theDT.dayOfWeek - 1], kDayLen);
	outDate += "\p, ";

	// date section
	if (theDT.day < 10)
		outDate += '0';
	outDate += TString<Str15>((Int32)theDT.day);
	outDate += ' ';
	outDate.Append(kMonth[theDT.month - 1], kMonthLen);
	outDate += ' ';
	TString<Str15> theYear((Int32)theDT.year);
	theYear.Remove(1,2);
	outDate += theYear;
	outDate += ' ';
	
	// time section
	if (theDT.hour < 10)
		outDate += '0';
	outDate += TString<Str15>((Int32)theDT.hour);
	outDate += ':';
	
	if (theDT.minute < 10)
		outDate += '0';
	outDate += TString<Str15>((Int32)theDT.minute);
	outDate += ':';

	if (theDT.second < 10)
		outDate += '0';
	outDate += TString<Str15>((Int32)theDT.second);
	outDate += ' ';

	// zone (must mask value of gmt field)
	Int32 theGMT = theML.u.gmtDelta & 0x00FFFFFF;
	// sign extend
	if (theGMT & 0x00800000)
		theGMT |= 0xFF000000;
			
	// this is in seconds east of GMT
	theGMT = theGMT / (60 * 60);
	if (theGMT < 0) 
	{ // east of GMT
		theGMT *= -1;
		outDate += '-';
	}
	else 
	{ //west of gmt
		outDate += '+';
	}
	
	if (theGMT < 10)
		outDate += '0';
	outDate += TString<Str15>((Int32)theGMT);
	outDate += "\p00";
		
}

