//	------------------------------------------------------------------------------------------	
//
//	Copyright  1996 Apple Computer, Inc., All Rights Reserved
//
//
//		You may incorporate this sample code into your applications without
//		restriction, though the sample code has been provided "AS IS" and the
//		responsibility for its operation is 100% yours.  However, what you are
//		not permitted to do is to redistribute the source as "DSC Sample Code"
//		after having made changes. If you're going to re-distribute the source,
//		we require that you make it clear in the source that the code was
//		descended from Apple Sample Code, but that you've made changes.
//
//		Authors:
//			Chris De Salvo
//			Michael Evans
//
//	------------------------------------------------------------------------------------------	

//	------------------------------	Includes

#include <DiskInit.h>
#include <ToolUtils.h>

#include "DrawSprocket.h"
#include "InputSprocket.h"

#include "ErrorHandler.h"
#include "EventHandler.h"
#include "MenuHandler.h"
#include "SprocketInvaders.h"

//	------------------------------	Private Definitions
//	------------------------------	Private Types
//	------------------------------	Private Variables

static SInt32 	gEventTime;
static SInt16	gSleepTime;

//	------------------------------	Private Functions

static void EventDispatch(EventRecord *theEvent);
static Boolean DefaultMouseDown(EventRecord *theEvent);
static Boolean DefaultKeyDown(EventRecord *theEvent);
static Boolean DoDiskEvent(EventRecord *theEvent);
static void DoSuspend(void);
static void EnableDeviceClass(OSType deviceClass, Boolean enable);

//	------------------------------	Public Variables

Boolean				gDone = false;
SInt32				gDeltaTime;
EventHandlerSet		gEventHandlers;
ISpElementReference	gInputElements[numInputs];
GameKeys			gGameKeys;

//	--------------------	EventInit

short
EventInit(void)
{
SInt16		modifiers = 0;
EventRecord	dummy;
OSStatus	theError;

//	This structure defines the input needs that we'll be requesting from InputSprocket
ISpNeed		myNeeds[numInputs] =
{
	{
		"\pGreen Movement",
		200,
		0,
		kISpElementKind_Axis,
		kISpElementLabel_Axis_XAxis,
		0,
		0,
		0,
		0
	},
	{
		"\pGreen Fire",
		201,
		0,
		kISpElementKind_Button,
		kISpElementLabel_Btn_Fire,
		0,
		0,
		0,
		0
	},
	{
		"\pRed Movement",
		200,
		0,
		kISpElementKind_Axis,
		kISpElementLabel_Axis_XAxis,
		0,
		0,
		0,
		0
	},
	{
		"\pRed Fire",
		201,
		0,
		kISpElementKind_Button,
		kISpElementLabel_Btn_Fire,
		0,
		0,
		0,
		0
	},
	{
		"\pAbort",
		202,
		0,
		kISpElementKind_Button,
		kISpElementLabel_Btn_PauseResume,
		0,
		0,
		0,
		0
	}
};

	EventAvail(everyEvent, &dummy);
	modifiers |= dummy.modifiers;
	EventAvail(everyEvent, &dummy);
	modifiers |= dummy.modifiers;
	EventAvail(everyEvent, &dummy);
	modifiers |= dummy.modifiers;

	RegisterEventHandlers(DefaultKeyDown, DefaultKeyDown, DefaultMouseDown, nil, nil);
	gEventHandlers.diskHandler = DoDiskEvent;

	//	This sets up how often we'll call WaitNextEvent() in out event loop.
	//	WNE is still emulated so it's sort of a time waster if you call it too often.
	//	So, we check DblTime which is the user-selected double-click speed.  We make sure
	//	that we check inputs at least often enough to notice double-clicks.
	gEventTime = GetDblTime();
	gEventTime /= 2;

	//	This is how much time we'll give to background apps when we're not actively playing
	//	a game.
	gSleepTime = 32767;

	//	Setup the input sprocket elements
	theError = ISpElement_NewVirtualFromNeeds(numInputs, myNeeds, gInputElements, 0);
	if (theError)
		FatalError("Could not create ISp virtual controls from needs.");

	//	Init InputSprocket and tell it our needs
	theError = ISpInit(numInputs, myNeeds, gInputElements, 'SInv', '0002', 0, 0, 0);
	if (theError)
		FatalError("Could not initialize ISp.");
	
	//	Turn on the keyboard and mouse handlers
	EnableDeviceClass(kISpDeviceClass_Keyboard, true);
	EnableDeviceClass(kISpDeviceClass_Mouse, true);
	
	ISpSuspend();

	return (modifiers);
}

//	--------------------	EventLoop

void
EventLoop(void)
{
EventRecord	theEvent;

static SInt32	lastTime = 0L;
static SInt32	nextTime = 0L;

SInt32			thisTime;
Boolean			wasProcessed;

	while (! gDone)
	{
		thisTime = TickCount();

		//	WNE is still emulated so we throttle the calls to it
		if (thisTime > nextTime)
		{
			//	We don't call WNE while we're playing the game
			if (! gGameInProgress)
			{
				if (WaitNextEvent(everyEvent, &theEvent, gSleepTime, nil))
				{
					//	See if DrawSprocket wants to handle this event
					DSpProcessEvent(&theEvent, &wasProcessed);

					if (! wasProcessed)
						EventDispatch(&theEvent);
				}
				else
				{
					//	Run any idle tasks
					if (gEventHandlers.idleHandler)
					{
						(*gEventHandlers.idleHandler)(&theEvent);
					}
				}
			}
			
			nextTime = TickCount() + gEventTime;
		}

		//	gDeltaTime tells the game engine how much time has passed since the last
		//	game loop.
		if (lastTime > 0)
			gDeltaTime = thisTime - lastTime;
			
		//	We must tell it that at least SOME time has passed.
		if (gDeltaTime < 1)
			gDeltaTime = 1;

		if (gGameInProgress)
		{
			GameLoop();
			gSleepTime = 1;
		}
		else
		{
			gSleepTime = 32767;
		}

		lastTime = TickCount();
	}
}

//	--------------------	EventDispatch

static void
EventDispatch(EventRecord *theEvent)
{
	switch (theEvent->what)
	{
		case mouseDown:
			if (gEventHandlers.clickHandler != nil)
				if ((*gEventHandlers.clickHandler)(theEvent))
				{
					break;
				}
					
			DefaultMouseDown(theEvent);
			 break;
		 
		case keyDown:
			if (gEventHandlers.keyHandler != nil)
				if ((*gEventHandlers.keyHandler)(theEvent))
				{
					break;
				}
			
			DefaultKeyDown(theEvent);
			 break;

		case autoKey:
			if (gEventHandlers.autoKeyHandler != nil)
				if ((*gEventHandlers.autoKeyHandler)(theEvent))
				{
					break;
				}
					
			DefaultKeyDown(theEvent);
			 break;
		 
		case updateEvt:
			if (gEventHandlers.updateHandler != nil)
				if ((*gEventHandlers.updateHandler)(theEvent))
				{
					break;
				}
					
			BeginUpdate((WindowRef) theEvent->message);
			EndUpdate((WindowRef) theEvent->message);
			 break;
		 
		case diskEvt:
			if (gEventHandlers.diskHandler != nil)
				(*gEventHandlers.diskHandler)(theEvent);
			 break;

		case osEvt:
			if (theEvent->message & 0x01000000)		//	Suspend/resume event
			{
				if (theEvent->message & 0x00000001)	//	Resume
				{
				}
				else
				{
					DoSuspend();					//	Suspend
				}
			}
			break;
			
		case kHighLevelEvent:
			AEProcessAppleEvent(theEvent);
			break;
	}
}

//	------------------------------	DefaultMouseDown

static Boolean
DefaultMouseDown(EventRecord *theEvent)
{
SInt16		whatPart;
SInt32		menuResult;
WindowRef	whichWindow;
	
	whatPart = FindWindow(theEvent->where, &whichWindow);
	
	switch (whatPart)
	{
		case inGoAway:
			break;
	
		case inMenuBar:
			DrawMenuBar();
			menuResult = MenuSelect(theEvent->where);
			
			if (HiWord(menuResult) != 0)
				MenuDoChoice(menuResult);
			break;

		case inSysWindow:
			SystemClick(theEvent, whichWindow);
			break;
	}

	return (true);
}

//	------------------------------	DefaultKeyDown

static Boolean
DefaultKeyDown(EventRecord *theEvent)
{
SInt8	theKey;
SInt8	theCode;
SInt32	menuResult;

	theKey = theEvent->message & charCodeMask;
	theCode = (theEvent->message & keyCodeMask) >> 8;
	if ((theEvent->modifiers & cmdKey) != 0)
	{
		menuResult = MenuKey(theKey);
		if (HiWord(menuResult) != 0)
			MenuDoChoice(menuResult);
			
		return (true);
	}

	return (true);
}

//	------------------------------	DoDiskEvent

static Boolean
DoDiskEvent(EventRecord *theEvent)
{
Point	thePoint;
SInt16	value;

	if (theEvent->message & 0xFFFF0000)		//	Error mounting disk
	{
		SetPt(&thePoint, 100, 100);
		value = DIBadMount(thePoint, theEvent->message);
	}
	
	return (true);
}

//	------------------------------	RegisterEventHandlers

void
RegisterEventHandlers(EventHandlerProc keyDown, EventHandlerProc autoKey, EventHandlerProc mouseDown,
					EventHandlerProc update, EventHandlerProc idle)
{
	gEventHandlers.keyHandler = keyDown;
	gEventHandlers.autoKeyHandler = autoKey;
	gEventHandlers.clickHandler = mouseDown;
	gEventHandlers.updateHandler = update;
	gEventHandlers.idleHandler = idle;
}

//	------------------------------	ModifyEventHandlers

void
ModifyEventHandlers(EventHandlerProc keyDown, EventHandlerProc autoKey, EventHandlerProc mouseDown,
					EventHandlerProc update, EventHandlerProc idle)
{
	if (keyDown)
		gEventHandlers.keyHandler = keyDown;
	
	if (autoKey)
		gEventHandlers.autoKeyHandler = autoKey;
		
	if (mouseDown)
		gEventHandlers.clickHandler = mouseDown;
		
	if (update)
		gEventHandlers.updateHandler = update;
		
	if (idle)
		gEventHandlers.idleHandler = idle;
}

//	--------------------	DoSuspend
//
//	This is a small, local event loop.  All it does is fetch screen update
//	events and wait to get put back in the foreground.

static void
DoSuspend(void)
{
EventRecord	theEvent;

	while (true)
	{
		//	Only check for resume and update events.  While doing so, give TONS of time to other apps
		if (WaitNextEvent(osMask, &theEvent, 32767, nil))
		{
			if (theEvent.message & 0x01000000)		//	Suspend/resume event
			{
				if (theEvent.message & 0x00000001)	//	Resume
				{
				Boolean	dummy;

					//	Make sure that DrawSprocket knows that we've resumed
					DSpProcessEvent(&theEvent, &dummy);
				}
				break;
			}
		}
	}
}

//	--------------------	EnableDeviceClass

static void
EnableDeviceClass(OSType deviceClass, Boolean enable)
{
enum
{
	kDeviceList_COUNT = 100
};
	
OSStatus			err;
UInt32				count;
ISpDeviceReference	deviceList[kDeviceList_COUNT];
	
	// NOTE: This is not the correct way to handle the list count thing.  We
	// should actually call once with NULL for the device list, malloc a list of
	// that size, and call again.
	
	err = ISpDevices_ExtractByClass(deviceClass, kDeviceList_COUNT, &count, deviceList);
	
	if (err)
	{
		FatalError(" ISpDevices_ExtractByClass returned some error");
		return;
	}
	
	if (count > kDeviceList_COUNT)
		count = kDeviceList_COUNT;
	
	if (enable)
	{
		err = ISpDevices_Activate(count, deviceList);

		if (err)
		{
			FatalError(" ISpDevices_Activate returned some error");
			return;
		}
	}
	else
	{
		err = ISpDevices_Deactivate(count, deviceList);
		
		if (err)
		{
			FatalError(" ISpDevices_Deactivate returned some error");
			return;
		}
	}
}
