/*	TVector Class methods	*/

#include "Tangrams.h"

/* DEFINITIONS */

#define k_ZeroError (0.001)	/* Needs to be dynamic */

/*	TVector is a polar vector
	Angle is measured in clockwise radians
	Absolute angles are measured relative to the h=0 axis */

float zeroCorrect (float q);
float zeroCorrect (float q)
{
	return (q > -k_ZeroError && q < k_ZeroError) ? 0: q;
}

float angleCorrect (float q);
float angleCorrect (float q)
{
	while (q > pi)
		q = q - 2 * pi;
	while (q < - pi)
		 q = q + 2 * pi;
	return zeroCorrect (q);
}

TVector::TVector()
{
	next = 0;
	set (0,0);
}

TVector::TVector(TVector& v)
{
	*this = v;
}

TVector::TVector(float dir, float len)
{
	set (dir, len);
}

/*	Polar Vector assignment
	Only assigns direction and length */
TVector& TVector::operator = (TVector& aVector)
{
	angle = aVector.angle;
	length = aVector.length;
	return (*this);
}

/*	Polar Vector addition */
TVector& TVector::operator += (TVector& aVector)
{
	MyVertex	p1, p2;
	
	getHV (&p2);
	aVector.getHV (&p1);
	p2.h += p1.h;
	p2.v += p1.v;
	p2.h = zeroCorrect(p2.h);
	p2.v = zeroCorrect(p2.v);
	setHV (&p2);
	return (*this);
}

/*	Polar Vector subtraction */
TVector& TVector::operator -= (TVector& aVector)
{
	MyVertex	p1, p2;
	
	getHV (&p2);
	aVector.getHV (&p1);
	p2.h -= p1.h;
	p2.v -= p1.v;
	p2.h = zeroCorrect(p2.h);
	p2.v = zeroCorrect(p2.v);
	setHV (&p2);
	return (*this);
}

/*	Polar Vector equality */
Boolean TVector::operator == (TVector& aVector)
{
	if (0 == zeroCorrect (aVector.length - length))
		if (0 == angleCorrect (aVector.angle - angle))
			return true;
	return false;
}

/*	Polar Vector angle compare */
Boolean TVector::operator > (TVector& aVector)
{
	if (0 != zeroCorrect (aVector.length - length))
		if (0 < angleCorrect (aVector.angle - angle))
			return true;
	return false;
}

/*	Polar Vector angle compare */
Boolean TVector::operator < (TVector& aVector)
{
	if (0 != zeroCorrect (aVector.length - length))
		if (0 < angleCorrect (aVector.angle - angle))
			return true;
	return false;
}
/*	Test if parallel */
Boolean TVector::operator || (TVector& aVector) 
{
	float diff = angleCorrect (aVector.angle - angle);
	if (0 == diff)
		return true;
	else if (0 == angleCorrect (diff + pi))
		return true;
	else return false;
}

void TVector::set (float dir, float len)
{
	angle = dir; length = len;
}

/*
void TVector::setLength (float len)
{
	length = len;
}
*/
void TVector::setAngle (float dir)
{
	angle = dir;
}

void TVector::setNext (TVector *aVector)
{
	next = aVector;
}

/*	Transformation */
void TVector::rotate (float anAngle)
{
	angle += anAngle;
	if (angle > pi)
		angle -= 2*pi;
	else if (angle < -pi)
		angle += 2*pi;
	angle = zeroCorrect (angle);
}

void TVector::reverse (void)
{
	if (angle > 0)
		angle = zeroCorrect (angle - pi);
	else
		angle = zeroCorrect (angle + pi);
}

void TVector::flip (float around)
{
	angle = angleCorrect (2 * around - angle);
}

/*	Access */
float TVector::getLength (void)
{
	return length;
}

void TVector::setLength (float len)
{
	length = len;
}

float TVector::getAngle (void)
{
	return angle;
}

TVector*  TVector::getNext (void)
{
	return (TVector *)next;
}

/*	Cartesian/Polar conversion */
void TVector::getHV (MyVertex *pt)
{
	pt->h = zeroCorrect(length * sinf(angle));
	pt->v = zeroCorrect(length * cosf(angle));
}

void TVector::setHV (MyVertex *pt)
{
	setHV (pt->h, pt->v);
}

void TVector::setHV (float h, float v)
{
	if (zeroCorrect(v))
	{
	/*	|v| != 0 */
		angle = zeroCorrect(atanf(h/v));
	/*	We have an angle between -pi/2 and pi/2
		If v < 0 we need to put the angle in the opposite quadrant */
		if (v < 0)
			if (angle < 0)
				angle += pi;
			else
				angle -= pi;
	}
	else if (h < 0)
		angle = -pi/2;
	else
		angle = pi/2;

	length = sqrtf(h*h + v*v);
}

/*	Test for Intersection */
Boolean Intercept (TVector *p1start, TVector *p1,
				TVector *p2start, TVector *p2)
{
/*	p1start is the first point on the inside vector p1
	p2start is the first point on the outside vector p2 */
	
	Boolean result = false;

/*	x1 and x2 are the start and end points of the inside vector */
	TVector x1 = *p1start;
	TVector	x2 = *p1;
	x2 += x1;
	
/*	Transform the inside vector so that the outside vector
	lies along the positive h=0 axis */
	x1 -= *p2start;	/* translate inside points so that outside start point = 0,0 */
	x2 -= *p2start;
	x1.length = zeroCorrect (x1.length);
	float	a1 = angleCorrect(x1.angle - p2->angle); /* rotate so that outside lies along the axis */
	float	a2 = angleCorrect(x2.angle - p2->angle);
	
/*	Inside can get outside by being concurrent with it and longer */
	if (a1 == 0 && a2 == 0)
		if (x1.length < 0 || zeroCorrect (x2.length - p2->length) > 0)
			return true;
			
/*	Inside can get outside by starting at axis and going in the wrong direction */
	if ((x1.length == 0 || a1 == 0) && a2 < 0)
			return true;	

	if (	(a1 > 0 != a2 > 0) && /* The endpoints are on opposite sides of the h=0 axis, */
			a1 != 0 && x1.length != 0 &&
			a2 != 0 && x2.length != 0 &&  /* neither is on the axis, */
			(a1 < pi || a2 < pi)) /* and at least one end is above the v=0 axis */

		
		
	{
		float	denom = ((sinf(a1) * x1.length) - (sinf(a2) * x2.length));
		if (denom > k_ZeroError || denom < - k_ZeroError)
		{
			float diff = p2->length - sinf(a1 - a2) * x1.length * x2.length /denom;
			
			if (diff > k_ZeroError)
				result = true;
		}
		else
			result = false;
	}
	return result;
}
