/*
	File:		som_ImagePart.cpp

	Contains:	Sample Part's SOM based part public interface implementation

	Written by:	Steve Smith
	
	Copyright:	 1995 by Apple Computer, Inc., all rights reserved.
	
	----------------------------------------------------------------
	
	Notes:		ImagePart is implemented using a "wrapped" class strategy.
				This means that the SOM class is just a smart delegator and
				the real implementation lives in a C++ class. The reason for
				doing this is twofold. First, this strategy all but elimintates
				the need for you to have to modify the IDL file and run the
				SOM compiler/emitter. Secondly, given that all code changes
				will occur in the C++ class, build turn around time is reduced
				to however fast the compiler can crunch code.
				
				This is not to say that implementing pure SOM classes isn't
				possible, nor necessary. There will be objects (ie. Extensions)
				which need to be exported for other parts to use which will need
				to be written in IDL and SOM implementation files.
*/

// Notification that this is a SOM source file
#define KSS_som_ImagePart_Class_Source

// define underscore field names (ie. _fSelf)
#define VARIABLE_MACROS

// -- Compiler/Preprocessor Switches --

#ifndef _COMPILERDEFS_
#include "CompDefs.h"
#endif

// -- OpenDoc Utilities --

#ifndef _EXCEPT_
// Exceptions define several important macros (ie. CHECKENV)
// which are used in the SOM method dispatch glue. If Except.h
// is not included early enough, exceptions may not be thrown
// correctly when returning from a SOM method with the "ev" parameter set.
#include <Except.h>
#endif

// -- ImagePart Includes --

#ifndef SOM_KSS_som_ImagePart_xih
#include "som_ImagePart.xih"
#endif

#ifndef _IMAGEPART_
#include "ImagePart.h"
#endif

#ifndef _IMAGEPARTDEF_
#include "ImagePartDef.h"
#endif

// -- OpenDoc Includes --

#ifndef _ODTYPES_
#include <ODTypes.h>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#pragma segment somImagePart

//====================================================================
// som_ImagePart
//====================================================================

SOM_Scope void  SOMLINK som_ImagePart__somInit(KSS_som_ImagePart *somSelf)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__somInit");

	// somInit and somUninit methods behave like C++ constructors in that the
	// inherited methods are called automatically. Thus, there is no need to
	// call the parent class' somInit or somUninit.
	
	// There is also no need to set instance variables to zero/NULL
	// since SOM guarantees that a newly constructed object is zeroed.
	
	// Lastly, it is not valid to perform any operation here which may throw
	// an exception.
}

SOM_Scope void  SOMLINK som_ImagePart__somUninit(KSS_som_ImagePart *somSelf)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__somUninit");

	delete _fPart;
}

SOM_Scope void  SOMLINK som_ImagePart__InitPart(KSS_som_ImagePart *somSelf, Environment *ev,
		ODStorageUnit* storageUnit,
		ODPart* partWrapper)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__InitPart");

	SOM_TRY
		// Create the real C++ ImagePart.		
		_fPart = new ImagePart;

		KSS_som_ImagePart_parent_ODPart_InitPart(somSelf,ev,storageUnit,partWrapper);
		_fPart->InitPart(ev,storageUnit,partWrapper);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__InitPartFromStorage(KSS_som_ImagePart *somSelf, Environment *ev,
		ODStorageUnit* storageUnit,
		ODPart* partWrapper)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__InitPartFromStorage");

	SOM_TRY
		// Create the real C++ ImagePart.		
		_fPart = new ImagePart;

		KSS_som_ImagePart_parent_ODPart_InitPartFromStorage(somSelf,ev,storageUnit,partWrapper);
		_fPart->InitPartFromStorage(ev,storageUnit,partWrapper);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__Externalize(KSS_som_ImagePart *somSelf, Environment *ev)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Externalize");

	SOM_TRY
		KSS_som_ImagePart_parent_ODPart_Externalize(somSelf,ev);
		_fPart->Externalize(ev);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope ODSize  SOMLINK som_ImagePart__Purge(KSS_som_ImagePart *somSelf, Environment *ev,
		ODSize size)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Purge");

	ODSize result;

	SOM_TRY
		result = _fPart->Purge(ev,size);
	SOM_CATCH_ALL
		result = 0;
	SOM_ENDTRY
	
	return result;
}

SOM_Scope void  SOMLINK som_ImagePart__Release(KSS_som_ImagePart *somSelf, Environment *ev)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Release");

	SOM_TRY
		KSS_som_ImagePart_parent_ODPart_Release(somSelf,ev);
		_fPart->Release(ev);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__ReleaseAll(KSS_som_ImagePart *somSelf, Environment *ev)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ReleaseAll");

	SOM_TRY
		_fPart->ReleaseAll(ev);
		KSS_som_ImagePart_parent_ODPart_ReleaseAll(somSelf,ev);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope ODExtension*  SOMLINK som_ImagePart__AcquireExtension(KSS_som_ImagePart *somSelf, Environment *ev,
		ODType extensionName)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__GetExtension");

	ODExtension* result;

	SOM_TRY
		// We don't support any extensions, but one of our parent classes might;
		result = KSS_som_ImagePart_parent_ODPart_AcquireExtension(somSelf,ev,extensionName);
	SOM_CATCH_ALL
		result = kODNULL;
	SOM_ENDTRY

	return result;
}

SOM_Scope ODBoolean  SOMLINK som_ImagePart__HasExtension(KSS_som_ImagePart *somSelf, Environment *ev,
		ODType extensionName)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__HasExtension");

	ODBoolean result;

	SOM_TRY
		// We don't support any extensions, but one of our parent classes might;
		result = KSS_som_ImagePart_parent_ODPart_HasExtension(somSelf,ev,extensionName);
	SOM_CATCH_ALL
		result = kODFalse;
	SOM_ENDTRY

	return result;
}

SOM_Scope void  SOMLINK som_ImagePart__ReleaseExtension(KSS_som_ImagePart *somSelf, Environment *ev,
		ODExtension* extension)
{
    KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ReleaseExtension");

	SOM_TRY
		// We don't support any extensions, but one of our parent classes might;
	    KSS_som_ImagePart_parent_ODPart_ReleaseExtension(somSelf,ev,extension);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__FulfillPromise(KSS_som_ImagePart *somSelf, Environment *ev,
		ODStorageUnitView* /*promiseSUView*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FulfillPromise");
}

SOM_Scope void  SOMLINK som_ImagePart__DropCompleted(KSS_som_ImagePart *somSelf, Environment *ev,
		ODPart* ,
		ODDropResult dropResult)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DropCompleted");
}

SOM_Scope ODDragResult  SOMLINK som_ImagePart__DragEnter(KSS_som_ImagePart *somSelf, Environment *ev,
		ODDragItemIterator* /*dragInfo*/,
		ODFacet* /*facet*/,
		ODPoint* /*where*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DragEnter");

	return kODFalse;
}

SOM_Scope ODDragResult  SOMLINK som_ImagePart__DragWithin(KSS_som_ImagePart *somSelf, Environment *ev,
		ODDragItemIterator* /*dragInfo*/,
		ODFacet* /*facet*/,
		ODPoint* /*where*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DragWithin");

	return kODFalse;
}

SOM_Scope void  SOMLINK som_ImagePart__DragLeave(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* /*facet*/,
		ODPoint* /*where*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DragLeave");
}

SOM_Scope ODDropResult  SOMLINK som_ImagePart__Drop(KSS_som_ImagePart *somSelf, Environment *ev,
		ODDragItemIterator* /*dropInfo*/,
		ODFacet* /*facet*/,
		ODPoint* /*where*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Drop");

	return kODDropFail;
}

SOM_Scope ODBoolean  SOMLINK som_ImagePart__RevealFrame(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*embeddedFrame*/,
		ODShape* /*revealShape*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RevealFrame");

	return kODFalse;
}

SOM_Scope ODBoolean  SOMLINK som_ImagePart__EditInLinkAttempted(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
    KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__EditInLinkAttempted");

    return kODFalse;
}

SOM_Scope void  SOMLINK som_ImagePart__EmbeddedFrameUpdated(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame,
		ODUpdateID change)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__EmbeddedFrameChanged");
}

SOM_Scope void  SOMLINK som_ImagePart__EmbeddedFrameSpec(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*embeddedFrame*/,
		ODObjectSpec* /*spec*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__EmbeddedFrameSpec");
}

SOM_Scope void  SOMLINK som_ImagePart__ContainingPartPropertiesUpdated(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame,
		ODStorageUnit* propertyUnit)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ContainingPartPropertiesChanged");
}

SOM_Scope ODEmbeddedFramesIterator*  SOMLINK som_ImagePart__CreateEmbeddedFramesIterator(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CreateEmbeddedFramesIterator");

	// Scripting know if you part supports embedding by calling this method.
	// Since we dont' support embedding, we let ODPart handle it. This is the
	// only "unimplemented" method that we really have to handle.
	KSS_som_ImagePart_parent_ODPart_CreateEmbeddedFramesIterator(somSelf,ev,frame);
	return kODNULL;
}

SOM_Scope void  SOMLINK som_ImagePart__DisplayFrameAdded(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DisplayFrameAdded");

	SOM_TRY
		_fPart->DisplayFrameAdded(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__DisplayFrameRemoved(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DisplayFrameRemoved");

	SOM_TRY
		_fPart->DisplayFrameRemoved(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__DisplayFrameConnected(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DisplayFrameConnected");

	SOM_TRY
		_fPart->DisplayFrameConnected(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__DisplayFrameClosed(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DisplayFrameClosed");

	SOM_TRY
		_fPart->DisplayFrameClosed(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__AttachSourceFrame(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame,
		ODFrame* sourceFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__AttachSourceFrame");

	SOM_TRY
		_fPart->AttachSourceFrame(ev,frame,sourceFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__FrameShapeChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FrameShapeChanged");

	SOM_TRY
		_fPart->FrameShapeChanged(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__ViewTypeChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ViewTypeChanged");

	SOM_TRY
		_fPart->ViewTypeChanged(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__PresentationChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*frame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__PresentationChanged");
}

SOM_Scope void  SOMLINK som_ImagePart__SequenceChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*frame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__SequenceChanged");
}

SOM_Scope void  SOMLINK som_ImagePart__ClonePartInfo(KSS_som_ImagePart *somSelf, Environment *ev,
		ODDraftKey key,
		ODInfoType partInfo,
		ODStorageUnitView* storageUnitView,
		ODFrame* scopeFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ClonePartInfo");

	SOM_TRY
		_fPart->ClonePartInfo(ev,key,partInfo,storageUnitView,scopeFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__WritePartInfo(KSS_som_ImagePart *somSelf, Environment *ev,
		ODInfoType partInfo,
		ODStorageUnitView* storageUnitView)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__WritePartInfo");

	SOM_TRY
		_fPart->WritePartInfo(ev,partInfo,storageUnitView);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope ODInfoType  SOMLINK som_ImagePart__ReadPartInfo(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame,
		ODStorageUnitView* storageUnitView)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ReadPartInfo");

	ODInfoType result;

	SOM_TRY
		result = _fPart->ReadPartInfo(ev,frame,storageUnitView);
	SOM_CATCH_ALL
		result = kODNULL;
	SOM_ENDTRY

	return result;
}

SOM_Scope ODID  SOMLINK som_ImagePart__Open(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Open");

	ODID result;

	SOM_TRY
		result = _fPart->Open(ev,frame);
	SOM_CATCH_ALL
		result = kODNULLID;
	SOM_ENDTRY

	return result;
}

SOM_Scope ODFrame*  SOMLINK som_ImagePart__RequestEmbeddedFrame(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* containingFrame,
		ODFrame* baseFrame,
		ODShape* frameShape,
		ODPart* embeddedPart,
		ODTypeToken viewType,
		ODTypeToken presentation,
		ODBoolean isOverlaid)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RequestEmbeddedFrame");

	return kODNULL;
}

SOM_Scope void  SOMLINK som_ImagePart__RemoveEmbeddedFrame(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*embeddedFrame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RemoveEmbeddedFrame");
}

SOM_Scope ODShape*  SOMLINK som_ImagePart__RequestFrameShape(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*embeddedFrame*/,
		ODShape* /*frameShape*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RequestFrameShape");

	return kODNULL;
}

SOM_Scope void  SOMLINK som_ImagePart__UsedShapeChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*embeddedFrame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__UsedShapeChanged");
}

SOM_Scope ODStorageUnit*  SOMLINK som_ImagePart__AcquireContainingPartProperties(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__GetContainingPartProperties");

	return kODNULL;
}

SOM_Scope ODShape*  SOMLINK som_ImagePart__AdjustBorderShape(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* /*embeddedFacet*/,
		ODShape* /*shape*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__AdjustBorderShape");

	return kODNULL;
}

SOM_Scope void  SOMLINK som_ImagePart__FacetAdded(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* facet)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FacetAdded");

	SOM_TRY
		_fPart->FacetAdded(ev,facet);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__FacetRemoved(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* facet)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FacetRemoved");

	SOM_TRY
		_fPart->FacetRemoved(ev,facet);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__CanvasChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* /*facet*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CanvasChanged");
}

SOM_Scope void  SOMLINK som_ImagePart__GeometryChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* facet,
		ODBoolean clipShapeChanged,
		ODBoolean externalTransformChanged)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__GeometryChanged");

	SOM_TRY
		_fPart->GeometryChanged(ev,facet,clipShapeChanged,externalTransformChanged);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__Draw(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* facet,
		ODShape* invalidShape)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__Draw");

	SOM_TRY
		_fPart->Draw(ev,facet,invalidShape);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__CanvasUpdated(KSS_som_ImagePart *somSelf, Environment *ev,
		ODCanvas* /*canvas*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CanvasUpdated");
}

SOM_Scope void  SOMLINK som_ImagePart__HighlightChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFacet* facet)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__HighlightChanged");

	SOM_TRY
		_fPart->HighlightChanged(ev,facet);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope ODULong  SOMLINK som_ImagePart__GetPrintResolution(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*frame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__GetPrintResolution");

	return kMinImagingResolution;
}

SOM_Scope ODLinkSource*  SOMLINK som_ImagePart__CreateLink(KSS_som_ImagePart *somSelf, Environment *ev,
		ODByteArray* /*data*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CreateLink");

	return kODNULL;
}

SOM_Scope void  SOMLINK som_ImagePart__LinkUpdated(KSS_som_ImagePart *somSelf, Environment *ev,
		ODLink* link,
		ODUpdateID change)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__LinkUpdated");
}

SOM_Scope void  SOMLINK som_ImagePart__RevealLink(KSS_som_ImagePart *somSelf, Environment *ev,
		ODLinkSource* /*linkSource*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RevealLink");
}

SOM_Scope void  SOMLINK som_ImagePart__LinkStatusChanged(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* /*frame*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__LinkStatusChanged");
}

SOM_Scope ODBoolean  SOMLINK som_ImagePart__BeginRelinquishFocus(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeToken focus,
		ODFrame* ownerFrame,
		ODFrame* proposedFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__BeginRelinquishFocus");

	ODBoolean result;

	SOM_TRY
		result = _fPart->BeginRelinquishFocus(ev,focus,ownerFrame,proposedFrame);
	SOM_CATCH_ALL
		result = kODFalse;
	SOM_ENDTRY

	return result;
}

SOM_Scope void  SOMLINK som_ImagePart__CommitRelinquishFocus(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeToken focus,
		ODFrame* ownerFrame,
		ODFrame* proposedFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CommitRelinquishFocus");

	SOM_TRY
		_fPart->CommitRelinquishFocus(ev,focus,ownerFrame,proposedFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__AbortRelinquishFocus(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeToken focus,
		ODFrame* ownerFrame,
		ODFrame* proposedFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__AbortRelinquishFocus");

	SOM_TRY
		_fPart->AbortRelinquishFocus(ev,focus,ownerFrame,proposedFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__FocusAcquired(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeToken focus,
		ODFrame* ownerFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FocusAcquired");

	SOM_TRY
		_fPart->FocusAcquired(ev,focus,ownerFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__FocusLost(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeToken focus,
		ODFrame* ownerFrame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__FocusLost");

	SOM_TRY
		_fPart->FocusLost(ev,focus,ownerFrame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__CloneInto(KSS_som_ImagePart *somSelf, Environment *ev,
		ODDraftKey key,
		ODStorageUnit* toSU,
		ODFrame* scope)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__CloneInto");

	SOM_TRY
		KSS_som_ImagePart_parent_ODPart_CloneInto(somSelf,ev,key,toSU,scope);
		_fPart->CloneInto(ev,key,toSU,scope);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__ExternalizeKinds(KSS_som_ImagePart *somSelf, Environment *ev,
		ODTypeList* kindset)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ExternalizeKinds");

	SOM_TRY
		_fPart->ExternalizeKinds(ev,kindset);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__ChangeKind(KSS_som_ImagePart *somSelf, Environment *ev,
		ODType kind)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__ChangeKind");

	SOM_TRY
		_fPart->ChangeKind(ev,kind);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope ODBoolean  SOMLINK som_ImagePart__HandleEvent(KSS_som_ImagePart *somSelf, Environment *ev,
		ODEventData* event,
		ODFrame* frame,
		ODFacet* facet,
		ODEventInfo* eventInfo)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__HandleEvent");

	ODBoolean result;

	SOM_TRY
		result = _fPart->HandleEvent(ev,event,frame,facet,eventInfo);
	SOM_CATCH_ALL
		result = kODFalse;
	SOM_ENDTRY

	return result;
}

SOM_Scope void  SOMLINK som_ImagePart__AdjustMenus(KSS_som_ImagePart *somSelf, Environment *ev,
		ODFrame* frame)
{
	KSS_som_ImagePartData *somThis = KSS_som_ImagePartGetData(somSelf);
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__AdjustMenus");

	SOM_TRY
		_fPart->AdjustMenus(ev,frame);
	SOM_CATCH_ALL
	SOM_ENDTRY
}

SOM_Scope void  SOMLINK som_ImagePart__UndoAction(KSS_som_ImagePart *somSelf, Environment *ev,
		ODActionData* /*actionState*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__UndoAction");
}

SOM_Scope void  SOMLINK som_ImagePart__RedoAction(KSS_som_ImagePart *somSelf, Environment *ev,
		ODActionData* /*actionState*/)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__RedoAction");
}

SOM_Scope void  SOMLINK som_ImagePart__DisposeActionState(KSS_som_ImagePart *somSelf, Environment *ev,
		ODActionData* ,
		ODDoneState doneState)
{
	KSS_som_ImagePartMethodDebug("KSS_som_ImagePart","som_ImagePart__DisposeActionState");
}
