/********* ********* ********* ********* ********* ********
	
	October 1996 MacTech Programmer's Challenge
	
	Author:		Andy Antoniewicz
				Copyright  1996 By Andy Antoniewicz of
				Mountain View, California USA.
				All rights reserved worldwide.
	
	Problem:
		DNA string match with wildcard & constant distance
	
	Notes:
	No setup calculations on the database are performed.
	It seemed kind of pointless to attempt the equivalent
	of a 50 to 1 lossless compression of the database
	( it was 1000 to 1 for the first problem statement ).

	This is a simple and quick single byte per pass index
	and count type algorithm. It uses three tables:
		a byte to code index table "aTable",
		a fragment index list array "alphaList",
		and a hit count circular array "matchQueue".
	A precursor list is built and used once to build the
	alphaList array, and I do not clean up after ( plenty 
	of allocated storage ).
	
	Once built, the aTable contains an index into the
	alphaList for each given alphabet letter in the
	fragment.  The alphaList contains a sequential list of
	all occurrences of that character in the fragment
	string.
	
	The algorithm then increments the match count for all
	possible alignments of the fragment for each database
	character.  Since the maximum mis-alignment is less
	than the fragment size, only the previous fragmentSize
	database characters need to be considered for matches
	at any particular time.  Hence the circular queue.
	
	The database search execution time is order( pN ) where 
	  N = number of database characters
	  p = average fragment entries per alphabet char
	  ( for given example: A=3,C=3,G=2,T=2 --> p = 2.5 )
	 
	The storage used is almost totally dependent on the
	fragment size. Storage used in bytes
			= 20					( storage struct )
			+ 256					( 16bit aTable )
			+ 2 * fragment chars	( 16bit precursor )
			+ 2 * fragment chars	( 16bit alphaList )
			+ 2 * alphabet chars	(   more alphaList )
			+ 4 * fragment chars	( 2x16bit matchQueue )

 ********* ********* ********* ********* ********* ********/
 
#define	kAlphaTableSize 	128
#define kEndOfAlphaList		-1
 
typedef struct {
	long	storageSize;	// total storage available
	long	usedStorage;	// total storage used
	long	fragmentSize;	// size in chars of fragment
	short	*alphaList;		// start of fragment index list
	short	*matchQueue;	// start of match count queue
	
	short	alphaTable[ kAlphaTableSize ];
	short	precursor[];	// start of match count queue
} DNAStore;

/********* ********* ********* ********* ********* ********
	Function Prototypes
 ********* ********* ********* ********* ********* ********/

void InitMatch(
	char	*alphabet,		// legal characters in database
	char	*database,		// the reference database
	void	*storage,		// pre-allocated storage
	long	storageSize		// size of storage in bytes
);

long FindMatch(				// return number of matches
	char  *alphabet,		// legal characters in database
	char  *database,		// the reference database
	void  *storage,			// pre-allocated storage
	char  *fragment,		// the fragment to find
	long  diffsAllowed,		// num of diffs allowed between
							//   a "match" and the database
	long  matchPosition[]	// match return array
);

void BuildAlphaList(
	char  		*alphabet,	// legal characters in database
	char		*fragment,	// the fragment to find; 0 term
	DNAStore	*storage	// my storage area
);

/********* ********* ********* ********* ********* ********
 *	InitMatch
 *		This routine does nothing except to store the
 *		size of memory allocated by the calling program.
 *		All of the structures used are based on the fragment
 *		to be searched for.
 ********* ********* ********* ********* ********* ********/

void InitMatch(
	char	*alphabet,		// legal characters in database
	char	*database,		// the reference database
	void	*storage,		// pre-allocated storage
	long	storageSize		// size of storage in bytes
)
{
	((DNAStore*)storage)->storageSize = storageSize;
	
}	// end of InitMatch


/********* ********* ********* ********* ********* ********
 *	BuildAlphaList
 *	
 *	This routine has three parts:
 *		Build the AlphaTable
 *			The AlphaTable is an index table used to find
 *			the start location inside the AlphaList
 *			that corresponds to the given alphabet
 *			character.
 *		Build the AlphaList precursor
 *			This is a linked list of the indexes to each
 *			character in the fragment. It is used once
 *			to build the AlphaList and never re-used.
 *		Build the AlphaList
 *			This is a sorted list of indexes for each
 *			alphabet character. The head of each list is
 *			stored in the AlphaTable, and each list is
 *			ended by a -1.
 *			
 ********* ********* ********* ********* ********* ********/

void BuildAlphaList(
	char  	 *alphabet,		// legal characters in database
	char	 *fragment,		// the fragment to find; 0 term
	DNAStore *storage		// my storage area
)
{
	short	*aTable;		// ptr in the AlphaTable
	short	*precursor;		// the AlphaList precursor
	short	*alphaList;		// the AlphaList

	char	*aString;		// ptr to a character string
	long	aChar;			// a character from a string
	
	long	count;			//
	short	index;

    /*********
	 * Initialize alphabet entries to kEndOfAlphaList
	 *	 Proper function for characters not in the alphabet
	 *	 requires that the other entries be preset to zero.
	 *	 This is automatic if the storage is cleared by
	 *	 the calling program and if the alphabet does not
	 *	 change between calls to InitMatch.
	 */
	aTable	= storage->alphaTable;
	aString = alphabet - 1;
	while( (aChar = (long)(*(++aString)) ) > 0x00 )
	{
		*(aTable+aChar) = kEndOfAlphaList;
	}
		
    /*********
	 * Build precursor linked index list
	 *	 This list is only used to produce the 
	 *	 AlphaList below. The precursor and
	 *	 the AlphaList are rebuilt for each
	 *	 new fragment that will be searched for.
	 */
	precursor	= storage->precursor;
	aString		= fragment - 1;
	index		= 0;
	count		= 0;
	while( (aChar = (long)(*(++aString))) > 0x00 )
	{
		index = *(aTable+aChar);	// get prev head
		*(aTable+aChar) = count;	// put new head
		*(precursor+count) = index;	// store prev head
		count++;
	}
	storage->fragmentSize = count;
	
    /*********
	 * Build AlphaList
	 *	 by walking each alphabet character's precursor 
	 *	 list and writing the index list into the 
	 *	 AlphaList the algorithm gets a sorted list of
	 *	 indexes into the fragment for each letter in the
	 *	 alphabet. The aTable will point to the first entry
	 *	 in the AlphaList and the last entry of each 
	 *	 character list is equal to the constant 
	 *	 kEndOfAlphaList ( = -1 ).  Note that AlphaList
	 *	 location 0 is used to ignore database characters
	 *	 which are not in the given alphabet (it was free).
	 */
	alphaList = precursor + count;
	*(alphaList) = kEndOfAlphaList;
	aString = alphabet - 1;
	count = 1;
	while( (aChar = (long)(*(++aString)) ) > 0x00 )
	{
		index = *(aTable+aChar);
		*(aTable+aChar) = count;
		while( index != kEndOfAlphaList )
		{
			*(alphaList+count) = index;
			index = *(precursor+index);
			count++;
		}
		*(alphaList+count) = kEndOfAlphaList;
		count++;
	}
	storage->alphaList	= alphaList;
	storage->matchQueue	= alphaList + count + 1;
	
}	// end of BuildAlphaList


/********* ********* ********* ********* ********* ********
 *	FindMatch
 *
 *	For each character in the database increment every
 *	match count that has the same character ( a hit ).
 *	If after all possible alignments have been tallied,
 *	the match count is greater than or equal to the value
 *	of the threshold, then a match has been found.
 *	Add the matching entry to the return array and continue
 *	until all database characters have been tested.
 *		
 ********* ********* ********* ********* ********* ********/

long FindMatch(				// return number of matches
	char  *alphabet,		// legal characters in database
	char  *database,		// the reference database
	void  *storage,			// pre-allocated storage
	char  *fragment,		// the fragment to find
	long  diffsAllowed,		// num of diffs allowed between
							//   a "match" and the fragment
	long  matchPosition[]	// match return array
)
{
	DNAStore	*theStore;	// typed storage
	short		*matchTop;	// top of match array
	short		*matchCur;	// current match array entry
	short		*aTable;	// alpha to aList index table
	short		*aList;		// array of hit offsets
	short		*hitOffset;	// current hit offset entry
	char		*dbString;	// current database char entry
	long		dbChar;		// current database character
	long		count;		// current database location
	long		numMatch;	// number of matches found
	long		fragSize;	// fragment size in bytes
	long		loop;		// loop counter
	long		hitPos;		// current hit offset position
	long		threshold;	// the threshold for matching
	
	theStore = (DNAStore*)storage;
	BuildAlphaList( alphabet,fragment,theStore);
	fragSize = theStore->fragmentSize;
	matchTop = theStore->matchQueue;
	matchCur = matchTop + fragSize;
	
	theStore->usedStorage
			= (long)(matchCur+fragSize)-(long)theStore;
	/*********
	 * Clear the Match Count Array
	 *	 clear out the match counts for the entire array
	 */
	for( count=0; count<fragSize; count++ )
	{
		*(matchCur+fragSize) = 0;
		*matchCur-- = 0;
	}

	/*********
	 * Walk the Database
	 *
	 *	The match count array is a double size circular
	 *	queue of match counts.  It is double size so I do
	 *	not need to check the array bounds in the inner
	 *	match count increment loop.
	 */
	 
	dbString = database - 1;
	aTable = theStore->alphaTable;
	aList = theStore->alphaList;
	threshold = fragSize-diffsAllowed;

	numMatch = 0;
	count = -fragSize;	// count = current database loc.
	
	while( (dbChar = (long)(*(++dbString))) > 0x00 )
	{
		// circular queue reset to center
		if( matchCur == matchTop )
		{
			matchCur += fragSize;
		}
		
		// check for a match to the fragment
		if((*matchCur+*(matchCur+fragSize)) >= threshold )
		{
			matchPosition[ numMatch ] = count;
			numMatch++;
		}
		
		// clear both old counts
		*matchCur = 0;
		*(matchCur+fragSize) = 0;

		// increment match counts for all possible
		//	 fragment alignments
		hitOffset = aList + *(aTable + dbChar) - 1;
		while( (hitPos = (long)(*(++hitOffset))) >= 0 )
		{
			*(matchCur + hitPos ) += 1;
		}
		
		// The match count queue is walked in reverse
		//	 order through memory because the alphaList
		//	 indexes are positive.
		matchCur--;
		count++;
	}
	
	/*********
	 * Check the remaining match count entries for
	 *	any fragments that extend beyond the end of
	 *	the database.
	 */
	 
	for( loop=0; loop < diffsAllowed; loop++)
	{
		if( matchCur == matchTop )
		{
			matchCur += fragSize;
		}
		
		if((*matchCur+*(matchCur+fragSize)) >= threshold )
		{
			matchPosition[ numMatch ] = count;
			numMatch++;
		}
		matchCur--;
		count++;		
	}
	
	return(numMatch);
}

/********* ********* ********* ********* ********* ********
 ********* ********* ********* ********* ********* ********/
// end of DNA_Match.c
